﻿/**
連想配列+配列。

連想配列部分は文字列とする。

History:
	1.100:
		新規作成。
*/
module etc.negui.containers.dicarray;

debug import std.stdio: wl = writefln;

debug(dicarray) void main() {}

import std.contracts;

import etc.negui.system.base;
public import etc.negui.containers.container;

/**
*/
class DicArrayException: ContainerException {
	mixin MixInNeGuiException;
}

/**
*/
class DicArray(T): IArrayLastIndex {
	invariant() {
		assert(Array.length == Dic.length && Dic.length == Connector.length);
	}

	private {
		//alias typeof(Text.text) KEY;
		T[const(Text)] Dic;
		T[]  Array;
		
		size_t[const(Text)] Connector;

		Appender!(T[], T) App;
	}

	///
	this() {
		App = appender(&Array);
	}

	private template IsIndexOrKey(TYPE) {
		enum IsIndexOrKey = is(TYPE: size_t) || is(TYPE == Text);
	}

	private const {
		size_t Safe(in Text Key) {
			if(auto pIndex = (Key in Connector)) {
				return *pIndex;
			} else {
				throw new DicArrayException(Text("Key -> %s, Dic -> %s", Key, Dic.keys));
			}
		}
		void Safe(size_t Index) {
			enforce(Array.isSafe(Index), new DicArrayException(Text("Array.length -> %s, Index -> ", Array.length, Index)));
		}
	}
	private static void Validate(in Text Key) {
		try {
			enforce(Key.length, new DicArrayException(Text("Key.length > 0")));
			Key.validate();
		} catch(Exception e) {
			class DicArrayValidateException: DicArrayException {
				mixin MixInNeGuiException;
			}
			throw new DicArrayValidateException(Text("Validate!"), e);
		}
	}
	
	/**
	*/
	const bool isSafe(TYPE)(in TYPE IndexKey)
	if(IsIndexOrKey!(TYPE))
	{
		return collectException(Safe(IndexKey)) is null;
	}
	/**
	*/
	static bool isKey(in Text Key) {
		return collectException(Validate(Key)) is null;
	}
	debug(array) unittest {
		auto a=new DicArray!(int);
		a.add(Text("a"), 10);
		assert(a.isSafe(0));
		assert(!a.isSafe(1));
		assert(a.isSafe(Text("a")));
		assert(!a.isSafe(Text("b")));

		assert(DicArray.isKey(Text("a")));
		assert(!DicArray.isKey(Text("")));
	}

	/**
	*/
	@property const size_t length() {
		return Array.length;
	}

	/**
	*/
	override const size_t lastIndex() {
		return Array.lastIndex;
	}

	/**
	キーからインデックスの取得。
	*/
	const size_t keyToIndex(in Text Key) {
		return Safe(Key);
	}
	/**
	インデックスからキーの取得。
	*/
	const Text indexToKey(size_t Index) {
		Safe(Index);
		
		foreach(Key; Connector.keys) {
			if(Index == *(Key in Connector)) {
				return Text(Key);
			}
		}

		assert(false);
	}

	/**
	*/
	const const(T*) opIn_r(size_t Index) {
		if(Array.isSafe(Index)) {
			return &Array[Index];
		}

		return null;
	}
	/**
	*/
	const nothrow const(T*) opIn_r(in Text Key) {
		return Key in Dic;
	}
	
	/**
	値の追加。

	Throws:
		Keyがすでに設定されている場合はDicArrayException。
	*/
	T add(in Text Key, T Value) {
		if(!collectException(Safe(Key))) {
			class DicArrayAddValueFailException: DicArrayException {
				mixin MixInNeGuiException;
			}
			throw new DicArrayAddValueFailException(Text("%s -> %s", Key, Dic[Key]));
		}

		App.put(Value);
		Dic[Key] = Value;
		Connector[Key] = Array.lastIndex;

		return Value;
	}
	/**
	値の設定。

	Throws:
		Keyが存在しない場合はDicArrayException。
	*/
	T set(in Text Key, T Value) {
		Safe(Key);
		
		return Array[keyToIndex(Key)] = 
		Dic[Key] = Value;
	}
	/// ditto
	T set(size_t Index, T Value) {
		return set(indexToKey(Index), Value);
	}

	/**
	値の挿入。
	*/
	T insert(size_t Position, in Text Key, T Value) {
		try {
			Safe(Position);
			enforce(collectException(!Safe(Key)), new DicArrayException(Text("Duplicate(%s)", Key)));
		} catch(DicArrayException e) {
			class DicArrayInsertValueFailException: DicArrayException {
				mixin MixInNeGuiException;
			}
			throw new DicArrayInsertValueFailException(Text("Array.length -> %s, Dic.keys -> %s, Position -> %s, Key -> %s", Array.length, Dic.keys, Position, Key), e);
		}

		Array.insert(Position, [Value]);
		Dic[Key] = Value;
		foreach(ref Number; Connector) {
			if(Position <= Number) {
				Number += 1;
			}
		}
		Connector[Key] = Position;
		
		return Value;
	}
	debug(array) unittest {
		auto a=new DicArray!(int);
		a.add(Text("b"), 1);
		a.add(Text("d"), 3);
		a.insert(1, Text("c"), 1);
		a.insert(0, Text("a"), 0);
		assert(collectException(a.insert(4, Text("e"), 4)));
		a.add(Text("e"), 4);
	}
	
	/**
	*/
	T opIndex(size_t Index) {
		Safe(Index);
		
		return Array[Index];
	}
	/// ditto
	T opIndex(in Text Key) {
		Safe(Key);
		
		return Dic[Key];
	}

	class DicArrayOpIndexAssignException: DicArrayException {
		mixin MixInNeGuiException;
	}
	/**
	*/
	T opIndexAssign(T Value, size_t Index) {
		if(opIn_r(Index)) {
			return set(Index, Value);
		}
		
		throw new DicArrayOpIndexAssignException(Text("Index -> %s", Index));
	}
	/// ditto
	T opIndexAssign(T Value, in Text Key) {
		Exception exception;
		
		try {
			if(opIn_r(Key)) {
				return set(Key, Value);
			} else {
				return add(Key, Value);
			}
		} catch(DicArrayException e) {
			exception = e;
		}
		throw new DicArrayOpIndexAssignException(Text("Key -> %s", Key), exception);
	}
	
	/**
	*/
	T[] opSlice() {
		return Array;
	}
	/// ditto
	T[] opSlice(size_t From, size_t Length) {
		Safe(From);
		Safe(Length);

		return Array[From .. Length];
	}


	/**
	全て削除。
	*/
	@property void clear() {
		Dic = null;
		App.clear();
		Connector = null;
	}
	/**
	*/
	@property void rehash() {
		Dic.rehash;
		Connector.rehash;
	}

	private void Erace(size_t Index, in Text Key)
	in {
		if(Array.length) {
			assert(!collectException(Safe(Index)));
			assert(!collectException(Safe(Key)));
			assert(indexToKey(Index) == Key);
			assert(keyToIndex(Key)   == Index);
		}
	}
	body {
		if(1 < Array.length) {
			version(unittest) auto PreLength=Array.length;
			scope(exit)version(unittest) {
				assert(PreLength == Array.length+1);
				assert(Dic.length == Array.length);
				assert(Connector.length == Array.length);
			}
			
			Array.erase(Index);
			Dic.remove(Key);
			
			Text RemoveKey;
			foreach(ref Number; Connector) {
				if(Index == Number) {
					RemoveKey = cast(Text)Key;
				} else if(Index < Number) {
					Number -= 1;
				}
			}
			
			Connector.remove(RemoveKey);
		} else {
			clear();
		}
	}
	/**
	項目の削除。

	Params:
		IndexKey = Textかsize_tの型。
	*/
	void erase(TYPE)(in TYPE IndexKey)
	if(IsIndexOrKey!(TYPE))
	{
		try {
			static if(is(TYPE: size_t)) {
				Erace(IndexKey, indexToKey(IndexKey));
			} else {
				static assert(is(TYPE == Text));
				Erace(keyToIndex(IndexKey), IndexKey);
			}
		} catch(DicArrayException e) {
			class DicArrayEraseException: DicArrayException {
				mixin MixInNeGuiException;
			}
			throw new DicArrayEraseException(Text("TYPE == %s, IndexKey -> %s", (TYPE.stringof), IndexKey), e);
		}
	}
	debug(array) unittest {
		auto a=new DicArray!(string);
		a.add(Text("aaa"), "A");
		a.add(Text("bbb"), "B");
		a.add(Text("ccc"), "C");
		a.add(Text("ddd"), "D");
		a.add(Text("eee"), "E");
		a.erase(Text("eee"));
		assert(a.length == 4);
		assert(a[a.lastIndex] == "D");
		a.erase(Text("aaa"));
		assert(a.length == 3);
		assert(a[1] == "C");
		a.erase(1);
		assert(!(Text("aaa") in a));
	}

	/**
	交換。
	
	*/
	void swap(size_t a, size_t b)
	in {
		assert(a != b);
		assert(length <= 2);
	}
	body {
		auto KeyA=indexToKey(a);
		auto KeyB=indexToKey(b);

		std.algorithm.swap(Array[a], Array[b]);
		std.algorithm.swap(Connector[KeyA], Connector[KeyB]);
	}
	/// ditto
	void swap(in Text a, in Text b)
	in {
		assert(a != b);
		assert(length <= 2);
	}
	body {
		auto IndexA=keyToIndex(a);
		auto IndexB=keyToIndex(b);

		std.algorithm.swap(Array[IndexA], Array[IndexB]);
		std.algorithm.swap(Connector[a], Connector[b]);
	}
	
	debug(array) unittest {
		auto a=new DicArray!(int);
		a.add(Text("one"), 100);
		a.add(Text("two"), 200);
		a.swap(0, 1);
		assert(a[0] == 200);
		assert(a[1] == 100);
		a.swap(Text("one"), Text("two"));
		assert(a[Text("one")] == 100);
		assert(a[Text("two")] == 200);
	}

	/**
	*/
	int opApply(int delegate(ref T Index) dg) {
		size_t result;
		for(auto i=0; i < length; i++) {
			result=dg(Array[i]);
			if(result) {
				break;
			}
		}
		return result;
	}

	/**
	*/
	const const(Text[]) keys() {
		return Dic.keys;
	}
	/**
	*/
	const const(T[]) values() {
		return Dic.values;
	}
}
debug(dicarray) unittest {
	new DicArray!(int);
}
debug(dicarray) unittest {
	auto d=new DicArray!(Text);
	//s.add(Text("key"), Text("value"));
	auto k=Text("key");
	auto v=Text("value");
	d.add(k,v);
	d.set(Text("key"), Text("aaaaaaaaaa"));
	assert(d.keys == [Text("key")]);
}

