﻿/**
カレキ。

移植及び仕様変更。

Note:
	最適化未考慮。

History:
	1.100:
		新規作成。
*/
module etc.negui.containers.kareki;

debug import std.stdio: wl = writefln;
debug(kareki) void main() {}

import std.contracts;
import std.traits;
import std.conv;
import std.math;
import std.string: hexdigits;
import std.uni;
import std.array;

import etc.negui.system.base;
public import etc.negui.containers.container;

///
class KarekiBaseException: ContainerException {
	mixin MixInNeGuiException;
}

version(X86) {
	/// 整数型(int)
	alias int Integer;
} else version(X86_64) {
	/// 整数型(long)
	alias long Integer;
} else static assert(false, "むり");

/// バイナリ型。
//typedef invariant(ubyte)[] Binary;
typedef ubyte[] Binary;

/**
ファイル型情報に関連付けられる文字。
*/
enum HAGATA: wchar {
	INTEGER     = ':',  /// 整数型。
	REAL        = '*',  /// 実数型。
	TEXT        = '\\', /// 文字列型。
	BINARY      = '?',  /// バイナリ型。

	TSV_INTEGER = '=',  /// TSV形式整数型。
	TSV_REAL    = '@',  /// TSV形式実数型。
	TSV_TEXT    = '$',  /// TSV形式文字列型。
	TSV_BINARY  = '!',  /// TSV形式バイナリ型。
}

///
enum ELEMENT: wchar {
	SECTION = '>',
	COMMENT = ';',
}


///
class ConvertException: KarekiBaseException {
	mixin MixInNeGuiException;
}

private template IsSingle(T, TYPE=Unqual!T) {
	enum IsSingle
		=  is(TYPE: Integer)
		|| is(TYPE: Text)
		|| is(TYPE: real)
		|| is(TYPE: Binary)
	;
}
debug(kareki) unittest {
	assert(IsSingle!(Integer));
	assert(IsSingle!(const(Integer)));
	
	assert(IsSingle!(Text));
	assert(IsSingle!(const(Text)));
	
	assert(IsSingle!(real));
	assert(IsSingle!(const(real)));
	
	assert(IsSingle!(Binary));
	assert(IsSingle!(const(Binary)));

	assert(IsSingle!(short));
	assert(IsSingle!(float));
	assert(!IsSingle!(ubyte[]));
}
private template IsTSV(T: T[], TYPE=T) {
	enum IsTSV = IsSingle!(TYPE);
}

debug(kareki) unittest {
	assert(IsTSV!(Integer[]));
	assert(IsTSV!(const(Integer[])));
	
	assert(IsTSV!(Binary[]));
}

string SMixInTypeToHagata(T, TYPE=Unqual!T)()
if(IsSingle!(TYPE))
{
	static if(is(TYPE: Integer)) {
		return "HAGATA.INTEGER";
	} else static if(is(TYPE: Text)) {
		return "HAGATA.TEXT";
	} else static if(is(TYPE: real)) {
		return "HAGATA.REAL";
	} else static if(is(TYPE: Binary)) {
		return "HAGATA.BINARY";
	}
}
string SMixInTypeToHagata(T: T[], TYPE=Unqual!T)()
if(IsTSV!(T[]))
{
	static if(is(TYPE: Integer)) {
		return "HAGATA.TSV_INTEGER";
	} else static if(is(TYPE: Text)) {
		return "HAGATA.TSV_TEXT";
	} else static if(is(TYPE: real)) {
		return "HAGATA.TSV_REAL";
	} else static if(is(TYPE: Binary)) {
		return "HAGATA.TSV_BINARY";
	}
}

HAGATA TypeToHagata(T, TYPE=Unqual!T)()
if(IsSingle!(TYPE))
{
	return mixin(SMixInTypeToHagata!(TYPE));
}
HAGATA TypeToHagata(T: T[], TYPE=T)()
if(IsTSV!(T[]))
{
	return mixin(SMixInTypeToHagata!(T[]));
}
debug(kareki) unittest {
	assert(TypeToHagata!(Integer) == HAGATA.INTEGER);
	assert(TypeToHagata!(Integer[]) == HAGATA.TSV_INTEGER);
}

/**
*/
struct ENTRY {
	///
	Text name;
	///
	Text data;
	///
	HAGATA type;

	/**
	*/
	const ref Text toText() {
		return Text("[%s] [%s] [%s]", name, type, data);
	}
	/**
	*/
	string toString() {
		return toText.str;
	}
}


/**
*/
struct CONVERT {
	///
	enum :wchar {
		EMPTY = '*', ///
		SEP = '\t',  ///
	}

static:
	/**
	*/
	pure nothrow bool isEmpty(SINGLE_TYPE, TYPE=Unqual!SINGLE_TYPE)(in SINGLE_TYPE Value)
	if(IsSingle!(SINGLE_TYPE))
	body {
		static if(is(SINGLE_TYPE: Integer)) {
			return Value == Integer.min;
		} else static if(is(TYPE: Text)) {
			return !Value.length;
		} else static if(is(SINGLE_TYPE: real)) {
			return isnan(Value);
		} else static if(is(SINGLE_TYPE: Binary)) {
			return !Value.length;
		}
	}
	debug(kareki) unittest {
		assert(isEmpty(Integer.min));
		assert(!isEmpty(Integer.max));
		assert(!isEmpty(Integer.min+1));
		
		assert(isEmpty(Text("")));
		assert(!isEmpty(Text("a")));
		const t=Text('a');
		assert(!isEmpty(t));
		
		assert(isEmpty(real.nan));
		assert(!isEmpty(0.0));
	}
	/**
	*/
	ref Text toText(Integer Value) {
		if(isEmpty(Value)) {
			return Text(EMPTY);
		} else {
			return Text(Value);
		}
	}
	debug(kareki) unittest {
		assert(CONVERT.toText(Integer.min) == Text("*"));
		assert(CONVERT.toText(100) == Text("100"));
		assert(CONVERT.toText(-100) == Text("-100"));
	}
	/// ditto
	ref Text toText(in Text Value) {
		if(isEmpty(Value)) {
			return Text(`\e`);
		} else {
			auto t=Value.replace(Text('\\'), Text(`\\`));
			t.replaceRaw(Text('\t'), Text(`\t`));
			t.replaceRaw(Text('\r'), Text(`\r`));
			return t.replaceRaw(Text('\n'), Text(`\n`));
		}
	}
	debug(kareki) unittest {
		assert(CONVERT.toText(Text(``)) == Text("\\e"));
		assert(CONVERT.toText(Text(`abc`)) == Text("abc"));
		assert(CONVERT.toText(Text("a\r\nb\tc")) == Text(`a\r\nb\tc`));
	}
	/// ditto
	ref Text toText(real Value) {
		if(isEmpty(Value)) {
			return Text(EMPTY);
		} else {
			return Text(Value);
		}
	}
	debug(kareki) unittest {
		assert(CONVERT.toText(real.nan) == Text("*"));
		assert(CONVERT.toText(3.14) == Text("3.14"));
	}
	/// ditto
	ref Text toText(in Binary Value) {
		if(isEmpty(Value)) {
			return Text(EMPTY);
		} else {
			auto s=new wchar[(Value.length << 1) + Value.length / 8];
			wchar h1=void, h2=void;
			for(size_t i=0, j=0; i < Value.length; i++, j+=2) {

				if(Value[i] >= 16) {
					h1 = hexdigits[Value[i] >> 4];
					h2 = hexdigits[Value[i] % 16];
				} else {
					h1 = '0';
					h2 = hexdigits[Value[i]];
				}

				s[j]  = h1;
				s[j+1]= h2;

				if(j+2 < s.length && !((i+1)%8)){
					s[j+2] = (i+1)%16 ? '_': ' ';
					j++;
				}

			}

			return Text.wrap(s);
		}
	}
	debug(kareki) unittest {
		assert(toText(cast(Binary)"ABCDEFGHIJKLMNOPQ") == Text("4142434445464748_494A4B4C4D4E4F50 51"));
	}
	private ref Text ToText(TSV)(in TSV Values)
	if(IsTSV!(TSV))
	{
		auto Results=new Text[Values.length];
		for(auto i=0; i < Results.length; i++) {
			Results[i] = toText(Values[i]);
		}
		return join(Results, Text(SEP));
	}
	private string SMixInToText(TSV)()
	if(IsTSV!(TSV))
	{
		return `
		ref Text toText(in ` ~ TSV.stringof ~ ` Values) {
			return ToText(Values);
		}
		`;
	}
	mixin(SMixInToText!(Integer[]));
	mixin(SMixInToText!(Text[]));
	mixin(SMixInToText!(real[]));
	mixin(SMixInToText!(Binary[]));
	

	/**
	*/
	Integer toValue(T: Integer)(in Text Value) {
		if(Value.length == 1 && Value[0] == EMPTY) {
			return Integer.min;
		}
		
		return cutSeparator(Value).toNumber!(Integer);
	}
	debug(kareki) unittest {
		assert(toValue!(Integer)(Text("123")) == 123);
		assert(toValue!(Integer)(Text("123456")) == 123456);
		assert(toValue!(Integer)(Text("1 2 3 4 5 6")) == 123456);
		assert(toValue!(Integer)(Text("1_2_3_4_5_6")) == 123456);
		assert(toValue!(Integer)(Text("-12345")) == -12345);
		assert(toValue!(Integer)(Text("*")) == Integer.min);
	}
	/// ditto
	real toValue(T: real)(in Text Value) {
		if(Value.length == 1 && Value[0] == EMPTY) {
			return real.nan;
		}
		
		return cutSeparator(Value).toNumber!(real);
	}
	debug(kareki) unittest {
		assert(toValue!(real)(Text("0.123")) == 0.123);
		assert(toValue!(real)(Text("123.456")) == 123.456);
		assert(toValue!(real)(Text("-12.345")) == -12.345);
		assert(isnan(toValue!(real)(Text("*"))));
		assert(toValue!(real)(Text("1e-1")) == 0.1);
	}

	/// ditto
	Text toValue(T: Text)(in Text Value) {
		auto s=Value.text.dup;

		for(size_t i=0; i < s.length; i++)
		if(s[i] == '\\' && i+1 < s.length) {
			auto t=s;

			switch(t[i+1]) {
				case '\\': /+t[i] = '\\';+/ break;
				case 't':  t[i] = '\t'; break;
				case 'r':  t[i] = '\r';  break;
				case 'n':  t[i] = '\n';  break;
				case 'e':
					s = t[0..i]~t[i+2..$];
					i--;
					continue;
				default: t[i] = t[i+1];
			}
			s = t[0..i+1] ~ t[i+2..$];
		}

		return Text.wrap(s);
	}
	debug(kareki) unittest {
		assert(toValue!(Text)(Text(`jy\ngt`)) == Text("jy\ngt"));
		assert(toValue!(Text)(Text(`C\rR`)) == Text(("C\rR")));
		assert(toValue!(Text)(Text(`L\nF`)) == Text(("L\nF")));
		assert(toValue!(Text)(Text(`TA\tB`)) == Text(("TA\tB")));
		assert(toValue!(Text)(Text(`W\\Q`)) == Text(("W\\Q")));
		assert(toValue!(Text)(Text(`a\bc`)) == Text(("abc")));
		assert(toValue!(Text)(Text(`1\e2`)) == Text(("12")));
		assert(toValue!(Text)(Text(`12%3`)) == Text(("12%3")));
		assert(toValue!(Text)(Text(`\e`)) == Text(("")));
		assert(toValue!(Text)(Text(`\e\e`)) == Text(("")));
		assert(toValue!(Text)(Text(`a\eb\ec`)) == Text(("abc")));
	}

	/// ditto
	Binary toValue(T: Binary)(in Text _Value) {

		auto Value=cutSeparator(_Value);

		if(Value.length == 1 && Value[0] == EMPTY) {
			return null;
		}

		enforce(!(Value.length % 2), new ConvertException(Text("%s not even", Value)));
		
		ubyte to10(in wchar[] s)
		in {
			assert(s.length == 2);
		}
		body {
			size_t r, n=void;
			for(size_t i = 0; i < s.length; i++) {
				if(s[i] >= 'A' && s[i] <= 'F') {
					n = s[i] - 'A' + 10;
				} else if(s[i] >= 'a' && s[i] <= 'f') {
					n = s[i] - 'a' + 10;
				} else if('0' <= s[i] && s[i] <= '9') {
					n = s[i] - '0';
				} else {
					throw new ConvertException(Text("%s bad", s));
				}
				r = (r << 4) + n;
			}
			assert(0 <= r && r <= 0xFF);

			return cast(ubyte)r;
		}
	
		auto b=new ubyte[Value.length / 2];

		for(size_t i=0, j=0; i < b.length; i++, j+=2) {
			b[i] = to10(Value.text[j..j+2]);
		}

		return cast(Binary)b;
	}
	debug(kareki) unittest {
		//scope t=new Timer;
		Binary b=[0xaa, 0xff, 0x00, 0x41, 0xC0, 0xFF, 0xEE];
		//foreach(n; 0..10000)assert(toValue!(Binary)("aaff0041C0FFEE")==b, toString(n));
		assert(toValue!(Binary)(Text("aAff0041C0FFEE"))==b);
		assert(toValue!(Binary)(Text("*")) == null);
		try {
			toValue!(Binary)(Text("123"));
			assert(false);
		} catch(ConvertException) {
			assert(true);
		}
		try {
			toValue!(Binary)(Text("Ga"));
			assert(false);
		} catch(ConvertException) {
			assert(true);
		}
		try {
			toValue!(Binary)(Text("aG"));
			assert(false);
		} catch(ConvertException) {
			assert(true);
		}
	}

	/// ditto
	auto toValue(T: T[], TYPE=Unqual!T)(in Text Value)
	if(IsTSV!(TYPE[]))
	{
		enforce(Value.length, new ConvertException(Text("Value.length -> 0")));
		enforce(isTSV(Value), new ConvertException(Text("not tsv -> %s", Value)));
		return toValue!(TYPE[])(splitTSV(Value));
	}
	/// ditto
	auto toValue(T: T[], TYPE=Unqual!T)(in Text[] Values)
	if(IsTSV!(TYPE[]))
	{
		enforce(Values.length, new ConvertException(Text("Values.length -> 0")));
		auto app=appender!(TYPE[]);
		auto tc=THROWABLECHAIN(new ConvertException(Text("toValue, T -> %s, TYPE -> %s", T.stringof, TYPE.stringof)));

		foreach(Value; Values) {
			try {
				app.put(toValue!(TYPE)(Value));
			} catch(Throwable e) {
				tc.chain(e);
			}
		}
		if(auto e=tc()) {
			throw e;
		}

		return app.data;
	}
	debug(kareki) unittest {
		assert(toValue!(real[])([Text("nan")]) == [real.nan]);
	}

	/**
	文字が型情報か調べる。

	Params:
		c = 調べたい文字。

	Returns:
		型情報であればtrue、型情報でなければfalseを返す。
	*/
	pure nothrow bool isHagata(dchar c) {
		switch(c) {
			case
				HAGATA.INTEGER,     HAGATA.TEXT,     HAGATA.REAL,     HAGATA.BINARY,
				HAGATA.TSV_INTEGER, HAGATA.TSV_TEXT, HAGATA.TSV_REAL, HAGATA.TSV_BINARY
			:
				return true;
			default:
				return false;
		}
	}
	debug(kareki) unittest {
		assert(!isHagata('a'));
		assert(isHagata(':'));
		assert(isHagata('*'));
		assert(isHagata('\\'));
		assert(isHagata('?'));
		assert(isHagata('='));
		assert(isHagata('@'));
		assert(isHagata('$'));
		assert(isHagata('!'));
	}

	/**
	*/
	pure nothrow bool isElement(dchar c) {
		switch(c) {
			case ELEMENT.SECTION, ELEMENT.COMMENT:
				return true;
			default:
				return false;
		}
	}
	/**
	*/
	pure nothrow bool isLine(in Text Value) {
		foreach(c; Value.text) {
			if(c == '\r' || c == '\n') {
				return false;
			}
		}

		return true;
	}
	
	private pure nothrow bool IsElementText(in Text Value, ELEMENT Element) {
		if(!Value.length) {
			return false;
		}

		if(Value.text[0] == Element) {
			if(Value.length > 1) foreach(c; Value.text) {
				if(c == '\r' || c == '\n') {
					return false;
				}
			}

			return true;
		}
		
		return false;
	}
	/**
	*/
	pure nothrow bool isSection(in Text Value) {
		return IsElementText(Value, ELEMENT.SECTION);
	}
	/**
	*/
	pure nothrow bool isComment(in Text Value) {
		return IsElementText(Value, ELEMENT.COMMENT);
	}

	/**
	*/
	ref Text commentToText(in Text[] Comments)
	in {
		foreach(Comment; Comments) {
			assert(isComment(Comment), Comment.str);
		}
	}
	body {
		Text[] Results=new Text[Comments.length];
		
		foreach(i, ref Comment; Comments) {
			assert(Comment.length);
			Results[i] = Comment[1 .. Comment.length];
		}

		return join(Results, Text.newline);
	}
	/**
	*/
	ref Text commentToText(in Text Comment) {
		return commentToText(Comment.splitlines);
	}
	debug(kareki) unittest {
		auto s=Texts(";aaa", ";bbb");
		assert(commentToText(s) == Text("%s%s%s", "aaa", Text.newline, "bbb"));
	}
	/**
	*/
	ref Text textToComment(in Text[] Values) {
		wchar[][] Results;
		Results.length = Values.length;
		foreach(i, ref Result; Results) {
			Result.length = Values[i].length + 1;
			Result[0] = ELEMENT.COMMENT;
			if(Result.length > 1) {
				Result[1 .. $] = Values[i].text.dup;
			}
		}
		
		return join(Texts(Results), Text.newline);
	}
	debug(kareki) unittest {
		auto s=Texts("aaa", "bbb");
		assert(textToComment(s) == Text("%s%s%s", ";aaa", Text.newline, ";bbb"),Text("%s%s%s", ";aaa", Text.newline, ";bbb").str);
		s=Texts("", "");
		assert(textToComment(s) == Text("%s%s%s", ";", Text.newline, ";"));
	}

	/**
	*/
	nothrow pure bool isEntry(in Text Value) {
		if(!Value.length) {
			return false;
		}

		foreach(c; Value.text) {
			if(isElement(c) || isHagata(c)) {
				return false;
			}
		}

		return true;
	}

	/**
	*/
	ref ENTRY split(in Text Value) {
		enforce(Value.length >= "eTd".length, new ConvertException(Text("Value.length -> %s", Value.length)));

		auto entry=new ENTRY;

		size_t SplitPoint;
		foreach(i, c; Value.text) {
			if(isHagata(c)) {
				SplitPoint = i;
				break;
			}
		}
		enforce(SplitPoint && Value.length != SplitPoint, new ConvertException(Text("SplitPoint -> %s", SplitPoint)));

		entry.name = Value[0 .. SplitPoint];
		entry.data = Value[SplitPoint+1 .. Value.length];
		assert(isHagata(Value[SplitPoint]));
		entry.type = cast(HAGATA)Value[SplitPoint];

		return *entry;
	}


	/**
	*/
	pure nothrow bool isTSV(in Text Value) {
		if(!Value.length) {
			return false;
		}
		
		wchar c=SEP;
		foreach(s; Value.text) {
			if(s == SEP && s == c) {
				return false;
			}
			c = s;
		}
		return c != SEP;
	}
	debug(kareki) unittest {
		assert(isTSV(Text("t	s	v")));
		assert(isTSV(Text("tsv")));
	}

	/**
	*/
	Text[] splitTSV(in Text Value)
	in {
		assert(isTSV(Value));
	}
	body {
		if(Value.count(SEP)) {
			return Value.split(SEP);
		} else if(Value.length) {
			return [Value.dup];
		} else {
			return null;
		}
	}
	debug(kareki) unittest {
		assert(splitTSV(Text("t	s	v")).length == 3);
		assert(splitTSV(Text("tsv")).length == 1);
	}

	/**
	*/
	ref Text cutSeparator(in Text Value) {
		return Value.filter(['_', ' ']);
	}


	/**
	*/
	void splitDocumentComment(Text[] Lines, out Text[] Comments, out Text[] Datas) {
		size_t i=0;
		while(true) {
			if(!i && !isComment(Lines[i])) {
				Comments = null;
				Datas    = Lines;
				return;
			}

			if(isComment(Lines[i++])) {
				continue;
			}

			if(Lines[i-1].length) {
				Comments = null;
				Datas    =Lines;
				return;
			}

			break;
		}
		Comments = Lines[0..i-1];
		Datas    = Lines[i..Lines.length];
	}

	debug(kareki) unittest {
		auto doc=Text(";document
;comennt
;title

a/b\\c
");
		Text[][] sp;sp.length=2;
			splitDocumentComment(doc.splitlines, sp[0], sp[1]);
			assert(sp[0] == Texts(";document", ";comennt", ";title"));
			assert(sp[1] == Texts(["a/b\\c"]));

		auto doc2=Text("1/2:3
");
			splitDocumentComment(doc2.splitlines, sp[0], sp[1]);
			assert(sp[0] == Texts(""));
			assert(sp[1] == Texts(["1/2:3"]));

	}
	
}

/**
*/
abstract class KarekiBase: IText {
	///
	Text comment;

	/**
	*/
	protected const Text[] CommentLines() {
		return comment.splitlines;
	}

	/**
	*/
	protected const ref Text ToFormatetText(in Text Value) {
		auto Result=new Text;

		if(comment.length) {
			*Result = CONVERT.textToComment(CommentLines) ~ Text.newline;
		} else {
			*Result = Text.init;
		}

		return *Result ~ Value ~ toText;
	}
	version(ITEXT) {
		mixin MixInITextToString!(KarekiBase);
	} else {
		override string toString() {
			return toText.str;
		}
	}
}

///
class EntryException: KarekiBaseException {
	mixin MixInNeGuiException;
}
/**
*/
class Entry: KarekiBase, IText {
	protected {
		///
		HAGATA Type;
		///
		bool HaveValue;
	}
	/**
	*/
	override protected const ref Text ToFormatetText(in Text Value) {
		if(HaveValue) {
			return super.ToFormatetText(Value);
		}
		return Text.init;
	}
	
	mixin(SMixInClassGet!(bool)("have", q{HaveValue}, false));
	/**
	*/
	void del() {
		HaveValue = false;
		_data = null;
	}
	protected const void ExceptionHaveValue(string file=__FILE__, in int line=__LINE__) {
		class EntryNotHaveValueException: EntryException {
			mixin MixInNeGuiException;
		}
		enforce(HaveValue, new EntryNotHaveValueException(Text("not value"), null, file, line));
	}
	protected const void ExceptionTsv(string file=__FILE__, in int line=__LINE__) {
		class EntryNotTsvException: EntryException {
			mixin MixInNeGuiException;
		}
		enforce(isTSV, new EntryNotTsvException(Text("not tsv"), null, file, line));
	}

	/**
	*/
	@property const HAGATA type() {
		ExceptionHaveValue();

		return Type;
	}

	private static string SMixInTypeToVarName(T, TYPE=Unqual!T)()
	if(IsSingle!(TYPE))
	{
		return "_" ~ cast(char)toUniLower(TYPE.stringof[0]);
	}
	private static string SMixInTypeToVarName(T: T[], TYPE=Unqual!T)()
	if(IsTSV!(T[]))
	{
		return SMixInTypeToVarName!(T) ~ "t";
	}
	
	private static string SMixInEntryInit(T)()
	if(IsSingle!(T))
	{
		immutable TYPE=T.stringof;

		immutable TYPES=[TYPE, TYPE ~ "[]"];
		immutable NAMES=[SMixInTypeToVarName!(T), SMixInTypeToVarName!(T[])];

		auto Hagatas = [SMixInTypeToHagata!(T), SMixInTypeToHagata!(T[])];
		
		string s;
		for(auto i=0; i < TYPES.length; i++) {
			s ~= TYPES[i] ~ " " ~ NAMES[i] ~ ";";
			s ~= `
			@property ` ~ TYPES[i] ~ ` data(` ~ TYPES[i] ~ ` Value) {
				Type = ` ~ Hagatas[i] ~ `;
				HaveValue = true;
				return ` ~ NAMES[i] ~ ` = Value;
			}
			`;
		}

		return s;
	}
	union {
		mixin(SMixInEntryInit!(Integer));
		mixin(SMixInEntryInit!(Text));
		mixin(SMixInEntryInit!(real));
		mixin(SMixInEntryInit!(Binary));
		private void* _data;
	}
	private static string SMixInThis(T)()
	if(IsSingle!(T))
	{
		immutable TYPES=[T.stringof, T.stringof ~ "[]"];
		string s;
		for(auto i=0; i < TYPES.length; i++) {
			s ~= `this(` ~ TYPES[i] ~ ` Value) {
				data = Value;
			}
			`;
		}
		return s;
	}

	debug(kareki) unittest {
		auto a=[1,2,3,4];
		auto e=new Entry();
		e.data =a;
		a.length = 3;
		assert(e.get!(int[]) == [1,2,3,4]);
	}

	
	/**
	*/
	this() {
		del();
	}
	mixin(SMixInThis!(Integer));
	mixin(SMixInThis!(Text));
	mixin(SMixInThis!(real));
	mixin(SMixInThis!(Binary));

	debug(kareki) unittest {
		assert((new Entry(1)).type == HAGATA.INTEGER);
		assert((new Entry(Text(""))).type == HAGATA.TEXT);
		assert((new Entry(3.14)).type == HAGATA.REAL);
		assert((new Entry([cast(ubyte)1])).type == HAGATA.BINARY);
		assert((new Entry([1])).type == HAGATA.TSV_INTEGER);
	}

	/**
	*/
	const bool isTSV() {
		ExceptionHaveValue();
		
		final switch(Type) {
			case HAGATA.INTEGER,     HAGATA.TEXT,     HAGATA.REAL,     HAGATA.BINARY:
				return false;
			case HAGATA.TSV_INTEGER, HAGATA.TSV_TEXT, HAGATA.TSV_REAL, HAGATA.TSV_BINARY:
				return true;
		}
	}

	/**
	*/
	const size_t tsvLength() {
		ExceptionTsv();

		switch(Type) {
			case HAGATA.TSV_INTEGER: return _it.length;
			case HAGATA.TSV_REAL   : return _rt.length;
			case HAGATA.TSV_TEXT   : return _tt.length;
			case HAGATA.TSV_BINARY : return _bt.length;
			default: assert(false);
		}
	}

	
	private const ref Text ToText()
	in {
		assert(HaveValue);
	}
	body {
		auto Result=new Text;

		*Result ~= Text(cast(wchar)Type);
		final switch(Type) {
			case HAGATA.INTEGER:
				*Result ~= CONVERT.toText(_i);
				break;
			case HAGATA.TEXT:
				*Result ~= CONVERT.toText(_t);
				break;
			case HAGATA.REAL:
				*Result ~= CONVERT.toText(_r);
				break;
			case HAGATA.BINARY:
				*Result ~= CONVERT.toText(_b);
				break;
			case HAGATA.TSV_INTEGER:
				*Result ~= CONVERT.toText(_it);
				break;
			case HAGATA.TSV_TEXT:
				*Result ~= CONVERT.toText(_tt);
				break;
			case HAGATA.TSV_REAL:
				*Result ~= CONVERT.toText(_rt);
				break;
			case HAGATA.TSV_BINARY:
				*Result ~= CONVERT.toText(_bt);
				break;
		}
		
		return *Result;
	}
	
	/**
	*/
	override const Text toText() {
		if(HaveValue) {
			return ToText;
		}
		return Text.init;
	}
	//override string toString() {return toText.str;}
	

	class EntryValueGetException: EntryException {
		mixin MixInNeGuiException;
	}

	private static string SMixInGet(T)() {
		return `

		assert(HaveValue);
		if(Type == ` ~ SMixInTypeToHagata!(T) ~ `) {
			return cast(` ~ T.stringof ~ `)` ~ SMixInTypeToVarName!(T) ~ `;
		}
		throw new EntryValueGetException(Text("T -> %s, Type -> %s", Text(` ~ SMixInTypeToHagata!(T) ~ `), Text(Type)));

		`;
	}

	/**
	*/
	const auto get(T, TYPE=Unqual!T, string file=__FILE__, int line=__LINE__)()
	if(IsSingle!(TYPE))
	{
		ExceptionHaveValue();
		
		mixin(SMixInGet!(TYPE));
	}
	/// ditto
	const auto get(T: T[], TYPE=Unqual!T, string file=__FILE__, int line=__LINE__)()
	if(IsTSV!(TYPE[]))
	{
		ExceptionHaveValue();
		
		mixin(SMixInGet!(TYPE[]));
	}
	debug(kareki) unittest {
		Entry e;
		e=new Entry([1]);
		assert(e.get!(int[]) == [1]);
	}

	/**
	*/
	const bool tryGet(TYPE)(out TYPE var) {
		if(HaveValue && Type == TypeToHagata!(TYPE)) {
			var = get!(TYPE);
			return true;
		}
		
		return false;
	}

	debug(kareki) unittest {
		Entry e;
		e=new Entry(Text("@"));
		Text a;
		assert(e.tryGet(a));
		assert(a == Text("@"));
		a~=Text("a");
		e=new Entry([1]);
		int[] i;
		assert(e.tryGet(i));
		assert(i == [1]);
	}

	/**
	*/
	const TYPE defGet(TYPE)(TYPE def=TYPE.init)  {
		if(HaveValue && Type == TypeToHagata!(TYPE)) {
			return get!(TYPE);
		}
		
		return def;
	}
	/**
	*/
	const TYPE unSafe(TYPE)() {
		return cast(TYPE)mixin(SMixInTypeToVarName!(TYPE));
	}
	
	/**
	*/
	const Entry dup() {
		auto entry=new Entry();

		// コメント
		entry.comment = comment.dup;
		
		if(!HaveValue) {
			return entry;
		}

		entry.Type = Type;

		final switch(Type) {
			case HAGATA.INTEGER    : entry._i = _i; break;
			case HAGATA.REAL       : entry._r = _r; break;
			case HAGATA.TEXT       : entry._t = _t.dup; break;
			case HAGATA.BINARY     : entry._b = cast(Binary)_b.dup; break;
			case HAGATA.TSV_INTEGER: entry._it = _it.dup; break;
			case HAGATA.TSV_REAL   : entry._rt = _rt.dup; break;
			case HAGATA.TSV_TEXT :
				auto tt=new Text[_tt.length];
				for(auto i=0; i < tt.length; i++)
					tt[i] = _tt[i].dup;
				entry._tt = tt;
				break;
			case HAGATA.TSV_BINARY : 
				auto bt=new Binary[_bt.length];
				for(auto i=0; i < bt.length; i++)
					bt[i] = cast(Binary)_bt[i].dup;
				entry._bt = bt;
				break;
		}
		return entry;
	}

	/**
	*/
	override bool opEquals(Object o) {
		auto entry=cast(Entry)o;

		enforce(entry, new EntryException(Text("opEquals(null)")));
		
		bool OpEquals(T, TYPE=Unqual!T)(in T source, in T target)
		if(IsSingle!(TYPE))
		{
			static if(is(TYPE: real)) {
				// 手抜き。
				return source == target;
			} else {
				return source == target;
			}
		}

		if(Type == entry.Type) {
			if(isTSV) {
				size_t Length=void;
				bool Result=(Length = tsvLength) == entry.tsvLength;

				if(!Result) {
					return false;
				}

				for(auto i=0; i < Length && Result; i++) {
					switch(Type) {
						case HAGATA.TSV_INTEGER: Result = OpEquals(_it[i], entry._it[i]); break;
						case HAGATA.TSV_REAL   : Result = OpEquals(_rt[i], entry._rt[i]); break;
						case HAGATA.TSV_TEXT   : Result = OpEquals(_tt[i], entry._tt[i]); break;
						case HAGATA.TSV_BINARY : Result = OpEquals(_bt[i], entry._bt[i]); break;
						default: assert(false);
					}
				}

				return Result;
			}

			switch(Type) {
				case HAGATA.INTEGER: return OpEquals(_i, entry._i);
				case HAGATA.REAL   : return OpEquals(_r, entry._r);
				case HAGATA.TEXT   : return OpEquals(_t, entry._t);
				case HAGATA.BINARY : return OpEquals(_b, entry._b);
				default: assert(false);
			}
		}

		return false;
	}
	debug(kareki) unittest {
		Entry e1, e2;
		e1 = new Entry(1);
		e2 = new Entry(1);
		assert(e1 == e2);
		e1 = new Entry(1.1);
		e2 = new Entry(1.1);
		assert(e1 == e2);
		e1 = new Entry(Text("abc"));
		e2 = new Entry(Text("abc"));
		assert(e1 == e2);
		e1 = new Entry(cast(Binary)[1,2,3]);
		e2 = new Entry(cast(Binary)[1,2,3]);
		assert(e1 == e2);
		
		assert(e1 != new Entry());
		assert(e1 != new Entry([1,2,3]), "Type != Type");
	}

	/**
	*/
	override int opCmp(Object o) {
		auto entry=cast(Entry)o;

		enforce(entry, new EntryException(Text("opCmp(null)")));

		int OpCmp(T, TYPE=Unqual!T)(in T source, in T target)
		if(IsSingle!(TYPE))
		{
			return Cmp(source, target);
		}

		if(Type == entry.Type) {
			if(isTSV) {
				size_t Length=void;
				int Result=(Length = tsvLength) == entry.tsvLength;

				for(auto i=0; i < Length && Result; i++) {
					switch(Type) {
						case HAGATA.TSV_INTEGER : Result = OpCmp(_it[i], entry._it[i]); break;
						case HAGATA.TSV_REAL    : Result = OpCmp(_rt[i], entry._rt[i]); break;
						case HAGATA.TSV_TEXT    : Result = OpCmp(_tt[i], entry._tt[i]); break;
						case HAGATA.TSV_BINARY  : Result = OpCmp(_bt[i], entry._bt[i]); break;
						default: assert(false);
					}
				}
				
				return Result;
			}

			switch(Type) {
				case HAGATA.INTEGER : return OpCmp(_i, entry._i);
				case HAGATA.REAL    : return OpCmp(_r, entry._r);
				case HAGATA.TEXT    : return OpCmp(_t, entry._t);
				case HAGATA.BINARY  : return OpCmp(_b, entry._b);
				default: assert(false);
			}
		}

		if(Type > entry.Type) {
			return 1;
		}
		
		return -1;
	}
	debug(kareki) unittest {
		Entry e1, e2;
		e1=new Entry(0);
		e2=new Entry(1);
		assert(e1 < e2);
		e1=new Entry(1);
		assert(e1 <= e2);
	}
}

///
class SectionException: KarekiBaseException {
	mixin MixInNeGuiException;
}

/**
*/
class Section: KarekiBase, IText {

	protected {
		///
		Entry[Text] Entrys;
	}
	///
	alias Entrys this;

	/**
	*/
	Entry opIndexAssign(T)(T Value, in Text Key) {
		enforce(CONVERT.isEntry(Key), new SectionException(Text("EntryName -> %s", Key)));
		static if(is(T: Entry)) {
			enforce(Value, new SectionException(Text("Value(T=%s) is null", T.stringof)));
			return Entrys[cast(Text)Key] = Value;
		} else {
			return Entrys[cast(Text)Key] = new Entry(Value);
		}
	}
	debug(kareki) unittest {
		auto s=new Section();
		s[Text("entry")]=new Entry(1);
		s[Text("entry")].comment=Text(123);
		
		assert(s[Text("entry")].get!(Integer) == 1);
		try {
			s[Text("a:9")]=new Entry(1);
			assert(false);
		} catch(SectionException e) {
			assert(true);
		}
		if(auto p=Text("entry") in s) {
			assert((*p).toText == Text(":1"));
		} else {
			assert(false);
		}
		assert(!(s[Text(123)] = new Entry).have);
		assert(s.keys == Texts(["entry", "123"]));
		s[Text(__DATE__)] = new Entry(2.5);
		s[Text(__DATE__)].comment=Text("asd\nsafd");
		s[Text(__FILE__)] = 0;
		s[Text(__FILE__)] = 0.0002;
	}
	/**
	*/
	override protected const ref Text ToFormatetText(in Text Value)
	in {
		assert(CONVERT.isLine(Value));
	}
	body {
		return super.ToFormatetText(Text("%s%s%s", ELEMENT.SECTION, Value, Text.newline));
	}
	/**
	*/
	override const Text toText() {
		auto Results=new Text[Entrys.length];
		size_t i;
		foreach(ref Key; Entrys.keys.sort) {
			auto entry=&Entrys[Key];
			if(entry.have) {
				Results[i++] = entry.ToFormatetText(Key);
			}
		}
		if(i) {
			return join(Results[0 .. i], Text.newline);
		} else {
			return Text.init;
		}
	}
	version(ITEXT) {
		mixin MixInITextToString!(typeof(this));
	}
}

///
class KarekiException: KarekiBaseException {
	mixin MixInNeGuiException;
}
///
class KarekiParseException: KarekiException {
	mixin MixInNeGuiException;
}

/**
*/
class Kareki: KarekiBase {
	
	protected {
		Section[Text] Sections;
	}
	///
	alias Sections this;

	/**
	*/
	Section opIndexAssign(T)(T Value, in Text Key) {
		enforce(Key.length, new KarekiException(Text("SectionName.length -> 0")));
		enforce(CONVERT.isLine(Key), new KarekiException(Text("SectionNameName -> %s", Key)));
		static if(is(T: Section)) {
			enforce(Value, new KarekiException(Text("Value(T=%s) is null", T.stringof)));
			return Sections[cast(Text)Key] = Value;
		} else static if(is(T: Text)) {
			auto section=new Section();
			section.comment = Value;
			return opIndexAssign(section, Key);
		} else static assert(false);
	}


	/**
	*/
	protected const ref Text ToFormatetText() {
		if(comment.length) {
			return CONVERT.textToComment(CommentLines) ~ Text.newline.repeat(2);
		} else {
			return Text.init;
		}
	}

	/**
	*/
	const Text toText() {
		auto Results=new Text[Sections.length];
		size_t i;
		foreach(ref Key; Sections.keys.sort) {
			Results[i++] = Sections[Key].ToFormatetText(Key);
		}
		auto TopComment=ToFormatetText;
		if(i) {
			return TopComment ~ join(Results[0 .. i], Text.newline);
		} else {
			return TopComment;// ~ Text.emptyText;
		}
	}

	private static class DataBlock: IText {
		Text[] Comments;
		Text   Value;
		
		this(in Text Data) {
			enforce(!CONVERT.isComment(Data), new KarekiException(Text("comment:", Data)));
			
			this.Value = Data.dup;
		}

		override const Text toText() {
			return Text("%s, [%s]", Value, Comments);
		}
		debug {
			/+
			override string toString() {
				return toText.str;
			}
			+/
			version(ITEXT) {
				mixin MixInITextToString!(DataBlock);
			}
		}
	}
	private static DataBlock[] MakeDataBlocks(Text[] Datas) {
		//DataBlock[] DataBlocks;
		auto DataBlocks=appender!(DataBlock[]);

		struct COMMENT {
			size_t start;
			size_t end;
			bool now;

			void toStart(size_t pos) {
				now   = true;
				start = end = pos;
			}
		}
		COMMENT Comment;
		
		for(size_t i=0; i < Datas.length; i++) {
			
			if(!Datas[i].length) {
				Comment = Comment.init;
			} else if(CONVERT.isComment(Datas[i])) {
				if(Comment.now) {
					Comment.end = i;
				} else {
					Comment.toStart(i);
				}
			} else {//if(CONVERT.isSection(Datas[i]) || CONVERT.isEntry(Datas[i])) {//}
				auto db = new DataBlock(Datas[i]);
				if(Comment.now) {
					db.Comments = Datas[Comment.start..Comment.end+1];
					//Comment = Comment.init;
				}
				//DataBlocks ~= db;
				DataBlocks.put(db);
			//} else {
				Comment = Comment.init;
			}
		}

		return DataBlocks.data;
	}

	/**
	*/
	this() {}
	/// ditto
	this(in Text Value) {
		load(Value, this);
	}

	/**
	*/
	static Throwable load(in Text Value, out Kareki kareki) {
		auto k=new Kareki;
		
		auto tc=THROWABLECHAIN(new KarekiException(Text("load")));

		if(!Value.length) {
			tc.chain(new KarekiParseException(Text("Value.length -> 0")));
		}

		Text[] Comments, Datas;
		CONVERT.splitDocumentComment(Value.splitlines, Comments, Datas);

		if(Comments.length) {
			k.comment = CONVERT.commentToText(Comments);
		}

		if(Datas.length) {
			auto DataBlocks=MakeDataBlocks(Datas);
			Text NowSectionName;
			
			Section[Text] sections;
			
			foreach(DataBlock; DataBlocks) {
				try {
					if(CONVERT.isSection(DataBlock.Value)) {
						auto SectionName=DataBlock.Value[1 .. DataBlock.Value.length];
						enforce(SectionName.length, new KarekiParseException(Text("SectionName -> %s", SectionName)));
						NowSectionName = SectionName;
						auto section=new Section();
						section.comment = CONVERT.commentToText(DataBlock.Comments);
						k[SectionName] = section;
						NowSectionName = SectionName;
					} else {
						enforce(NowSectionName.length, new KarekiParseException(Text("NowSectionName.length -> 0")));
						auto EntryData=CONVERT.split(DataBlock.Value);
						auto entry=new Entry;
						entry.comment = CONVERT.commentToText(DataBlock.Comments);

						// いつかmixinする
						final switch(EntryData.type) {
							case HAGATA.INTEGER:
								entry.data=CONVERT.toValue!(Integer)(EntryData.data);
								break;
							case HAGATA.TEXT:
								entry.data = CONVERT.toValue!(Text)(EntryData.data);
								break;
							case HAGATA.REAL:
								entry.data = CONVERT.toValue!(real)(EntryData.data);
								break;
							case HAGATA.BINARY:
								entry.data = CONVERT.toValue!(Binary)(EntryData.data);
								break;
							case HAGATA.TSV_INTEGER:
								entry.data=CONVERT.toValue!(Integer[])(EntryData.data);
								break;
							case HAGATA.TSV_TEXT:
								entry.data = CONVERT.toValue!(Text[])(EntryData.data);
								break;
							case HAGATA.TSV_REAL:
								entry.data = CONVERT.toValue!(real[])(EntryData.data);
								break;
							case HAGATA.TSV_BINARY:
								entry.data = CONVERT.toValue!(Binary[])(EntryData.data);
								break;
						}
						k[NowSectionName][EntryData.name] = entry;
					}
				} catch(Throwable e) {
					tc.chain(e);
				}
			}
			kareki = k;
		}
		return tc();
	}
}

