﻿/**
スタック。

基本的なものなんだろうけど作ったことも(明示的に)使ったこともなかったので作ってみた。

History:
	1.100:
		新規作成。
*/
module etc.negui.containers.stack;

debug import std.stdio: wl = writefln;

debug(stack) void main() {}

import std.contracts;

import etc.negui.system.base;
public import etc.negui.containers.container;

///
class StackException: ContainerException {
	mixin MixInNeGuiException;
}

/**
*/
abstract class Stack: IText {
	private size_t Position;
	final mixin(SMixInClassGet!(typeof(Position))("position", q{Position}, false));
	
	/**
	*/
	abstract @property const size_t length();

	version(ITEXT) {
		mixin MixInITextToString!(DateTime);
	} else {
		override string toString() {
			return toText.str;
		}
	}
}

/**
*/
interface IStack(T) {
	/**
	peek/popは使用可能か。
	*/
	@property const nothrow bool peekSafe();
	/**
	pushは使用可能か。
	*/
	@property const nothrow bool pushSafe();
	
	/**
	一番上の値を取得。

	pop()と違い元のスタック自体からは取り出されない。

	Returns:
		取得した値。
	*/
	@property const T peek();
	
	/**
	一番上の値を取り出す。

	Returns:
		取り出した値。
	*/
	@property T pop();
	
	/**
	スタックに積む。

	Params:
		Value = 積む値。
	*/
	@property void push(T Value);
}

/***/
mixin template MixInStack(T, EXCEPTION: StackException) {
	protected T[] Datas;

	/**
	*/
	this(size_t Length) {
		enforce(Length, new EXCEPTION(Text("length -> 0")));
		Datas.length = Length;
	}

	/**
	*/
	override @property const size_t length() {
		return Datas.length;
	}

	/**
	*/
	override @property const T peek() {
		enforce(peekSafe, new EXCEPTION(Text("length -> %s, position -> %s", length, position)));
		return cast(T)Datas[Position-1];
	}

	/**
	*/
	override @property const nothrow bool peekSafe() {
		return Datas.isSafe(Position-1);
	}
	
	/**
	*/
	override @property T pop() {
		enforce(peekSafe, new EXCEPTION(Text("length -> %s, position -> %s", length, position)));
		return Datas[--Position];
	}

	/**
	*/
	const Text toText() {
		Text[] Results=new Text[length];
		foreach(i, Data; Datas[0 .. this.length]) {
			auto Format = i+1 == position ? "%s * %s": "%s - %s";
			static if(is(T: IText)) {
				if(Data) {
					Results[i] = Text(Format, i, Data.toText);
				} else {
					Results[i] = Text(Format, i, Data);
				}
			} else {
				Results[i] = Text(Format, i, Data);
			}
		}
		return Results.join(Text.newline);
	}
}

///
class StaticStackException: StackException {
	mixin MixInNeGuiException;
}

/**

*/
class StaticStack(T): Stack, IStack!(T) {
	mixin MixInStack!(T, StaticStackException);

	/**
	*/
	override void push(T Value) {
		enforce(Datas.length > Position, new StaticStackException(Text("length -> %s, position -> %s", length, position)));
		Datas[Position++] = Value;
	}

	/**
	*/
	override @property const nothrow bool pushSafe() {
		return Datas.length > Position;
	}
}
debug(stack) unittest {
	auto s=new StaticStack!(int)(3);
	s.push(0);
	assert(s.peek == 0);
	assert(s.position == 1);
	assert(s.pop  == 0);
	assert(s.position == 0);
	s.push(0);
	s.push(0);
	s.push(0);
	assert(s.position == 3);
	s.pop;
	s.pop;
	assert(s.peekSafe);
	s.pop;
	assert(!s.peekSafe);
}

class DynamicStackException: StackException {
	mixin MixInNeGuiException;
}
/**

*/
class DynamicStack(T): Stack, IStack!(T) {
	mixin MixInStack!(T, DynamicStackException);

	protected {
		size_t Length;
		size_t Increment;
	}
	final mixin(SMixInClassGetSet!(typeof(Increment))("increment", q{Increment}, false));
	
	/**
	*/
	@property size_t length(size_t Length) {
		enforce(Length, new DynamicStackException(Text("length -> 0")));

		this.Length = Length;
		if(Length < Position) {
			Position = Length;
		}
		return Length;
	}
	/**
	*/
	override @property const size_t length() {
		return Length;
	}
	
	/**
	*/
	override this(size_t Length, size_t Increment=BUFFER.INCREMENT) {
		this.Increment = Increment;
		Datas.length = Length;
		length = Length;
	}

	/**
	*/
	override @property const nothrow bool pushSafe() {
		return true;
	}
	
	/**
	*/
	override void push(T Value) {
		if(Datas.isSafe(Length-1)) {
			Length++;
		}
		Datas.autoIncrement(Position, Increment);
		Datas[Position++] = Value;
	}
}
debug(stack) unittest {
	auto s=new DynamicStack!(int)(3);
	assert(s.position == 0);
	s.push(1);
	assert(s.position == 1);
	s.push(2);
	assert(s.position == 2);
	s.push(3);
	assert(s.position == 3);
	s.push(4);
	assert(s.position == 4);
	s.push(5);
	assert(s.position == 5);
	assert(s.length == 5);
	s.length = 2;
	assert(s.position == 2);
}

debug(stack) unittest {
	auto s=new StaticStack!(DateTime)(10);
	s.push(new DateTime);
	s.push(null);
	s.push(new DateTime(DateTime.MODE.STRICT, 1999));
	assert(s.peek.year == 1999);
}


