﻿/**
ファイルとパス。

Phobosがchar[]大好きだったんで必要分だけ再構築。

std.pathとstd.fileの役割を統合。

History:
	1.000:
		[S] NeGuiが依存している部分あったんでpackageをNeGuiに移動、全然GUIじゃない。

	1.00β16:
		std.fileをimportしないように変更。
*/
module etc.negui.file.file;

debug import std.stdio: wl = writefln;
debug(file) void main() {}

import std.contracts;
import std.contracts;
import std.traits;

import win32.windows;
import win32.shlwapi;

import etc.negui.system.base;
import etc.negui.system.type.record;
import etc.negui.file.find;
import etc.negui.file.find;
import etc.negui.environ.security;
import etc.negui.containers.bit;
import etc.negui.system.utility.number;

unittest {
	assert(!(PATH.MAX_NAME % 4));
	assert(!(PATH.MAX_PATH % 4));
}

///
abstract class NeGuiFileBaseException: NeGuiException {
	mixin MixInNeGuiException;
}
///
class NeGuiPathException: NeGuiFileBaseException {
	mixin MixInNeGuiException;
}
///
class NeGuiFileException: NeGuiFileBaseException {
	mixin MixInNeGuiException;
}
///
class DriveBaseException: NeGuiFileException {
	mixin MixInNeGuiException;
}
///
class DriveException: DriveBaseException {
	mixin MixInNeGuiException;
}
///
class DiskException: DriveBaseException {
	mixin MixInNeGuiException;
}


/**
ファイルパス。

std.pathと違い文字列だけに作用するワケではない。

NeGuiの基準
<ul>
	<li>フォルダで統一。</li>
	<li>フォルダの最後は\</li>
</ul>
*/
static struct PATH {
	static:

	/**
	History:
		1.00β14:
			[S] nemuxi.baseから引越し。
	*/
	enum {
		/// 最大ファイル名長
		MAX_NAME = 255 + 1,
		/// 最大パス長。
		MAX_PATH = 32767 + MAX_NAME + 1,
	}
	/// パス中でのフォルダ区切り文字。
	static immutable Text sep={text:cast(tchar[])[std.path.sep[0]]};
	/// パスとパスの区切り文字。
	static immutable Text pathsep={text:cast(tchar[])[std.path.pathsep[0]]};
	/// カレントフォルダ。
	static immutable Text curdir={text:cast(tchar[])[std.path.curdir[0]]};
	/// 親フォルダを表す文字列。
	static immutable Text pardir={text:cast(tchar[])[std.path.pardir[0], std.path.pardir[1]]};


	/**
	フォルダパス判定。

	NeGui基準で最後がPATH.sepをフォルダとする。

	Params:
		Path = 調べるパス。

	Returns:
		フォルダならtrue、違えばfalse。

	History:
		1.100:
			[S] PATH.sepの使用。
	*/
	bool isFolderPath(in Text Path) {
		//return Path.length && Path[Path.length-1] == '\\';
		return Path.length && Path[Path.length - 1] == sep[0];
	}
	debug(file) unittest {
		assert(isFolderPath(`asd\asd\`w.toText));
		assert(!isFolderPath(`asd\asd`c.toText));
	}
	
	/**
	パスをフォルダに。

	Params:
		Path = パス。

	Returns:
		変換後のパス。
		Pathが元からフォルダならPath自身を返す。

	History:
		1.100:
			[S] 戻り値をrefへ。
			[S] PATH.sepの使用。
	*/
	ref Text addFolderSep(in Text Path) {
		if(isFolderPath(Path)) {
			return cast(Text)Path;
		}
		//return Path ~ Text('\\');
		return Path ~ sep;
	}
	debug(file) unittest {
		assert(addFolderSep(Text(``)) == PATH.sep);
		assert(addFolderSep(Text(`asd\asd\`)) == Text(`asd\asd\`));
		assert(addFolderSep(Text(`asd\asd`)) == Text(`asd\asd\`));
	}
	
	/**
	フォルダから非フォルダパスへ。

	終端のフォルダ区切り文字を削除。

	Params:
		Path = パス。

	Returns:
		変換後のパス。
		Pathの終端がフォルダ区切り文字でない場合はPathを。

	History:
		1.100:
			[S] 戻り値をrefへ。
	*/
	ref Text delFolderSep(in Text Path) {
		if(auto len=Path.length) {
			while(isFolderPath(Path[0 .. len])) {
				if(len--) continue; else break;
			}
			return Path[0 .. len];
		} else {
			return Text.init;
		}
	}
	debug(file) unittest {
		assert(delFolderSep(Text(`asd\asd\`)).text == `asd\asd`);
		assert(delFolderSep(Text(`asd\asd`)).text == `asd\asd`);
		assert(delFolderSep(Text(``)).text == ``);
		assert(delFolderSep(Text(`a`)).text == `a`);
		assert(delFolderSep(Text(`a\\\`)).text == `a`);
		assert(delFolderSep(Text(`LICENSE`)).text == `LICENSE`);
	}
	/**
	パスの結合。

	Params:
		RootPath = 結合元となるパス。

		AddPath = RootPathに結合するパス。
	
	Returns:
		結合後のパス。
	
	History:
		1.081:
			新規作成。
	*/
	Text join(in Text RootPath, in Text[] AddPath ... )
	out(r) {
		if(isFolderPath(AddPath[$-1])) {
			assert(isFolderPath(r));
		} else {
			assert(!isFolderPath(r));
		}
	}
	body {
		auto JoinPath=addFolderSep(RootPath);
		foreach(Path; AddPath) {
			JoinPath ~= addFolderSep(Path);
		}
		if(isFolderPath(AddPath[$-1])) {
			return JoinPath;
		} else {
			return delFolderSep(JoinPath);
		}
	}
	debug(file) unittest {
		Text t1, t2, t3;
		t1 = Text("a");
		t2 = Text("b");
		t3 = Text("c");
		assert(join(t1, t2, t3).text == `a\b\c`);
		t3 ~= Text(`\`);
		assert(join(t1, t2, t3).text == `a\b\c\`);
	}

	/**
	拡張子取得。

	Params:
		Path = 拡張子を取得したいパス。

	Returns:
		拡張子。

	History:
		1.100:
			[S] 戻り値をrefへ。
			[B] 拡張子なしの場合に.なければエラー。
	*/
	ref Text getExtension(in Text Path) {
		auto pStr=Path.ptr(true);
		if(auto p=PathFindExtension(pStr)) {
			if(&pStr[Path.length] == p) {
				return Text.init;
			}
			return Text(cast(wchar*)(p+1));
		} else {
			return Text.init;
		}
	}
	debug(file) unittest {
		assert(getExtension(Text(``)).text == ``);
		
		assert(getExtension(Text(`123456`)).text == ``);
		assert(getExtension(Text(`123.56`)).text == `56`);
		assert(getExtension(Text(`.23.56`)).text == `56`);
		assert(getExtension(Text(`.23456`)).text == `23456`);
		assert(getExtension(Text(`abc\def`)) == Text(``));
		assert(getExtension(Text(`abc\.def`)) == Text(`def`));
		Text t;
		assert(getExtension(t) == Text(``));
	}
	
	/**
	拡張子の変更。

	Params:
		Path = パス名。

		Extension = 変更する拡張子。

	Returns:
		Pathに拡張子が無い場合は.Extensionが追加される。

	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.063:
			新規作成。
	*/
	ref Text changeExtension(in Text Path, in Text Extension) {
		if(auto ExtLength=getExtension(Path).length) {
			return Path[0 .. Path.length - ExtLength] ~ Extension;
		} else {
			return Path ~ Text('.') ~ Extension;
		}
	}
	debug(file) unittest {
		assert(changeExtension(Text(``), Text(``)) == Text(`.`),changeExtension(Text(``), Text(``)).str);
		assert(changeExtension(Text(`abc.def`), Text(`ghi`)) == Text(`abc.ghi`));
		assert(changeExtension(Text(`abc`), Text(`def`)) == Text(`abc.def`));
		assert(changeExtension(Text(`.abc`), Text(`def`)) == Text(`.def`));
	}

	
	/**
	ファイル名取得。

	Params:
		path = ファイル名を取得したいパス。

	Returns:
		ファイル名。
		こいつは拡張子付き。
	
	History:
		1.100:
			[S] 戻り値をrefへ。
	*/
	ref Text getFileName(in Text path) {
		return Text(PathFindFileName(path.ptr));
	}
	debug(file) unittest {
		assert(getFileName(Text(``)).text == ``);
		assert(getFileName(Text(`Z:\aaa.txt`)).text == `aaa.txt`);
		assert(getFileName(Text(`Z:\aaatxt`)).text == `aaatxt`);
		assert(getFileName(Text(`Z:\aaatxt\`)).text == `aaatxt\`);
		assert(getFileName(Text(`LICENSE`)).text == `LICENSE`);
	}
	/**
	ファイル名取得。

	Params:
		path = ファイル名を取得したいパス。

	Returns:
		ファイル名。
		こいつは拡張子無し。

	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.00β16:
			[F] 拡張子のみ(.htaccess等の隠しファイル)をファイル名として扱うように変更。
	
		1.00β11:
			スライスで死ぬ対策。
	*/
	ref Text getName(in Text path) {
		auto name=delFolderSep(getFileName(path));
		
		auto ext =getExtension(path);
		
		if(ext.length && name.length > ext.length+1) {
			return name[0..name.length - ext.length-1];
		} else {
			return name[];
		}
	}
	debug(file) unittest {
		assert(getName(Text(``)).text == ``);
		assert(getName(Text(`.`)).text == `.`);
		assert(getName(Text(`.aaa`)).text == `.aaa`);
		assert(getName(Text(`Z:\aaa.txt`)).text == `aaa`);
		assert(getName(Text(`Z:\aaatxt\`)).text == `aaatxt`);
		assert(getName(Text(`aaabbb`)).text == `aaabbb`);
		assert(getName(Text(`aaa.bbb`)).text == `aaa`);
		assert(getName(Text(`aaa.bbb.ccc`)).text == `aaa.bbb`);
	}
	/**
	親フォルダ取得。

	Params:
		親フォルダを取得したいパス。

	Returns:
		親フォルダ。

	Out:
		戻り値に長さがあれば最後は\で終わってる。

	History:
		1.100:
			[S] 戻り値をrefへ。
			[S] PATH.sepの使用。
	
		1.00β16:
			[P] Phobosを使用しない様に変更。
			[S] 名前をgetOwnerFolderからownerFolderに変更。
	*/
	ref Text ownerFolder(in Text path) {
		if(path.length) {
			auto folder=delFolderSep(path);
			for(auto i=folder.length-1; i != 0; i--) {
				if(folder[i] == sep[0]) {
					return folder[0 .. i+1];
				} else if(folder[i] == ':') {
					//return folder[0 .. i+1] ~ Text('\\');
					return addFolderSep(folder[0 .. i+1]);
				}
			}
		}

		return Text(cast(tstring)['.', sep[0]]);
	}
	debug(file) unittest {
		assert(ownerFolder(Text(``)).text == `.\`);
		assert(ownerFolder(Text(`123`)).text == `.\`);
		assert(ownerFolder(Text(`123\456`)).text == `123\`);
		assert(ownerFolder(Text(`123\456\`)).text == `123\`);
		assert(ownerFolder(Text(`123\456\789`)).text == `123\456\`);
		assert(ownerFolder(Text(`C:\`)).text == `C:\`);
	}

	/**
	ファイルパス比較。

	単純にファイル文字列の比較であって相対パス等は考慮しない。
	
	History:
		1.00β11:
			新規作成。
	*/
	int cmp(in Text path1, in Text path2) {
		return path1.icmp(path2);
	}

	/**
	モジュールのパスを取得。
	
	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.00β17:
			新規作成。
	*/
	ref Text modulePath(HINSTANCE hInstance, size_t Length=PATH.MAX_PATH) {
		auto text=new wchar[Length];

		/+
		if(!GetModuleFileName(hInstance, text.ptr, text.length)) {
			throw new NeGuiPathException(ERR.toText);
		}
		+/
		enforce(GetModuleFileName(hInstance, text.ptr, text.length), new NeGuiPathException(ERR.toText));

		return Text(text.ptr);
	}

	/+
	/**
	History:
		1.00β17:
			[P] 取得をmodulePathに委譲。
	
		1.00β14:
			[P] メモリ確保用の引数追加。
			[S] バカっぽかったのを少しだけアホっぽく。
	
		1.00β11:
			nemuxi.baseから移動。
	*/
	private Text MyPath(size_t Length) {
		try {
			return modulePath(GetModuleHandle(null), Length);
		} catch(NeGuiPathException e) {
			throw new NeGuiPathException(Text("自分の居場所さえ分からない"), e);
		}
	}
	+/
	/**
	自身のアドレス取得。
	
	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.00β14:
			MyPathの引数追加に合わせてデフォルト引数の追加。

		1.00β11:
			新規作成。
	*/
	ref Text myPath(size_t Length=PATH.MAX_PATH) {
		try {
			return modulePath(GetModuleHandle(null), Length);
		} catch(NeGuiPathException e) {
			throw new NeGuiPathException(Text("自分の居場所さえ分からない"), e);
		}
	}
	/**
	自身の親フォルダ。
	
	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.00β14:
			myPathの引数追加に合わせてデフォルト引数の追加。

		1.00β11:
			nemuxi.baseから移動。
	*/
	ref Text myFolder(size_t Length=PATH.MAX_PATH) {
		return ownerFolder(myPath(Length));
	}

	/**
	カレントフォルダ取得。
	
	History:
		1.100:
			[S] 戻り値をrefへ。
	
		1.00β16:
			新規作成。
	*/
	ref Text currentFolder() {
		immutable BUFFER=GetCurrentDirectory(0, null);
		auto path=new wchar[BUFFER];

		auto CmpLength=GetCurrentDirectory(path.length, path.ptr);
		/+
		if(!CmpLength) {
			new NeGuiPathException(Text("カレントディレクトリ取得失敗"));
		}
		+/
		enforce(CmpLength, new NeGuiPathException(ERR.toText));
		enforce(CmpLength == path.length-1, new NeGuiPathException(Text("カレントディレクトリ取得後に変わった？")));

		//Text text;
		auto text=new Text;
		text.text = path[0 .. $];
		//text.text[$-1] = '\\';
		text.text[$-1] = sep[0];//'\\';
		
		return *text;
	}
	///
	bool currentFolder(in Text Path, bool ThrowException=false) {
		auto ret=cast(bool)SetCurrentDirectory(Path.ptr);

		if(!ret) {
			enforce(ThrowException, new NeGuiPathException(ERR.toText));
		}
		
		return ret;
	}


	private {
		alias extern(Windows) DWORD function(LPCTSTR, LPTSTR, DWORD) FuncType;
		ref Text ConvertShortLongName(in Text Path, FuncType Func) {
			ERR.initialize;
			const pPath=Path.ptr;
			immutable Length=Func(pPath, null, 0);

			enforce(Length, new NeGuiPathException(Text("%s, Path -> %s", ERR.toText, Path)));
			auto pResult=new wchar[Length];
			enforce(Length-1==Func(pPath, pResult.ptr, Length), new NeGuiPathException(ERR.toText));

			return Text.wrap(pResult[0..Length-1]);
		}
	}

	ref Text toShortPath(in Text Path) {
		return ConvertShortLongName(Path, cast(FuncType)&GetShortPathName);
	}
	ref Text toLongPath(in Text Path) {
		return ConvertShortLongName(Path, cast(FuncType)&GetLongPathName);
	}
}




/**
ファイル。

色々あって実装してみたものの使う機会あんましない。
std.streamかstd.stdio.File使うべきだよ。

History:
	1.100:
		新規作成。
*/
class NeFile: HandleRaii {
	alias Handle hFile;
	mixin KillResource!("ファイル", hFile, CloseHandle, 0);

	protected bool IsSeek;
	mixin(SMixInClassGet!(typeof(IsSeek))("isSeek", q{IsSeek}, false));

	///
	enum ACCESS_MODE {
		NONE  = 0,             /// オブジェクトへのデバイス問い合わせアクセスを指定します。アプリケーションは、デバイスにアクセスすることなく、デバイス属性の問い合わせを行えます。
		READ  = GENERIC_READ,  /// オブジェクトへの読み取りアクセスを指定します。データの読み取りとファイルポインタの移動ができます。読み書きアクセスを行うには、GENERIC_WRITE と共に指定します。
		WRITE = GENERIC_WRITE, /// オブジェクトへの書き込みアクセスを指定します。データの書き込みとファイルポインタの移動ができます。読み書きアクセスを行うには、GENERIC_READ と共に指定します。
		ALL   = READ | WRITE,
	}

	///
	enum SHARE {
		NONE   = 0,
		DELETE = FILE_SHARE_DELETE, /// Windows NT/2000：これ以降、オブジェクトを開こうとしたときに削除アクセスを要求した場合のみ、その操作を許可します。
		READ   = FILE_SHARE_READ,   /// これ以降、オブジェクトを開こうとしたときに読み取りアクセスを要求した場合のみ、その操作を許可します。
		WRITE  = FILE_SHARE_WRITE,  /// これ以降、オブジェクトを開こうとしたときに書き込みアクセスを要求した場合のみ、その操作を許可します。
		ALL    = DELETE | READ | WRITE,
	}

	///
	enum DISPOSITION {
		NONE              = 0,
		CREATE_NEW        = CREATE_NEW        , /// 新しいファイルを作成します。指定したファイルが既に存在している場合、この関数は失敗します。
		CREATE_ALWAYS     = CREATE_ALWAYS     , /// 新しいファイルを作成します。指定したファイルが既に存在している場合、そのファイルを上書きし、既存の属性を消去します。
		OPEN_EXISTING     = OPEN_EXISTING     , /// ファイルを開きます。指定したファイルが存在していない場合、この関数は失敗します。デバイスに対して CreateFile 関数を呼び出す際に OPEN_EXISTING フラグを使うべき状況については、この関数の「解説」を参照してください。
		OPEN_ALWAYS       = OPEN_ALWAYS       , /// ファイルが存在している場合、そのファイルを開きます。指定したファイルが存在していない場合、この関数は dwCreationDisposition パラメータで CREATE_NEW が指定されていたと仮定して新しいファイルを作成します。
		TRUNCATE_EXISTING = TRUNCATE_EXISTING , /// ファイルを開き、ファイルのサイズを 0 バイトにします。呼び出し側プロセスは、dwDesiredAccess パラメータで、少なくとも GENERIC_WRITE アクセス権を指定しなければなりません。指定したファイルが存在していない場合、この関数は失敗します。		
	}
	///
	enum FLAG {
		NONE               = 0,
		WRITE_THROUGH      = FILE_FLAG_WRITE_THROUGH,      /// キャッシュに書き込まれたデータを直接ディスクに書き込むようシステムに指示します。システムは依然としてキャッシュ書き込み操作を行えますが、ディスクへの遅延書き込みは無効になります。
		OVERLAPPED         = FILE_FLAG_OVERLAPPED,         /// オブジェクトを初期化するようシステムに指示します。その結果、非常に長い時間のかかる処理に対して ERROR_IO_PENDING が返るようになります。この処理が終了すると、指定されたイベントはシグナル状態に設定されます。このフラグを指定したときは、ファイルの読み書きを行う関数で 1 個の 構造体を指定しなければなりません。つまり、FILE_FLAG_OVERLAPPED を指定した場合、アプリケーションはオーバーラップ読み書きを行わなければなりません。FILE_FLAG_OVERLAPPED を指定した場合、システムはファイルポインタを管理しません。ファイルの読み書きを行う関数に対し、1 個の OVERLAPPED 構造体を指している lpOverlapped パラメータの一部として、ファイルの位置を渡さなければなりません。また、このフラグを指定した場合、1 つのハンドルに対して複数の操作（たとえば、読み取りと書き込みそれぞれの操作）を同時に行うことができます。
		NO_BUFFERING       = FILE_FLAG_NO_BUFFERING,       /// バッファやキャッシュを使わずにファイルを開くようシステムに指示します。FILE_FLAG_OVERLAPPED フラグと共に指定すると、非同期操作のパフォーマンスが向上します。I/O 操作を行う際に、メモリマネージャの同期処理に依存する必要がなくなるからです。しかし、データがキャッシュ内に存在しないので、一部の I/O 操作には多くの時間がかかります。このフラグを指定した場合、アプリケーションは次の各条件を満たさなければなりません。<ul><li>ファイルアクセスの開始オフセットを、ボリュームのセクタサイズの整数倍にしなければなりません。</li><li>ファイルアクセスのバイト数を、ボリュームのセクタサイズの整数倍にしなければなりません。たとえば、セクタサイズが 512 バイトの場合、512 バイト、1,024 バイト、2,048 バイトの読み書きはできますが、335 バイト、981 バイト、7,171 バイトなどの読み書きはできません。</li><li>読み書き操作用のバッファのアドレス（メモリ内のアドレス）を、ボリュームのセクタサイズの整数倍に整列（セクタ整列）させなければなりません。ディスクによっては、この要件を満たせないこともあります。</li></ul>3 番目の条件を満たす、つまりバッファをボリュームのセクタサイズの整数倍の位置に整列させる 1 つの方法は、 関数を使ってバッファを割り当てることです。VirtualAlloc 関数が割り当てるメモリは、オペレーティングシステムのメモリページサイズの整数倍の位置に整列されています。メモリページとボリュームのセクタサイズは両方とも 2 のべき乗なので、このメモリは、ボリュームのセクタサイズの整数倍のアドレスにも整列されます。ボリュームのセクタサイズを求めるには、GetDiskFreeSpace 関数を使います。また、ページサイズを調べるには、 関数を使います。
		RANDOM_ACCESS      = FILE_FLAG_RANDOM_ACCESS,      /// このファイルをランダムアクセスすることをシステムに指示します。システムは、この指定をファイルのキャッシングを最適化するためのヒントとして使います。
		SEQUENTIAL_SCAN    = FILE_FLAG_SEQUENTIAL_SCAN,    /// ファイルを最初から最後までシーケンシャルに（連続）アクセスすることをシステムに指示します。システムは、この指定をファイルのキャッシングを最適化するためのヒントとして使います。この場合、アプリケーションはランダムアクセスを行う目的でファイルポインタを移動することはできますが、最適なキャッシングが達成できなくなる可能性があります。しかし、ランダムアクセスを正しく行えることは保証されています。このフラグを指定すると、大きなファイルをシーケンシャルアクセスするアプリケーションのパフォーマンスが向上します。基本的にシーケンシャルアクセスを行い、ときどき小規模な範囲をスキップするアプリケーションでは、特に大きな向上が見られることもあります。
		DELETE_ON_CLOSE    = FILE_FLAG_DELETE_ON_CLOSE,    /// FILE_FLAG_DELETE_ON_CLOSE を指定したハンドルを除き、このファイルに関連付けられているすべてのハンドルが閉じた段階で、そのファイルを即座に削除するようシステムに指示します。それ以降、FILE_SHARE_DELETE を指定しない限り、このファイルを開くよう要求しても失敗します。
		BACKUP_SEMANTICS   = FILE_FLAG_BACKUP_SEMANTICS,   /// Windows NT/2000：バックアップまたは復元操作の目的で、ファイルを開くまたは作成するようシステムに指示します。呼び出し側プロセスがファイルセキュリティに関して十分な権限を備えていることを前提として、システムは、そのプロセスがファイルセキュリティのチェックを無効にできることを保証します。これに関連する権限は、 SE_BACKUP_NAME と SE_RESTORE_NAME です。ディレクトリのハンドルを取得する目的で、このフラグを指定することもできます。特定の Win32 関数で、ファイルハンドルの代わりにディレクトリハンドルを指定することもできます。
		POSIX_SEMANTICS    = FILE_FLAG_POSIX_SEMANTICS,    /// POSIX（UNIX ベースのポータブル OS）の規則に従ってファイルにアクセスするようシステムに指示します。この規則には、ファイル名が同じで大文字と小文字だけが異なる複数のファイルを認めるファイルシステムに対し、それらのファイルへのアクセスを認めることが含まれます。このフラグを指定する場合、注意が必要です。MS-DOS または 16 ビット版 Windows からは、このフラグを指定して作成したファイルにアクセスできないからです。
		OPEN_REPARSE_POINT = FILE_FLAG_OPEN_REPARSE_POINT, /// このフラグを指定すると、NTFS の再解析ポイントで再解析を行うことを禁止します。このフラグを指定してファイルを開くと、再解析ポイントを制御するフィルタが動作しているかどうかにかかわりなく、ファイルハンドルが返ります。このフラグと共に CREATE_ALWAYS フラグを指定することはできません。
		OPEN_NO_RECALL     = FILE_FLAG_OPEN_NO_RECALL,     /// ファイルのデータを要求すると共に、そのデータを継続的にリモート記憶域に配置するべきことをシステムに指示します。そのファイルをローカル記憶域へ復帰させるべきではありません。このフラグは、リモート記憶域システム、言い換えると階層記憶管理システムで使うことを意図しています。
	}
	///
	override this(HANDLE Handle, bool Suicide, bool IsSeek) {
		this.IsSeek = IsSeek;
		super(Handle, Suicide);
	}
	///
	protected this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, in SECURITY* Security, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, NeFile Template, bool IsSeek) {
		ERR.initialize;
		
		auto hHandle = CreateFile(
			Path.ptr,
			AccessMode,
			Shere,
			Security ? Security.ptr: null,
			Disposition,
			Attribute | Flags,
			Template ? Template(): null
		);
		enforce(hFile != INVALID_HANDLE_VALUE, new NeGuiFileException(Text("%s, %s", ERR.toText, Path)));
		this(hHandle, true, IsSeek);
	}
	///
	this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, const ref SECURITY Security, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, NeFile Template, bool IsSeek) {
		this(Path, AccessMode, Shere, &Security, Disposition, Attribute, Flags, Template, IsSeek);
	}
	///
	this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, bool IsSeek) {
		this(Path, AccessMode, Shere, null, Disposition, Attribute, Flags, null, IsSeek);
	}
	private static NeFile rw(in Text Path, bool Write=false) {
		return new NeFile(
			Path,
			Write ? ACCESS_MODE.READ | ACCESS_MODE.WRITE: ACCESS_MODE.READ,
			SHARE.ALL,
			Write ? DISPOSITION.OPEN_ALWAYS: DISPOSITION.OPEN_EXISTING,
			FILE.ATTRIBUTE.NORMAL,
			FLAG.NONE,
			true
		);
	}
	///
	static NeFile reader(in Text Path) {
		return rw(Path, false);
	}
	///
	static NeFile writer(in Text Path) {
		return rw(Path, true);
	}

	///
	enum TYPE {
		UNKNOWN = FILE_TYPE_UNKNOWN, /// 指定されたファイルの種類は不明です。
		DISK    = FILE_TYPE_DISK,    /// 指定されたファイルは、ディスクファイルです。
		CHAR    = FILE_TYPE_CHAR,    /// 指定されたファイルは、LPT デバイスやコンソールのような文字ファイルです。
		PIPE    = FILE_TYPE_PIPE,    /// 指定されたファイルは、名前付きまたは名前なしパイプです。
		//REMOTE  = FILE_TYPE_REMOTE,  ///
	}
	///
	const TYPE type() {
		return cast(TYPE)GetFileType(hFile);
	}


	
	/**
	この関数の動作を修飾するフラグを指定します。このパラメータで、次の値のいずれかまたは任意の組み合わせを指定します。
	*/
	enum LOCK {
		FAIL_IMMEDIATELY = LOCKFILE_FAIL_IMMEDIATELY, /// 要求したロックを獲得できなかった場合、この関数はすぐに制御を返します。この値を指定しなかった場合は、ロックを獲得できるまで待ち続けます。
		EXCLUSIVE_LOCK   = LOCKFILE_EXCLUSIVE_LOCK,   /// この関数は、排他ロックを要求します。この値を指定しなかった場合は、共有ロックを要求します。
	}
	///
	bool lock(const ref LARGE Offset, const ref LARGE Size) {
		return cast(bool)LockFile(hFile, Offset.low, Offset.high, Size.low, Size.high);
	}
	///
	bool lock(LOCK Lock, const ref LARGE Size, ref OVERLAPPEDINFO OverlappedInfo) {
		return cast(bool)LockFileEx(hFile, Lock, 0, Size.low, Size.high, OverlappedInfo.ptr);
	}
	///
	bool unLock(const ref LARGE Offset, const ref LARGE Size) {
		return cast(bool)UnlockFile(hFile, Offset.low, Offset.high, Size.low, Size.high);
	}
	///
	bool unLock(const ref LARGE Size, ref OVERLAPPEDINFO OverlappedInfo) {
		return cast(bool)UnlockFileEx(hFile, 0, Size.low, Size.high, OverlappedInfo.ptr);
	}

	///
	enum SEEK {
		TOP = FILE_BEGIN,   /// 0、またはファイルの先頭が開始点になります。このフラグを指定すると、liDistanceToMove パラメータは符号のない値として解釈されます。
		NOW = FILE_CURRENT, /// 現在のファイルポインタの値が開始点になります。
		END = FILE_END,     /// 現在ファイルの終端の位置が開始点になります。
	}
	/**
	現在位置の変更。

	Returns:
		変更後の位置、
	*/
	const LARGE seekTo(const ref LARGE Position, SEEK Seek) {
		LARGE Result;
		enforce(SetFilePointerEx(hFile, Position.LargeInteger, Result.ptr, Seek), new NeGuiFileException(Text("%s, IsSeek -> %s", ERR.toText, IsSeek)));
		return Result;
	}
	const LARGE seekTo(long Position, SEEK Seek) {
		LARGE Large;
		Large.total = Position;
		return seekTo(Large, Seek);
	}
	

	/// 現在位置
	@property const LARGE position() {
		LARGE Large;
		Large.total = 0;
		return seekTo(Large, SEEK.NOW);
	}
	/// ditto
	@property const LARGE position(LARGE Large) {
		return seekTo(Large, SEEK.NOW);
	}

	/// 現在位置を先頭へ。
	bool seekToTop() {
		LARGE Large;
		return collectException(seekTo(Large, SEEK.TOP)) is null;
	}
	/// 現在位置を終端へ。
	bool seekToEnd() {
		LARGE Large;
		return collectException(seekTo(Large, SEEK.END)) is null;
	}
	
	///
	bool eof() {
		auto NowPosition=position;
		seekToEnd();
		auto LastPosition=position;
		scope(exit) position = NowPosition;
		
		return NowPosition == LastPosition;
	}
	///
	bool setEof() {
		return cast(bool)SetEndOfFile(hFile);
	}
	
	///
	final bool flush() {
		return cast(bool)FlushFileBuffers(hFile);
	}
	///
	final bool cancel() {
		return cast(bool)CancelIo(hFile);
	}

	protected {
		/**
		*/
		size_t Read(void* Buffer, size_t BufferSize) {
			size_t Size;

			enforce(ReadFile(hFile, Buffer, BufferSize, &Size, null), new NeGuiFileException(ERR.toText));
			
			return Size;
		}
		/**
		*/
		size_t Write(in void* Buffer, size_t BufferSize)
		in {
			assert(BufferSize, "use timeStamp()");
		}
		body {
			size_t Size;

			enforce(WriteFile(hFile, cast(void*)Buffer, BufferSize, &Size, null), new NeGuiFileException(ERR.toText));
			
			return Size;
		}

		void SeekExec(void delegate() SeekDg, void delegate() NonSeekDg=null) {
			if(IsSeek) {
				SeekDg();
			} else if(NonSeekDg) {
				NonSeekDg();
			}
		}
	}
	/**
	読み込み。

	Params:
		Buffer = 読み込む領域。

	Returns:
		正常に読み込めた場合はtrue、失敗した場合はfalseを返す。
		仮にshortを読み込もうとしてbyteしか読めなかった場合はシークを呼び出し前の位置に直してfalseを返す。
	*/
	bool get(T)(out T Buffer)
	if(!isArray!(T))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Read(&Buffer, T.sizeof);
		if(Size == T.sizeof) {
			return true;
		} else {
			SeekExec({seekTo(OldSeek, SEEK.TOP);});
			return false;
		}
	}
	/**
	読み込み。

	Params:
		Buffer = 読み込む領域。

	Returns:
		Bufferの有効なインデックスを返す。
	
	*/
	size_t read(T)(ref T[] Buffer)
	if(isArray!(T[]))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Read(Buffer.ptr, T.sizeof * Buffer.length);
		auto SafeIndex=Size / T.sizeof;
		if(SafeIndex != Buffer.length) {
			SeekExec({seekTo(OldSeek.total + T.sizeof * SafeIndex, SEEK.TOP);});
		}
		return SafeIndex;
	}

	/**
	書き込み。

	Params:
		Buffer = 書き込む値。

	Returns:
		正常に書き込めた場合はtrue、失敗した場合はfalseを返す。
		仮にshortを書き込もうとしてbyteしか書けなかった場合はシークを呼び出し前の位置に直してfalseを返す。
	*/
	bool put(T)(T Buffer)
	if(!isArray!(T))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Write(&Buffer, T.sizeof);
		if(Size == T.sizeof) {
			return true;
		} else {
			SeekExec({seekTo(OldSeek, SEEK.TOP);});
			return false;
		}
	}
	/**
	書き込み。

	Params:
		Buffer = 書き込む値。

	Returns:
		書き込まれたBufferの有効なインデックスを返す。
	
	*/
	size_t write(T)(in T[] Buffer)
	if(isArray!(T[]))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Write(Buffer.ptr, T.sizeof * Buffer.length);
		auto SafeIndex=Size / T.sizeof;
		if(SafeIndex != Buffer.length) {
			SeekExec({seekTo(OldSeek + T.sizeof * SafeIndex, SEEK.TOP);});
		}
		return SafeIndex;
	}
}


///
static struct FILE {
	static:

	///
	enum ATTRIBUTE {
		NONE                = 0xFFFFFFFF,
		ARCHIVE             = FILE_ATTRIBUTE_ARCHIVE,             /// アーカイブファイルまたはアーカイブディレクトリです。アプリケーションはこの属性を、ファイルのバックアップや削除のためのマークとして使います。
		COMPRESSED          = FILE_ATTRIBUTE_COMPRESSED,          /// 指定されたファイルまたはディレクトリは圧縮されています。ファイルの場合、ファイル内の全データが圧縮されていることを意味します。ディレクトリの場合、そのディレクトリ内に新しく作成されるファイルまたはサブディレクトリが、既定で圧縮状態になることを意味します。
		DEVICE              = FILE_ATTRIBUTE_DEVICE,              /// 予約済み。使わないでください。
		FOLDER              = FILE_ATTRIBUTE_DIRECTORY,           /// 指定されたハンドルは、ディレクトリに関連しています。
		ENCRYPTED           = FILE_ATTRIBUTE_ENCRYPTED,           /// 指定されたファイルまたはディレクトリは暗号化されています。ファイルの場合、ファイル内の全データストリームが暗号化されていることを意味します。ディレクトリの場合、そのディレクトリ内に新しく作成されるファイルまたはサブディレクトリが、既定で暗号化状態になることを意味します。
		HIDDEN              = FILE_ATTRIBUTE_HIDDEN,              /// 隠しファイルまたは隠しディレクトリです。通常のディレクトリリスティングでは表示されません。
		NORMAL              = FILE_ATTRIBUTE_NORMAL,              /// 指定されたファイルまたはディレクトリには、特に属性はありません。単独で返った場合にのみ、この属性は有効です。
		NOT_CONTENT_INDEXED = FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, /// Windows 2000：このファイルは、「インデックスサービス」の対象になっていません。
		OFFLINE             = FILE_ATTRIBUTE_OFFLINE,             /// Windows 2000：このファイルのデータは、すぐには利用できません。この属性は、ファイルのデータがオフラインの記憶装置へ物理的に移動されたことを示します。この属性は、 Windows 2000 の階層記憶管理ソフトウェアである「リモート記憶域」が利用するものです。アプリケーションは、任意にこの属性を変更するべきではありません。
		READONLY            = FILE_ATTRIBUTE_READONLY,            /// このファイルまたはディレクトリは読み取り専用です。アプリケーションはこのファイルの読み取りを行えますが、書き込みや削除はできません。ディレクトリの場合、アプリケーションは削除を行えません。
		REPARSE_POINT       = FILE_ATTRIBUTE_REPARSE_POINT,       /// このファイルには、再解析ポイントが関連付けられています。
		SPARSE_FILE         = FILE_ATTRIBUTE_SPARSE_FILE,         /// このファイルは、スパースファイル（疎なファイル、未使用の領域が多い、または同じ値が長く続くファイル）です。
		SYSTEM              = FILE_ATTRIBUTE_SYSTEM,              /// このファイルまたはディレクトリは、オペレーティングシステムの一部、またはオペレーティングシステム専用です。
		TEMPORARY           = FILE_ATTRIBUTE_TEMPORARY,           /// このファイルは、一時ファイルとして使われています。ファイルシステムは、データをハードディスクのような大容量記憶装置へ書き込む代わりに、高速なアクセスが行えるよう、すべてのデータをメモリ内に維持することを試みます。アプリケーションは、必要がなくなった段階で一時ファイルをすぐに削除するべきです。
	}
	/**
	*/
	ATTRIBUTE attributes(in Text path) {
		return cast(ATTRIBUTE)GetFileAttributes(path.ptr);
	}
	/// ditto
	bool attributes(in Text path, ATTRIBUTE Attribute) {
		return cast(bool)SetFileAttributes(path.ptr, Attribute);
	}
	/**
	指定アドレスは存在するか。

	Params:
		path = 調べたいアドレス。

	Returns:
		存在すればtrue、しなければfalse。
	*/
	bool isExistence(in Text path) {
		return attributes(path) != ATTRIBUTE.NONE;
	}
	/**
	指定アドレスはフォルダか。

	Params:
		path = 調べたいアドレス。

	Returns:
		フォルダならtrue。

	Throws:
		アドレス自体が存在しなければNeGuiFileException。
	*/
	bool isFolder(in Text path) {
		enforce(isExistence(path), new NeGuiFileException(Text("そもそもファイル無い。")));
		return (attributes(path) & ATTRIBUTE.FOLDER) != 0;
	}
	/**
	指定アドレスはファイルか。

	Params:
		path = 調べたいアドレス。

	Returns:
		ファイルならtrue。

	Throws:
		アドレス自体が存在しなければNeGuiFileException。
	*/
	bool isFile(in Text path) {
		return !isFolder(path);
	}

	
	/**
	フォルダ作成。

	Params:
		path = フォルダ
	*/
	bool makeFolder(in Text path, bool Recurse=true) {
		if(Recurse) {
			const left=PATH.ownerFolder(path);
			isExistence(left) || makeFolder(left, true);
			return makeFolder(path, false);
		} else {
			return cast(bool)CreateDirectory(PATH.delFolderSep(path).ptr, null);
		}
	}
	
	bool delFile(in Text path) {
		return cast(bool)DeleteFile(path.ptr);
	}
	/**
	History:
		1.100:
			[F] ファイルの有無による処理。

	Throws:
		AllDeleteがtrueでファイル列挙に失敗すればNeGuiExceptionを投げる
	*/
	bool delFolder(in Text path, bool AllDelete=false) {
		if(AllDelete) {
			try {
				enforce(isFolder(path), new NeGuiFileException(Text("%s is not found.", path)));
			} catch(Exception e) {
				throw new NeGuiFileException(Text("path -> %s", path), e);
			}
			
			FILEITEM FileItem=void;
			scope finder=new Finder(PATH.join(path, Text('*')));
			while(finder.find(FileItem)) if(FileItem.isFolderItems) {
				const FileName=PATH.join(path, FileItem.name);
				if(FileItem.isFolder) {
					delFolder(FileName, true);
				} else {
					delFile(FileName);
				}
			}
		}
		
		return cast(bool)RemoveDirectory(path.ptr);
	}

	private enum {
		MOVEFILE_REPLACE_EXISTING       = 0x00000001,
		MOVEFILE_COPY_ALLOWED           = 0x00000002,
		MOVEFILE_DELAY_UNTIL_REBOOT     = 0x00000004,
		MOVEFILE_WRITE_THROUGH          = 0x00000008,
		MOVEFILE_CREATE_HARDLINK        = 0x00000010,
		MOVEFILE_FAIL_IF_NOT_TRACKABLE  = 0x00000020,
	}
	/***/
	enum MOVEFLAG {
		COPY_ALLOWED          = MOVEFILE_COPY_ALLOWED          , /// 他のボリュームへファイルを移動するよう指示された場合、この関数は、CopyFile 関数と DeleteFile 関数を使って、その動作をシミュレートします。このフラグを、MOVEFILE_DELAY_UNTIL_REBOOT フラグと共に指定することはできません。
		CREATE_HARDLINK       = MOVEFILE_CREATE_HARDLINK       , /// 将来の使用に備えて予約されています。
		DELAY_UNTIL_REBOOT    = MOVEFILE_DELAY_UNTIL_REBOOT    , /// この関数は、オペレーティングシステムを再起動するまでファイルを移動しません。システムは、AUTOCHK（Windows NT/2000 に付属していて、Chkdsk に相当し、起動時に実行できます）を実行した直後、かつページングファイルを作成する前にファイルを移動します。その後、この値が指定されている場合は、この関数は前回のスタートアップに使ったページングファイルを削除できます。このフラグを指定できるのは、Administrators グループに所属するユーザーのコンテキスト、または LocalSystem アカウントのコンテキストだけです。このフラグを、MOVEFILE_COPY_ALLOWED フラグと共に指定することはできません。
		FAIL_IF_NOT_TRACKABLE = MOVEFILE_FAIL_IF_NOT_TRACKABLE , /// Windows 2000：移動元ファイルがリンクソースの場合、このファイルは失敗します。そして、移動後は、そのファイルのリンクトラッキングを行えません。FAT ファイルシステムでフォーマットされたボリュームを移動先として指定すると、このような問題が発生することがあります。
		REPLACE_EXISTING      = MOVEFILE_REPLACE_EXISTING      , /// lpNewFileName パラメータで指定されたファイル名が既に存在していた場合、その内容を lpExistingFileName パラメータで指定されたファイルに置き換えます。lpNewFileName と lpExistingFileName どちらかのパラメータでディレクトリを指定した場合は、この値を利用できません。
		WRITE_THROUGH         = MOVEFILE_WRITE_THROUGH         , /// ファイルがディスク上で実際に移動するまで、制御を返しません。このフラグを指定すると、コピーと削除の操作からなる移動操作をディスクへフラッシュした（実際に書き込んだ）後、この関数から制御が返ることが保証されます。コピー操作が終わった後、フラッシュが実施されます。このフラグは、MOVEFILE_DELAY_UNTIL_REBOOT フラグが指定されていると機能しません。
	}
	/***/
	bool move(in Text SrcPath, in Text NewPath) {
		return cast(bool)MoveFile(SrcPath.ptr, NewPath.ptr);
	}
	/// ditto
	bool move(in Text SrcPath, in Text NewPath, MOVEFLAG MoveFlags) {
		return cast(bool)MoveFileEx(SrcPath.ptr, NewPath.ptr, MoveFlags);
	}
	/***/
	bool copy(in Text SrcPath, in Text NewPath, bool Overwrite) {
		return cast(bool)CopyFile(SrcPath.ptr, NewPath.ptr, !Overwrite);
	}




	private final scope class FileHandle: NeFile {
		this(in Text Path, bool Write=false, string _file=__FILE__, int _line=__LINE__) {
			super(
				Path,
				Write ? ACCESS_MODE.READ | ACCESS_MODE.WRITE: ACCESS_MODE.READ,
				SHARE.ALL,
				null,
				DISPOSITION.OPEN_EXISTING,
				FILE.ATTRIBUTE.NORMAL,
				FLAG.NONE,
				null,
				true
			);
		}
	}

	/***/
	long fileSize(in Text Path) {
		scope Handle=new FileHandle(Path);
		//LARGE_INTEGER LargeInteger;
		LARGE Large;
		enforce(GetFileSizeEx(Handle(), Large.ptr), new NeGuiFileException(ERR.toText));
		return Large.total;
	}
	/***/
	ulong fileSizeCompressed(in Text Path) {
		ERR.initialize;
		DWORD HiSize;
		DWORD LoSize=GetCompressedFileSize(Path.ptr, &HiSize);
		if(LoSize == -1 && ERR.code != ERR.noError) {
			throw new NeGuiFileException(Text("%s, Path -> %s", ERR.toText, Path));
		}
		
		if(HiSize) {
			return cast(ulong)(HiSize * DWORD.max)+1 + LoSize;
		} else {
			return LoSize;
		}
	}

	/***/
	ref TIMESTAMP timeStamp(in Text Path, TIMESTAMP.FLAG Flags=TIMESTAMP.FLAG.ALL) {
		scope Handle=new FileHandle(Path);
		
		FILEDATETIME Create, Access, Update;
		GetFileTime(
			Handle(),
			Flags & TIMESTAMP.FLAG.CREATE ? Create.ptr: null,
			Flags & TIMESTAMP.FLAG.ACCESS ? Access.ptr: null,
			Flags & TIMESTAMP.FLAG.UPDATE ? Update.ptr: null
		);
		auto TimeStamp=new TIMESTAMP;
		with(*TimeStamp) {
			create = Flags & TIMESTAMP.FLAG.CREATE ? new DateTime(DateTime.MODE.STRICT, Create.toLocalFileDateTime): null;
			access = Flags & TIMESTAMP.FLAG.ACCESS ? new DateTime(DateTime.MODE.STRICT, Access.toLocalFileDateTime): null;
			update = Flags & TIMESTAMP.FLAG.UPDATE ? new DateTime(DateTime.MODE.STRICT, Update.toLocalFileDateTime): null;
		}
		TimeStamp.flags = Flags;
		
		return *TimeStamp;
	}
	/***/
	bool timeStamp(in Text Path, const ref TIMESTAMP TimeStamp) {
		scope Handle=new FileHandle(Path, true);
		with(TimeStamp) {
			return cast(bool)SetFileTime(
				Handle(),
				TimeStamp.flags & TIMESTAMP.FLAG.CREATE ? create.toFileDateTime.toSystemFileDateTime.ptr: null,
				TimeStamp.flags & TIMESTAMP.FLAG.ACCESS ? access.toFileDateTime.toSystemFileDateTime.ptr: null,
				TimeStamp.flags & TIMESTAMP.FLAG.UPDATE ? update.toFileDateTime.toSystemFileDateTime.ptr: null
			);
		}
	}

	private Text TempFile(in Text Folder, in Text Prefix, size_t Unique) {
		enforce(isFolder(Folder), new NeGuiFileException(Text("Folder -> %s", Folder)));
		
		auto Buffer=new wchar[PATH.MAX_NAME];
		
		auto Result=GetTempFileName(Folder.ptr, Prefix.ptr, Unique, Buffer.ptr);
		enforce(Result, new NeGuiFileException(ERR.toText));
		
		return Text(Buffer.ptr);
	}
	/***/
	Text makeTempFile(in Text Folder, in Text Prefix=Text.init) {
		return TempFile(Folder, Prefix, 0);
	}
	/***/
	Text getTempFile(in Text Folder, in Text Prefix=Text.init, size_t Unique=1) {
		enforce(Unique, new NeGuiFileException(Text("Unique -> 0")));
		return TempFile(Folder, Prefix, Unique);
	}

}

///
struct DISKSPACE {
	ULARGE available;
	ULARGE total;
	ULARGE free;

	const Text toText() {
		return Text(
			"available -> %s total -> %s free -> %s",
			available, total, free
		);
	}

}

/***/
struct DRIVE {
	static:

	/***/
	ref DISKSPACE diskSpace(in Text Folder) {
		auto DiskSpace=new DISKSPACE;

		enforce(GetDiskFreeSpaceEx(
			Folder.ptr(true),
			DiskSpace.available.ptr,
			DiskSpace.total.ptr,
			DiskSpace.free.ptr
		), new DiskException(ERR.toText));
		
		return *DiskSpace;
	}

	
	/**

	*/
	enum TYPE {
		UNKNOWN     = DRIVE_UNKNOWN,     /// 不明です。 
		NO_ROOT_DIR = DRIVE_NO_ROOT_DIR, /// 指定したルートディレクトリは存在しません。 
		REMOVABLE   = DRIVE_REMOVABLE,   /// ドライブからディスクを抜くことができます。 
		FIXED       = DRIVE_FIXED,       /// ドライブからディスクを抜くことができません。 
		REMOTE      = DRIVE_REMOTE,      /// ネットワーク ドライブです。 
		CDROM       = DRIVE_CDROM,       /// CD-ROM ドライブです。 
		RAMDISK     = DRIVE_RAMDISK,     /// RAM ディスク ドライブです。 
	}
	/***/
	TYPE type(in Text Drive) {
		return cast(TYPE)GetDriveType(Drive.ptr);
	}
	/***/
	BIT!(DWORD) drives() {
		return BIT!(DWORD)(GetLogicalDrives());
	}
	/***/
	Text[] names() {
		auto Length=GetLogicalDriveStrings(0, null);
		enforce(Length, new DriveException(ERR.toText));

		auto Buffer=new wchar[Length+1];
		enforce(GetLogicalDriveStrings(Length+1, Buffer.ptr), new DriveException(ERR.toText));

		auto TextNullList=TEXTNULLLIST(Buffer.ptr);
		return TextNullList.datas;
	}
}


