﻿/**
NeGui初期化。

ホントはmainとかWinMainにラップできればよかったんだかどなー。

History:
	1.100:
		[S] 名前変更(neguiinitialize -> initializer)

	1.031:
		新規作成。
*/
module etc.negui.initializer;

debug import std.stdio: wl = writefln;

import std.contracts;
public import core.runtime;

public import win32.windows;

public import etc.negui.system.base;
import etc.negui.com.com;
import etc.negui.window.newindow;
public import etc.negui.system.type.enumerated;
public import etc.negui.window.dialog.system;
import etc.negui.control.control;
public import etc.negui.input.commandline;

///
class NeGuiInitializeException: NeGuiException {
	mixin MixInNeGuiException;
}


private version(unittest) {
	bool NoCall=true;
}

/**
NeGui初期化。

NeGuiを使用する前に呼び出しが必要。

<ul>
	<li>etc.negui.window.newindow</li>
	<li>etc.negui.control.control</li>
	<li>etc.negui.com.com</li>
</ul>
上記三つのmoduleを使用する際に必要。
使いたくない場合は手で初期化(各module参照)。

History:
	1.051:
		[B] NeControlInitialize忘れ。
*/
void NeGuiInitialize() {
	version(unittest) assert(NoCall);
	scope(exit) version(unittest) NoCall = false;

	try {
		NeWindowInitialize();
		NeControlInitialize();
		CommonControlInitialize();
		ComInitialize();
	} catch(Exception e) {
		throw new NeGuiInitializeException(Text("negui initialize"), e);
	}
}

/**
NeGui破棄。

NeGuiを使用し終わった際に呼び出しが必要。
*/
void NeGuiUnInitialize() {
	ComUnInitialize();
}

/+
string SMixInNeGuiInitialize() {
	return `

	NeGuiInitialize();
	scope(exit) {
		NeGuiUnInitialize();
	}

	`;
}
+/

///
alias int function(CommandLine Argument) MainFunc;
/**
mainラッパー。

*/
mixin template main(alias UserMain) {
	version(all) {
		int main() {
			NeGuiInitialize();
			scope(exit) NeGuiUnInitialize();
			
			return UserMain(new CommandLine());
		}
	} else {
		static assert(false, "-version=CUI");
	}
}

///
alias int function(HINSTANCE hInstance, CommandLine Argument, SHOW Show) WinMainFunc;
/**
WinMainラッパー。

Runtime.initialize, NeGuiInitialize, NeGuiUnInitializeをそれぞれ処理。

Examples:
* ----------------------------------------
* int NeGuiMain(HINSTANCE hInstance, CommandLine Argument, SHOW Show) {
* 	...
* }
* mixin WinMain!(NeGuiMain);
* ----------------------------------------


*/
mixin template WinMain(alias UserMain) {
	version(CUI) {
		static assert(false, "mixin WinMain!(" ~ UserMain.stringof ~ "), -version=CUI error");
	}
	
	extern (Windows) int WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPCTSTR lpCmdLine, int nCmdShow) {
		
		void exceptionHandler(Throwable e) {
			throw e;
		}
		
		try {
			Runtime.initialize(&exceptionHandler);
			
			NeGuiInitialize();
			scope(exit) NeGuiUnInitialize();
			
			return UserMain(hInstance, new CommandLine(), cast(SHOW)nCmdShow);
		} catch(Object e) {
			MessageDialog.error(Text(e.toString));
			return typeof(return).min;
		} finally {
			Runtime.terminate(&exceptionHandler);
		}
	}
}
