﻿/**

History:
	1.100:
		新規作成。
*/
module etc.negui.process.console;

debug import std.stdio: wl = writefln;
debug(console) void main() {}

import std.contracts;
import std.ctype;

import win32.windows;

import etc.negui.system.base;
import etc.negui.system.type.base;
import etc.negui.file.file;
import etc.negui.process.pipe;
import etc.negui.input.keyboard.keyboard;

///
class ConsoleException: NeGuiException {
	mixin MixInNeGuiException;
}

/***/
class ConsoleHandle: NeFile {
	enum STD_IO {
		INPUT  = STD_INPUT_HANDLE,  /// 標準入力のハンドルを取得
		OUTPUT = STD_OUTPUT_HANDLE, /// 標準出力のハンドル取得
		ERROR  = STD_ERROR_HANDLE,  /// 標準エラーのハンドル取得
	}
	this(STD_IO StdIO) {
		auto hStdHandle=GetStdHandle(StdIO);
		enforce(hStdHandle != INVALID_HANDLE_VALUE, new ConsoleException(ERR.toText));
		super(hStdHandle, false, false);
	}
	/+
	override const LARGE seekTo(const ref LARGE Position, SEEK Seek) {
		//LARGE Result;
		//enforce(SetFilePointerEx(hFile, Position.LargeInteger, Result.ptr, Seek), new NeGuiFileException(ERR.toText));
		return LARGE.init;
	}
	+/
}

///
enum FILLATTRIBUTE: WORD {
	FORE_R = FOREGROUND_RED       , /// テキスト色を赤を含める
	FORE_G = FOREGROUND_GREEN     , /// テキスト色を緑を含める
	FORE_B = FOREGROUND_BLUE      , /// テキスト色を青を含める
	FORE_I = FOREGROUND_INTENSITY , /// テキストを高輝度で表示する
	
	BACK_R = BACKGROUND_RED       , /// 背景色に赤を含める
	BACK_G = BACKGROUND_GREEN     , /// 背景色に緑を含める
	BACK_B = BACKGROUND_BLUE      , /// 背景色に青を含める
	BACK_I = BACKGROUND_INTENSITY , /// 背景色を高輝度で表示する
}

/***/
struct CONSOLECURSORINFO {
	CONSOLE_CURSOR_INFO ConsoleCursorInfo;
	mixin(SMixInStructHiddenOriginal!(CONSOLE_CURSOR_INFO)(q{ConsoleCursorInfo}));
	
	mixin(SMixInStructGetSet!(size_t)("size", q{ConsoleCursorInfo.dwSize}));
	mixin(SMixInStructGetSet!(bool)("visible", q{ConsoleCursorInfo.bVisible}));
}

///
struct CONSOLEPOINT {
	COORD Coord;
	mixin(SMixInStructHiddenOriginal!(COORD)(q{Coord}));
	
	mixin(SMixInStructGetSet!(typeof(Coord.X))("x", q{Coord.X}));
	mixin(SMixInStructGetSet!(typeof(Coord.Y))("y", q{Coord.Y}));
}

///
struct CONSOLESCREENBUFFER {
	CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
	mixin(SMixInStructHiddenOriginal!(CONSOLE_SCREEN_BUFFER_INFO)(q{ConsoleScreenBufferInfo}));

	mixin(SMixInStructGetSet!(CONSOLEPOINT)("line", q{ConsoleScreenBufferInfo.dwSize}));
	mixin(SMixInStructGetSet!(CONSOLEPOINT)("cursor", q{ConsoleScreenBufferInfo.dwCursorPosition}));
	mixin(SMixInStructGetSet!(CONSOLEPOINT)("maxSize", q{ConsoleScreenBufferInfo.dwMaximumWindowSize}));

	mixin(SMixInStructGetSet!(FILLATTRIBUTE)("color", q{ConsoleScreenBufferInfo.wAttributes}));
	
	mixin(SMixInStructGetSet!(SMALL_RECT)("rect", q{ConsoleScreenBufferInfo.srWindow}));
}

/***/
struct CONSOLEWINDOWEVENTRECORD {
	WINDOW_BUFFER_SIZE_RECORD WindowBufferSizeRecord;
	mixin(SMixInStructHiddenOriginal!(WINDOW_BUFFER_SIZE_RECORD)(q{WindowBufferSizeRecord}));
	
	mixin(SMixInStructGetSet!(CONSOLEPOINT)("size", q{WindowBufferSizeRecord.dwSize}));
}


/***/
struct CONSOLEFOCUSEVENTRECORD {
	FOCUS_EVENT_RECORD ConsoleFocusEventRecord;
	mixin(SMixInStructHiddenOriginal!(FOCUS_EVENT_RECORD)(q{ConsoleFocusEventRecord}));
}

/**
*/
enum CONSOLEMODKEY {
	CAPSLOCK   = CAPSLOCK_ON,            /// 
	ENHANCED   = ENHANCED_KEY,           /// 
	ALT_LEFT   = LEFT_ALT_PRESSED,       /// 
	CTRL_LEFT  = LEFT_CTRL_PRESSED,      /// 
	NUMLOCK    = NUMLOCK_ON,             /// 
	ALT_RIGHT  = RIGHT_ALT_PRESSED,      /// 
	CTRL_RIGHT = RIGHT_CTRL_PRESSED,     /// 
	SCROOLLOCK = SCROLLLOCK_ON,          /// 
	SHIFT      = SHIFT_PRESSED,          /// 
	ALT        = ALT_LEFT  | ALT_RIGHT,  /// 
	CTRL       = CTRL_LEFT | CTRL_RIGHT, /// 
}

/***/
struct CONSOLEKEYEVENTRECORD {
	KEY_EVENT_RECORD ConsoleKeyEventRecord;
	mixin(SMixInStructHiddenOriginal!(KEY_EVENT_RECORD)(q{ConsoleKeyEventRecord}));
	
	mixin(SMixInStructGetSet!(bool)("keyDown", q{ConsoleKeyEventRecord.bKeyDown}));
	mixin(SMixInStructGetSet!(ushort)("repeatCount", q{ConsoleKeyEventRecord.wRepeatCount}));
	mixin(SMixInStructGetSet!(KEY)("key", q{ConsoleKeyEventRecord.wVirtualKeyCode}));
	mixin(SMixInStructGetSet!(WORD)("scan", q{ConsoleKeyEventRecord.wVirtualScanCode}));

	mixin(SMixInStructGetSet!(char)("ascii", q{ConsoleKeyEventRecord.uChar.UnicodeChar}));
	mixin(SMixInStructGetSet!(wchar)("unicode", q{ConsoleKeyEventRecord.uChar.AsciiChar}));

	mixin(SMixInStructGetSet!(CONSOLEMODKEY)("mod", q{ConsoleKeyEventRecord.dwControlKeyState}));
}
/***/
struct CONSOLEMENUVENTRECORD {
	MENU_EVENT_RECORD ConsoleMenuEventRecord;
	mixin(SMixInStructHiddenOriginal!(MENU_EVENT_RECORD)(q{ConsoleMenuEventRecord}));
}

/***/
struct CONSOLEMOUSEVENTRECORD {
	MOUSE_EVENT_RECORD ConsoleMouseEventRecord;
	mixin(SMixInStructHiddenOriginal!(CONSOLEMOUSEVENTRECORD)(q{ConsoleMouseEventRecord}));

	mixin(SMixInStructGetSet!(CONSOLEPOINT)("position", q{ConsoleMouseEventRecord.dwMousePosition}));

	/**
	*/
	enum BUTTON {
		LEFT_1 = FROM_LEFT_1ST_BUTTON_PRESSED,
		LEFT_2 = FROM_LEFT_2ND_BUTTON_PRESSED,
		LEFT_3 = FROM_LEFT_3RD_BUTTON_PRESSED,
		LEFT_4 = FROM_LEFT_4TH_BUTTON_PRESSED,
		LEFT   = LEFT_1 | LEFT_2 | LEFT_3 | LEFT_4,
		RIGHT  = RIGHTMOST_BUTTON_PRESSED,
	}
	mixin(SMixInStructGetSet!(BUTTON)("button", q{ConsoleMouseEventRecord.dwButtonState}));
	mixin(SMixInStructGetSet!(CONSOLEMODKEY)("mod", q{ConsoleMouseEventRecord.dwControlKeyState}));

	private enum MOUSE_HWHEELED=0x0008;
	/***/
	enum EVENT {
		DBLCLK    = DOUBLE_CLICK,
		WHEELED_H = MOUSE_HWHEELED,
		MOVED     = MOUSE_MOVED,
		WHEELED   = MOUSE_WHEELED,
	}
	mixin(SMixInStructGetSet!(EVENT)("event", q{ConsoleMouseEventRecord.dwEventFlags}));
}

/***/
struct CONSOLEINPUTRECORD {
	INPUT_RECORD ConsoleInputRecord;
	mixin(SMixInStructHiddenOriginal!(INPUT_RECORD)(q{ConsoleInputRecord}));

	///
	enum TYPE {
		FOCUS  = FOCUS_EVENT,              /// 
		KEY    = KEY_EVENT,                /// 
		MENU   = MENU_EVENT,               /// 
		MOUSE  = MOUSE_EVENT,              /// 
		WINDOW = WINDOW_BUFFER_SIZE_EVENT, /// 
	}
	mixin(SMixInStructGetSet!(TYPE)("type", q{ConsoleInputRecord.EventType}));

	/**
	*/
	ref CONSOLEWINDOWEVENTRECORD eventWindow() {
		assert(type == TYPE.WINDOW);
		return *cast(CONSOLEWINDOWEVENTRECORD*)&ConsoleInputRecord.Event;
	}
	/**
	*/
	ref CONSOLEFOCUSEVENTRECORD eventFocus() {
		assert(type == TYPE.FOCUS);
		return *cast(CONSOLEFOCUSEVENTRECORD*)&ConsoleInputRecord.Event;
	}
	/**
	*/
	ref CONSOLEKEYEVENTRECORD eventKey() {
		assert(type == TYPE.KEY);
		return *cast(CONSOLEKEYEVENTRECORD*)&ConsoleInputRecord.Event;
	}
	/**
	*/
	ref CONSOLEMENUVENTRECORD eventMenu() {
		assert(type == TYPE.MENU);
		return *cast(CONSOLEMENUVENTRECORD*)&ConsoleInputRecord.Event;
	}
	/**
	*/
	ref CONSOLEMOUSEVENTRECORD eventMouse() {
		assert(type == TYPE.MOUSE);
		return *cast(CONSOLEMOUSEVENTRECORD*)&ConsoleInputRecord.Event;
	}
	
	
	
}


/**
コンソール。

低レベル入出力やらサイズやらが未実装。
*/
class Console: HandleRaii {
	protected alias Handle hConsole;
	protected override void Kill() {
		enforce(FreeConsole(), new ConsoleException(ERR.toText));
		console = null;
	}

	/// 
	static Console console;

	///
	NeFile stdInput, stdOutput, stdError;
	
	///
	this() {
		enforce(AllocConsole(), new ConsoleException(ERR.toText));
		super(GetConsoleWindow(), true);

		stdInput  = new ConsoleHandle(ConsoleHandle.STD_IO.INPUT);
		stdOutput = new ConsoleHandle(ConsoleHandle.STD_IO.OUTPUT);
		stdError  = new ConsoleHandle(ConsoleHandle.STD_IO.ERROR);

		console = this;
	}
	///
	void text(in Text Value) {
		SetConsoleTitle(Value.ptr);
	}

	///
	size_t write(in Text Value) {
		auto Data=Value.text;
		size_t Size;
		enforce(WriteConsole(stdOutput(), cast(void*)Data.ptr, Data.length, &Size, null), new ConsoleException(ERR.toText));

		return Size;
	}
	/**
	文字列の書き込み。

	Valueが全て書き込めた場合に改行を行う。

	Params:
		Value = 書き込む文字列。

	Returns:
		書き込んだ文字数。
		Valueが全て書き込めなかった場合は書き込めたサイズ。
		Valueが全て書き込めた場合は改行分を除いた書き込みサイズ。
	*/
	size_t writeln(in Text Value) {
		auto Size=write(Value);
		if(Size == Value.length) {
			write(Text.newline);
		}
		return Size;
	}
	///
	bool event(ref CONSOLEINPUTRECORD InputRecord) {
		size_t Size;
		return cast(bool)ReadConsoleInput(stdInput(), InputRecord.ptr, 1, &Size);
	}
	///
	CONSOLEINPUTRECORD[] event(size_t Count) {
		auto InputRecords=new CONSOLEINPUTRECORD[Count];
		size_t Size;
		enforce(ReadConsoleInput(stdInput(), InputRecords.ptr.ptr, Count, &Size), new ConsoleException(ERR.toText));
		return InputRecords[0 .. Size];
	}
	/***/
	char get() {
		CONSOLEINPUTRECORD InputRecord;
		while(true) {
			if(event(InputRecord)) {
				if(InputRecord.type == CONSOLEINPUTRECORD.TYPE.KEY) {
					immutable KeyEvent=InputRecord.eventKey;
					if(KeyEvent.keyDown && isprint(KeyEvent.ascii)) {
						return KeyEvent.ascii;
					} else switch(KeyEvent.key) {
						case KEY.ENTER, KEY.BACK:
							return char.init;
						default:
							continue;
					}
				}
			}
		}
	}
	///
	Text read(size_t Length) {
		auto Data=new tchar[Length];
		size_t Size;
		enforce(ReadConsole(stdInput(), cast(void*)Data.ptr, Data.length, &Size, null), new ConsoleException(ERR.toText));

		return Text.wrap(Data[0 .. Size]);
	}

	///
	const CONSOLEPOINT maxWindowSize() {
		auto Size= cast(CONSOLEPOINT)GetLargestConsoleWindowSize(stdOutput());
		enforce(Size.x && Size.y, new ConsoleException(ERR.toText));
		return Size;
	}
	///
	const ref CONSOLESCREENBUFFER screenBuffer() {
		auto Buffer=new CONSOLESCREENBUFFER;
		GetConsoleScreenBufferInfo(stdOutput(), Buffer.ptr);
		return *Buffer;
	}

	///
	bool screenBufferSize(const ref CONSOLEPOINT Size) {
		return cast(bool)SetConsoleScreenBufferSize(stdOutput(), Size.Coord);
	}
	
	///
	bool color(FILLATTRIBUTE ConsoleColor) {
		return cast(bool)SetConsoleTextAttribute(stdOutput(), ConsoleColor);
	}
	///
	const FILLATTRIBUTE color() {
		return screenBuffer.color;
	}

	/**
	*/
	const ref CONSOLECURSORINFO cursorInfo() {
		auto Cursor=new CONSOLECURSORINFO;
		enforce(GetConsoleCursorInfo(stdOutput(), Cursor.ptr), new ConsoleException(ERR.toText));
		return *Cursor;
	}
	/**
	*/
	bool cursorInfo(const ref CONSOLECURSORINFO Cursor) {
		return cast(bool)SetConsoleCursorInfo(stdOutput(), Cursor.ptr);
	}
	
	/**
	*/
	bool cursor(const ref CONSOLEPOINT Point) {
		return cast(bool)SetConsoleCursorPosition(stdOutput(), Point.Coord);
	}
	/**
	*/
	const CONSOLEPOINT cursor() {
		return screenBuffer.cursor;
	}



}
debug(console) unittest {
	scope a=new Console();
	CONSOLEPOINT pp;

	size_t i;
	try {
		uint b;
		//auto cs = ['a', 'b', 'c'];
		//auto n=a.stdWrite(cs);
		//a.stdPut(Text(n).toString[0]);
		//WriteConsole(a.stdOutput(), cast(void*)Text(`abc`).ptr, 3, &b, null);
		//n=a.write(Text("ABCあいうABC"));
		//a.stdPut(Text(n).toString[0]);
		//auto t=a.read(1);
		//n=a.write(t);
		char c;
		a.writeln(Text("a"));
		assert(a.color(FILLATTRIBUTE.FORE_R));
		a.write(Text("あ"));
		wl("%s, %s", a.maxWindowSize(),0);
		CONSOLEPOINT p;
		p.x = p.y = 0;
		//a.cursor = p;
		assert(a.color(FILLATTRIBUTE.BACK_R));
		//a.write(Text("x=%s, y=%s ", a.cursor.x, a.cursor.y));
		//a.read(0);
		while((c=a.get()) != char.init) {
			a.stdOutput.put(c);
		}
		
	} catch(Object o) {
		wl("%s, %s", o, ERR.toText);
	}
}
