﻿/**
文字列。

かなーりstring⇔Text変換の飛び交うﾈﾑぃの心臓部。

Note:
	調整が必要。

	てかtext, exception, timerを同一moduleにすべきなんじゃないだろうか。

History:
	1.100:
		[S] wcharをtcharに。それに伴い変更可能部分を変更中。なんにせよ-Unicodeしないと使い物ならないしあんまし意味無い。

	1.00β19:
		[P] どう見てもNeGuiに癒着していたのでpackageをnemuxi.systemからetc.negui.systemに移動。

	1.00β14:
		std.formatでよく使うやつを実装して強化したい。
*/
module etc.negui.system.text;

debug import std.stdio: wl = writefln;
debug(text) void main() {}

//import std.array;
static import std.algorithm;
import std.math;
import std.conv;
import std.utf;
import std.format;
static import std.string;
//import std.c.string;

import etc.negui.builder;

import std.ctype;
import std.uni;
import std.traits;
import std.contracts;

import win32.winbase;


///
alias wchar tchar;
///
alias immutable(tchar)[] tstring;

/**
とりあえずWindowsW文字列を扱いやすく、それでいていい加減な文字列。

構造体だけど中身配列なんでほぼ参照。
外部(Windows)に渡すような関数では大抵これが引数。

* --------------------------------------------------
* Text str;
* str = Text("aaa");
* assert(str == Text("aaa"c));
* assert(str == Text("aaa"w));
* assert(str == Text("aaa"c));
*
* str = Text(123);
* assert(str == Text("123"));
*
* str = Text("%s + %s = %s", 1, 9, 10);
* assert(str == Text("1 + 9 = 10"));
* --------------------------------------------------


Bugs:
	うわー、こいつ命名規則完全に無視してんじゃないか。
	一括置き換えでいけるんかねー。

History:
	1.100:
		[F] opCall, opAssign回りの変更とTの限定。関連moduleの修正。
		[S] opCallの分割。
		[F] 数値とかと比較できなくした。
		[F] とりあえずrefへ。
*/
struct Text {
	debug(text) unittest {
		Text str;
		str = Text("aaa");
		assert(str == Text("aaa"c));
		assert(str == Text("aaa"w));
		assert(str == Text("aaa"c));

		str = Text(123);
		assert(str == Text("123"));

		str = Text("%s + %s = %s", 1, 9, 10);
		assert(str == Text("1 + 9 = 10"));
	}

	
	/+
	/// 実態。
	wchar[] text;
	+/
	/// 実態。
	tchar[] text=null;


	/// 型。
	static nothrow pure TypeInfo charType() {
		return typeid(tchar);
	}
	/// ditto
	static nothrow pure TypeInfo textType() {
		return typeid(tchar[]);
	}

	//alias text this;

	/**
	ラップ。

	tchar[]そのものをTextに取り込む。

	Patams:
		arg = 取り込む配列。

	Returns:
		ラップ後のText。

	Examples:
	* --------------------------------
	* tstring src="abc";
	*
	* tchar[] buffer=src.dup;
	* auto t=Text.wrap(buffer);
	* 
	* t[2] = 'C';
	* assert(buffer == "abC");
	* assert(buffer is t.text);
	* --------------------------------
	
	History:
		1.100:
			新規作成。
	*/
	static ref Text wrap(tchar[] arg) {
		auto t=new Text;
		t.text = arg;
		return *t;
	}
	debug(text) unittest {
		tstring src="abc";
		
		tchar[] buffer=src.dup;
		auto t=Text.wrap(buffer);
		
		t[2] = 'C';
		assert(buffer == "abC");
		assert(buffer is t.text);
	}
	/**
	ラップ。

	Text.wrapの書き換えないバージョン。
	
	History:
		1.100:
			新規作成。
	*/
	static pure ref const(Text) iwrap(tstring arg) {
		auto t=new Text;
		t.text = cast(wchar[])arg;
		return *t;
	}

	/**
	空文字列の作成。

	History:
		1.100:
			[S] とりあえずpure/refに。

		1.00β15:
			[S] finalだったのを修正。クラスにしようか迷っていたときの名残か…。

	Deprecated:
		Text.initでいいじゃん。
	*/
	deprecated @property static pure ref Text emptyText() {
		auto t=new Text;
		t.text = null;
		return *t;
	}
	debug(text) unittest {
		//assert(!Text.emptyText.length);
	}
	///
	static immutable Text newline={text:cast(tchar[])[std.string.newline[0], std.string.newline[1]]};
	version(Windows) {
		debug(text) unittest {
			assert(Text.newline.length == 2);
			assert(Text.newline[0] == '\r');
			assert(Text.newline[1] == '\n');
		}
	}
	/**
	Windows用文字列の生成。

	Params:
		NonTextIsNull = trueの場合に文字列が空ならnullを返す。

	Returns:
		textの0終端文字列を生成。
		誰かに渡さないとGCの対象になるんで注意。

	History:
		1.100:
			[P] 挙動の変更。
			[F] 引数追加。
	*/
	wchar* ptr(bool NonTextIsNull=false) const {
		if(NonTextIsNull && !text.length) {
			return null;
		}

		return (text.dup ~ cast(wchar)0).ptr;
	}
	debug(text) unittest {
		Text t;
		t.text = "123"w.dup;
		auto p=t.ptr;
		assert(p != t.text.ptr);
		size_t i;
		while(p[i++] != '\0') {};
		assert(t.text.length + 1 == i);
	}

	/**
	文字列の配列長を取得。
	UTF-16が一文字純正だったら素敵。

	Returns:
		文字列の長さ。
	*/
	@property pure nothrow size_t length() const {
		return text.length;
	}
	debug(text) unittest {
		Text t;
		t.text = "123456"w.dup;
		assert(t.length == "123456"w.length);
		t.text = "あいうえお"w.dup;
		assert(t.length == 5);
	}
	/**
	文字列の配列長の設定。

	Params:
		Size = 設定する配列長。

	History:
		1.100:
			[S] 配列のサイズを返す。
			[S] nothrow。
	*/
	@property nothrow size_t length(size_t Size) {
		return text.length = Size;
	}
	debug(text) unittest {
		Text t;
		t.text = "123"w.dup;
		t.length = 1;
		assert(t.text == "1");
	}

	/**
	Unicode文字列判定。

	現在の文字列がUnicodeで合法かどうかを判定。
	
	History:
		1.100:
			新規作成。
	*/
	const bool validate() {
		return collectException(std.utf.validate(text)) is null;
	}
	/**
	文字列長取得。

	現在のUnicode文字列長を取得。
	length()と違い文字列の長さを取得する。

	Returns:
		文字列長。

	In:
		文字列がUnicodeとして正しいこと。
	
	History:
		1.100:
			新規作成。
	*/
	@property const size_t countCharacter()
	in {
		assert(validate());
	}
	body {
		return std.utf.count(text);
	}

	/**
	文字列のコピー。

	Returns:
		コピーされた文字列。
	*/
	ref pure Text dup() const {
		/+
		Text t;
		t.text = text.dup;
		return t;
		+/
		auto t=new Text;
		t.text = text.dup;
		return *t;
	}
	debug(text) unittest {
		Text t;
		t.text = "abc"w.dup;
		Text t2=t.dup;
		assert(t.text !is t2.text);
		assert(t.text == t2.text);
	}
	/**
	文字列のimmutableなコピー。

	Returns:
		Dのwstring。
	*/
	tstring idup() const {
		return text.idup;
	}

	/**
	数値変換。

	現在の文字列からTに対応する数値を取得。

	Returns:
		Tに対応する数値。

	Throws:
		変換に失敗した場合はNeGuiException。

	Examples:
	* --------------------------------
	* auto t=Text("123");
	* assert(t.toNumber!(int) == 123);
	* --------------------------------
	
	History:
		1.100:
			新規作成
	*/
	const T toNumber(T, string _file=__FILE__, int _line=__LINE__)()
	if(std.traits.isNumeric!(T))
	{
		try {
			return to!(T)(text);
		} catch(Throwable e) {
			throw new NeGuiException(Text("%s toNumber(), text -> %s", T.stringof, text), e, _file, _line);
		}
	}
	debug(text) unittest {
		Text t;
		t = Text("123");
		assert(t.toNumber!(int) == 123);
		t = Text("123");
		assert(t.toNumber!(byte) == 123);
		t = Text("123");
		assert(t.toNumber!(byte) == 123);
		t = Text("1.1");
		assert(t.toNumber!(real) == 1.1);
	}
	/**
	数値変換。

	toNumber()と違い例外出力を行わない。

	Params:
		Value = 変換後の数値。

	Returns:
		変換成功の真偽値。
	
	Examples:
	* --------------------------------
	* auto t=Text("123");
	* int result;
	* if(t.tryNumber!(int)(result)) {
	* 	assert(result == 123);
	* }
	* --------------------------------
	
	History:
		1.100:
			新規作成
	*/
	const bool tryNumber(T)(out T Value) {
		try {
			Value = toNumber!(T);
			return true;
		} catch {
			return false;
		}
	}
	debug(text) unittest {
		Text t;
		t = Text("123");
		int a;
		assert(t.tryNumber!(int)(a));
		assert(!t.init.tryNumber!(int)(a));
	}
	/**
	数値に区切り追加。

	1000を1,000のように変換。

	Params:
		number = 変換する数値。

		SplitChar = 区切りにする文字。

		SplitCount = SplitCharを設定する桁数。

	Returns:
		区切り文字の追加された文字列。

	In:
		SplitCount > 0
	
	Examples:
	* --------------------------------
	* assert(Text.numberSplit(1000) == Text("1,000"));
	* assert(Text.numberSplit(1000, '.') == Text("1.000"));
	* assert(Text.numberSplit(1000, ',', 1) == Text("1,0,0,0"));
	* --------------------------------
	
	History:
		1.100:
			新規作成。
	*/
	static ref Text numberSplit(T)(T number, tchar SplitChar=',', size_t SplitCount=3)
	if(isIntegral!(T))
	in {
		assert(SplitCount > 0);
	}
	body {
		auto s=to!(string)(number);
		immutable Minus=(number < 0) ? 1: 0;
		
		if(s.length - Minus <= SplitCount) {
			return Text(s);
		}
		
		immutable Counter=(s.reverse.length - 1 - Minus) / SplitCount;
		auto t=new tchar[s.length + Counter];

		for(auto i=0, j=0; i < s.length; i++, j++) {
			if(i && !(i % SplitCount)) {
				t[j++] = SplitChar;
			}
			t[j] = s[i];
		}
		
		return Text.wrap(t.reverse);
	}
	debug(text) unittest {
		assert(Text.numberSplit(1000) == Text("1,000"));
		assert(Text.numberSplit(1000, '.') == Text("1.000"));
		assert(Text.numberSplit(1000, ',', 1) == Text("1,0,0,0"));
	
		assert(numberSplit(1) == Text("1"));
		assert(numberSplit(0) == Text("0"));
		assert(numberSplit(100) == Text("100"));
		assert(numberSplit(-100) == Text("-100"), numberSplit(-100).str);
		assert(numberSplit(999) == Text("999"));
		assert(numberSplit(1000) == Text("1,000"));
		assert(numberSplit(-1000) == Text("-1,000"), numberSplit(-1000).str);
		assert(numberSplit(10000) == Text("10,000"));
		assert(numberSplit(100000) == Text("100,000"));
		assert(numberSplit(1000000) == Text("1,000,000"));
		assert(numberSplit(10000000) == Text("10,000,000"));
		assert(numberSplit(100000000) == Text("100,000,000"));
		assert(numberSplit(1000000000) == Text("1,000,000,000"));
		assert(numberSplit(-1000000000) == Text("-1,000,000,000"));
		
		assert(numberSplit(1000000000, '.') == Text("1.000.000.000"));
		assert(numberSplit(1000000000, '.', 1) == Text("1.0.0.0.0.0.0.0.0.0"));
		assert(numberSplit(1000000000, ':', 10) == Text("1000000000"));
		assert(numberSplit(10000000000, ':', 10) == Text("1:0000000000"));
		assert(numberSplit(-10000000000, ':', 10) == Text("-1:0000000000"));
	}
	
	private immutable {
		auto NUMERAL_ZERO_CHAR=
			"0123456789"
			"ABCDEFGHIJKLM"
			"NOPQRSTUVWXYZ"
		;
		auto NUMERAL_ZERO_NUM = [
			0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
			11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
			24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36
		];
		static assert(NUMERAL_ZERO_CHAR.length == NUMERAL_ZERO_NUM.length);
	}
	/**
	文字からn進数取得。

	Params:
		c = n進数の最大値を表す文字。

	Returns:
		n進数。

	Throws:
		失敗時にNeGuiException。

	Note:
		一応公開してるけどユーザー側で使うことはまず無い。
	*/
	static int numeralChatToNum(char c) {
		int n;
		if('0' <= c && c <= '9') {
			n = c - '0';
		} else {
			n = std.string.indexOf(NUMERAL_ZERO_CHAR, c, std.string.CaseSensitive.no);
		}
		enforce(n >= 0, new NeGuiException(Text("c -> %s in [%s]", c, NUMERAL_ZERO_CHAR)));
		
		return NUMERAL_ZERO_NUM[n];
	}

	/**
	数値をn進数に変換。

	History:
		1.100:
			新規作成。
	*/
	static ref Text toNumeral(T, BASE: int)(in T number, in BASE n=16, in size_t Increment=BUFFER.INCREMENT)
	if(isIntegral!(T) && !isSomeChar!(BASE))
	in {
		assert(2 <= n && n <= NUMERAL_ZERO_CHAR.length);
	}
	body {
		Unqual!T Number=number;
		tchar[] t;
		size_t i;
		for(; Number; i++, Number /= n) {
			t.autoIncrement(i, Increment)[i] = NUMERAL_ZERO_CHAR[Number % n];
		}
		return Text.wrap(t[0..i].reverse);
	}
	/// ditto
	static ref Text toNumeral(T, BASE=char)(in T number, in BASE c='F', in size_t Increment=BUFFER.INCREMENT)
	if(isIntegral!(T) && isSomeChar!(BASE))
	{
		return toNumeral(number, numeralChatToNum(c), Increment);
	}
	/**
	n進数から10進数へ変換。
	*/
	static T toNumeral10(T, BASE: int)(in Text Numeral, BASE n)
	if(isIntegral!(T) && !isSomeChar!(BASE))
	in {
		assert(2 <= n && n <= NUMERAL_ZERO_CHAR.length);
	}
	body {
		T Number;
		size_t i = Numeral.text.length - 1;
		foreach(c; Numeral.text) {
			Number += numeralChatToNum(cast(char)c) * pow(cast(int)n, i--);
		}
		return Number;
	}
	/// ditto
	static T toNumeral10(T, BASE: char)(in Text Numeral, BASE c)
	if(isIntegral!(T) && isSomeChar!(BASE))
	{
		return toNumeral10!(T)(Numeral, numeralChatToNum(c));
	}
	debug(text) unittest {
		assert(toNumeral(2, 2).text == "10");
		assert(toNumeral(3, 2).text == "11");
		assert(toNumeral(4, 2).text == "100");
		assert(toNumeral(4, 16).text == "4");
		assert(toNumeral(10, 16).text == "A");
		assert(toNumeral(15, 16).text == "F");
		assert(toNumeral(16, 16).text == "10");
		assert(toNumeral(35, 36).text == "Z");
		assert(toNumeral(36, 36).text == "10");
		assert(toNumeral(2, '2').text == "10");
		assert(toNumeral(3, '2').text == "11");
		assert(toNumeral(4, '2').text == "100");
		assert(toNumeral(4, 'f').text == "4");
		assert(toNumeral(10, 'F').text == "A");
		assert(toNumeral(15, 'f').text == "F",toNumeral(15, 'f').str);
		assert(toNumeral(16, 'F').text == "10");
		assert(toNumeral(35, 'Z').text == "Z");
		assert(toNumeral(36, 'z').text == "10");
		assert(toNumeral10!(int)(Text("10"), 16) == 16);
		assert(toNumeral10!(int)(toNumeral(16, 16), 16) == 16);
		assert(toNumeral10!(int)(toNumeral(36, 'Z'), 36) == 36);
		assert(toNumeral10!(int,char)(toNumeral(36, 'Z'), 'z') == 36);
	}

	/// ソート。
	ref Text sort() {
		text.sort;
		return this;
	}
	/// 反転。
	ref Text reverse() {
		text.reverse;
		return this;
	}
	/**
	死んでしまえ。

	History:
		1.00β14:
			空の場合に若干変更
	*/
	const string toString() {
		return text8;//text.length ? toUTF8(text): null;
	}
	/**
	History:
		1.00β19:
			新規作成。
	*/
	string toString() {
		return text8;
	}
	/**
	UTF-8文字列取得。
	
	History:
		1.090:
			[S] 名前変更(text8 -> str)

		1.00β19:
			新規作成。
	*/
	const string str() {
		return text.length ? toUTF8(text): null;
	}
	/// UTF-9文字列。
	alias str text8;
	/// UTF-16文字列。
	alias idup text16;
	/// UTF-32文字列。
	dstring text32() const {
		return std.utf.toUTF32(text);
	}
	///
	const(char*) toStringz() const {
		return std.string.toStringz(text8);
	}

	/// 連結。
	const ref Text opCat(in Text Value) {
		return Text(text ~ Value.text);
	}
	debug(text) unittest {
		Text t;
		t = Text("123");
		assert((t ~ Text("456")).text == "123456");
	}

	/// 連結。
	ref Text opCatAssign(in Text Value) {
		text ~= Value.text;
		return this;
	}
	debug(text) unittest {
		Text t;
		t.text = "aaa"w.dup;

	}

	private static wchar[] TextFormat(...) {
		//wchar[] s;
		//wchar[2] Buffer;
		static if(is(tchar == wchar)) {
			tchar[2] Buffer;
		}
		//auto app=appender!(wchar[]);//(&s);
		auto app=appender!(tchar[]);//(&s);
		void putc(dchar c) {
			immutable Length=encode(Buffer, c);
			//app.put!(wchar[])(Buffer[0 .. Length]);
			app.put!(tchar[])(Buffer[0 .. Length]);
		}
		doFormat(&putc, _arguments, _argptr);
		return app.data;
	}
	debug(text) unittest {
		auto t1=TextFormat("%0.*s".idup, 10, 59);
		auto t2=Text      ("%0.*s"w, 10, 59);
		int a=10;
		auto t3=Text      ("%0.*s"w, a, 59);
	}
	/**
	文字からTextを生成。

	Examples:
	* --------------------------------
	* Text t;
	* 
	* t = Text('a');
	* assert(t == Text("a"));
	* t = Text('あ');
	* assert(t == Text("あ"));
	* --------------------------------
	
	*/
	static ref Text opCall(T)(in T arg)
	if(isSomeChar!(T))
	{
		auto ResultText=new Text;
		static if(T.sizeof > tchar.sizeof) {
			encode(ResultText.text, arg);
		} else {
			ResultText.text = [arg];
		}
		return *ResultText;
	}
	debug(text) unittest {
		Text t;
		t = Text('a');
		assert(t == Text("a"));
		t = Text('あ');
		assert(t == Text("あ"));
	}
	
	/**
	C形式文字列ポインタからText生成。
	
	Examples:
	* --------------------------------
	* Text t;
	*
	* t = Text(x"313233003435360037383900".ptr);
	* assert(t.text == "123");
	* t = Text(x"313233003435360037383900"w.ptr);
	* assert(t.text == "123");
	* t = Text(x"313233003435360037383900"d.ptr);
	* assert(t.text == "123");
	* --------------------------------
	*/
	static ref Text opCall(T)(in T* arg)
	if(isSomeChar!(T))
	{
		auto ResultText=new Text;
		
		if(arg !is null) {
			size_t i=0;
			while(arg[i]) i++;
			
			ResultText.text= to!(tstring)(arg[0 .. i]).dup;
		} else {
			ResultText.text = null;
		}
		
		return *ResultText;
	}
	debug(text) unittest {
		Text t;
		
		t = Text(x"313233003435360037383900".ptr);
		assert(t.text == "123");
		t = Text(x"313233003435360037383900"w.ptr);
		assert(t.text == "123");
		t = Text(x"313233003435360037383900"d.ptr);
		assert(t.text == "123");
	}
	
	/**
	クラスからText生成。

	TはITextを継承していること。
	*/
	static ref Text opCall(T)(in T arg)
	if(is(T: IText))
	{
		if(arg) {
			return arg.toText[];
		} else {
			return Text.wrap("null"w.dup);
		}
	}
	debug(text) unittest {
		class itext: IText {
			const Text toText() {
				return Text("aaa");
			}
			override string toString() {
				return "bbb";
			}
		}
		auto o=new itext();
		auto t=Text(o);
		t == Text("aaa");
	}
	/**
	構造体からText生成。

	TはtoTextを持っていること。
	*/
	static ref Text opCall(T)(const ref T arg)
	if(is(T == struct) && is(typeof(&T.init.toText)))
	{
		return arg.toText[];
	}
	debug(text) unittest {
		struct S1 {
			const Text toText() {
				return Text("a");
			}
			string toString() {
				return "b";
			}
		}
		struct S2 {
			const string toString() {
				return "c";
			}
		}
		S1 s1;
		S2 s2;
		auto t1=Text(s1);
		auto t2=Text(s2);
		assert(t1 == Text("a"));
		assert(t2 == Text("c"));
	}
	/**
	TextからText生成。

	Text.dup()を内部で呼び出し。
	*/
	static ref Text opCall(T: Text)(const ref T arg) {
		return arg.dup;
	}
	/**
	引数からText生成。

	Params:
		arg = inなのでObject.toStringはできない。
	*/
	static ref Text opCall(T)(in T arg) {
		auto ResultText=new Text;
		ResultText.text = (to!(tstring)(arg)).dup;
		return *ResultText;
	}
	debug(text) unittest {
		Text t;
		t = Text(`a`);
		assert(t.text == "a");
		t = Text(123);
		assert(t.text == "123");

		Text c = "c";
		Text w = "w"w;
		Text d = "d"d;
	}
	/**
	書式からText生成。

	Params:
		SrcFormat = 書式。
	
	Note:
		std.format.doFormat参照。

	*/
	static ref Text opCall(String, T...)(in String SrcFormat, T Args) {
		auto ResultText=new Text;
		//Text ResultText;
		//wchar[] FormatString;

		static if(is(String == Text)) {
			auto FormatString = SrcFormat.text;
		} else static if(isSomeString!String)  {
			auto FormatString = std.utf.toUTF16(SrcFormat.text);
		} else {
			static assert("String is " ~ String.stringof);
		}
		ResultText.text = TextFormat(FormatString, Args);

		return *ResultText;
	}
	debug(text) unittest {
		auto text=new wchar[100];
		text[0]='a';
		for(auto i='A', j=0; i < 'Z'+1; i++, j++)
		text[j] = i;
		text[26]=0;
		auto t=Text(text.ptr);
		assert(t.text == text[0..26], t.toString);
		auto a=Text(132);
		assert(a.text=="132");
		assert(Text(123).text == "123");
		//assert(Text(1234) == Text(123) ~ 4);
		assert(a == Text(132));
		assert(a == Text(`132`));
		//assert(a == Text(`13`) ~ 2);
		assert(a == Text(`13`) ~ Text(`2`));
		assert(a == Text(`13`) ~ Text('2'));
		auto b=Text(2);
		assert(a == (Text(`13`) ~ b));
		assert(a == Text(`13`) ~ Text(2));

		//Text t;
		t=Text("aaa");
		assert(t.text == "aaa");
		t=Text("bbb"w);
		assert(t.text == "bbb");
		t=Text("ccc"d);
		assert(t.text == "ccc");
		t=Text(__VENDOR__.dup);
		assert(t == Text(__VENDOR__));

		t=Text("aaa\0"w.dup.ptr);
		assert(t.text == "aaa");
		t=Text(Text("vvv"));
		assert(t.text == "vvv");

		t=Text(100);
		assert(t.text == "100");
		t=Text(1.5);
		assert(t.text == "1.5");
		t=Text(1.5);
		assert(t.text == "1.5");

		t=Text("%s", __VENDOR__);
		assert(t == Text(__VENDOR__));
		t=Text("<%s>", __TIMESTAMP__);
		assert(t == Text("<" ~ __TIMESTAMP__ ~ ">"));
		t=Text(Text("a=%05s, b=%s, c=%s"), 1, 1.5, "ABC");
		assert(t.text == "a=00001, b=1.5, c=ABC");
		t = Text("<%s>", 100);
		assert(t.text == "<100>");
		t = Text("<%s>"w, 100);
		assert(t.text == "<100>");
		t = Text("<%s>"d, 100);
		assert(t.text == "<100>");
		t = Text(Text("<%s>"), 100);
		assert(t.text == "<100>");

		//wl(__FILE__);
		//? t = Text("%.*s",2,12);
		//assert(t.text == "<100>");
	}

	/// [n]
	const tchar opIndex(size_t n) {
		return text[n];
	}
	/// [n] = value
	tchar opIndexAssign(tchar value, size_t n) {
		return text[n] = value;
	}
	/// []
	ref Text opSlice() {
		return this;//Text(text[]);
	}
	/**
	[n..m]

	History:
		1.00β17:
			[B] 事前条件がDの仕様と違っていたのを修正。
	*/
	const ref Text opSlice(in size_t start, in size_t end)
	in {
		assert(start <= end, std.string.format("%s[%s..%s]", text, start, end));
	}
	body {
		/+
		Text t;
		t.text = text[start..end].dup;
		return t;
		+/
		auto t=new Text;
		t.text = text[start..end].dup;
		return *t;
	}
	debug(text) unittest {
		Text t;
		t = Text("LICENSE");
		assert(t[0..1].text=="L");
		assert(t[0..2].text=="LI");
		assert(Text(t.text[0..3]).text == "LIC");
		assert(Text(t.text[0..7]).text == "LICENSE");
	}

	/+
	/**
	==
	*/
	bool opEquals(T)(in T arg) const {
		static if(is(T == Text)) {
			return text == arg.text;
		} else {
			return text == to!(wstring)(arg);
		}
	}
	+/
	/+
const hash_t toHash();
const bool opEquals(ref const KeyType s);
const int opCmp(ref const KeyType s);
	+/
	///
	const hash_t toHash() {
		hash_t Hash;

		foreach(c; text) {
			Hash = (Hash * text.length) + c;
		}

		return Hash;
	}
	///
	const bool opEquals(const ref Text arg) {
		return text == arg.text;
	}
	///
	const int opCmp(const ref Text arg) {
		return cmp(arg);
	}
	debug(text) unittest {
		Text[const(Text)] t;
		t[Text("aaa")] = Text("@@@@");
		assert(t[Text(`aaa`)] == Text(`@@@@`));
		auto k=Text.iwrap("aaa");
		assert(t[k] == Text(`@@@@`));
	}

	/+
	/**
	>= > < <=
	*/
	int opCmp(T)(in T arg) const {
		static if(is(T == Text) || is(T == const(Text))) {
			return cmp(arg);
		} else {
			return cmp(text, to!(wstring)(arg));
		}
	}
	+/

	debug(text) unittest {
		assert(Text(123).text == "123");
		assert(Text.init.text == "");
		assert(Text(123) == Text(123));
		assert(!(Text(123) > Text(456)));
	}

	/**
	文字列比較。
	
	History:
		1.00β11:
			新規作成。
	*/
	const int cmp(in Text text) {
		return std.string.cmp(this.text, text.text);
	}

	/**
	文字列比較。
	
	大文字小文字を区別しない。
	
	History:
		1.100:
			[S] 名前変更(cmpi -> icmp)。

		1.00β11:
			新規作成。
	*/
	const int icmp(in Text text) {
		return lstrcmpi(ptr, text.ptr);
	}
	deprecated alias icmp cmpi;


	/**
	文字列からDelimiter削除。

	Params:
		Delimiter = 取り去る文字列。
		            Text.initならCR,LF,CRLF、

	Note:
		まんまstd.string.chomp。
	
	History:
		1.100:
			[P] 引数追加。
	*/
	const ref Text chomp(in Text Delimiter=Text.init) {
		return Text(std.string.chomp(text, Delimiter.text));
	}
	debug(text) unittest {
		Text t="abc\r\n";
		assert(t.chomp.text == "abc", t.toString());
	}

	/**
	先頭空白文字削除。

	History:
		1.063:
			[B] 全部ホワイトスペースでアウト。

		1.010:
			[S] 属性変更。
	*/
	const ref Text stripl() {
		if(!text.length) return Text.init;
		size_t i;
		while(text.isSafe(i) && isspace(text[i])) i++;
		return Text(text[i..$]);
	}
	debug(text) unittest {
		Text t=" abc";
		assert(t.stripl.text == "abc", t.toString());
		t = Text("                              abc");
		assert(t.stripl.text == "abc", t.toString());
		t = Text("\t\t\t\t\tabc");
		assert(t.stripl.text == "abc", t.toString());
		t = Text("\r\r\r\r\rabc");
		assert(t.stripl.text == "abc", t.toString());
		t = Text("\n\n\n\n\nabc");
		assert(t.stripl.text == "abc", t.toString());
		t = Text(" ");
		assert(t.stripl.text == "", t.toString());
	}
	/**
	末尾空白文字削除。
	
	History:
		1.063:
			[B] 全部ホワイトスペースでアウト。

		1.010:
			[S] 属性変更。
	*/
	const ref Text stripr() {
		if(!text.length) return Text.init;
		size_t i=text.length-1;
		while(text.isSafe(i) && isspace(text[i])) i--;
		return Text(text[0..i+1]);
	}
	debug(text) unittest {
		Text t="abc ";
		assert(t.stripr.text == "abc", t.toString());
		t = Text("abc                              ");
		assert(t.stripr.text == "abc", t.toString());
		t = Text("abc\t\t\t\t\t");
		assert(t.stripr.text == "abc", t.toString());
		t = Text("abc\r\r\r\r\r");
		assert(t.stripr.text == "abc", t.toString());
		t = Text("abc\n\n\n\n\n");
		assert(t.stripr.text == "abc", t.toString());
		t = Text(" ");
		assert(t.stripr.text == "", t.toString());
	}
	/**
	両端空白文字削除。
	
	History:
		1.010:
			[S] 属性変更。
	*/
	const ref Text strip() {
		return this.stripl.stripr;
	}
	debug(text) unittest {
		Text t="      abc      ";
		assert(t.strip.text == "abc", t.toString());
	}

	
	/***/
	enum CASE_SENSITIVE {
		YES, /// 
		NO,  /// 
	}
	/***/
	enum START {
		HEAD, /// 
		TAIL, /// 
	}
	
	private static {
		/**
		大文字小文字を区別する
		*/
		bool DistinguishMatch(dchar c1, dchar c2) {
			return c1 == c2;
		}
		/**
		大文字小文字を区別しない
		*/
		bool IgnoreMatch(dchar c1, dchar c2) {
			if(isUniAlpha(c1) && isUniAlpha(c2)) {
				return toUniUpper(c1) == toUniUpper(c2);
			} else {
				return DistinguishMatch(c1, c2);
			}
		}
		pure nothrow bool function(dchar, dchar) GetCaceSensitiveFunc(CASE_SENSITIVE CaseSensitive) {
			final switch(CaseSensitive) {
				case CASE_SENSITIVE.YES: return &DistinguishMatch;
				case CASE_SENSITIVE.NO:  return &IgnoreMatch;
			}
		}
	}


	
	/**
	文字列カウント。
	
	Examples:
	* --------------------------------
	* Text t;
	*
	* t = Text("abc;d;e;;f;g");
	* assert(t.count(';') == 5);
	*
	* t = Text("abcdefg");
	* assert(t.count(';') == 0);
	*
	* t=Text("");
	* assert(t.count(';') == 0);
	*
	* t=Text(";");
	* assert(t.count(';') == 1);
	* --------------------------------
	*/
	const size_t count(in tchar c) {
		size_t Counter=0;

		foreach(t; text) {
			if(t == c) {
				Counter++;
			}
		}
		
		return Counter;
	}
	debug(text) unittest {
		Text t;
		
		t = Text("abc;d;e;;f;g");
		assert(t.count(';') == 5);

		t = Text("abcdefg");
		assert(t.count(';') == 0);

		t=Text("");
		assert(t.count(';') == 0);
		
		t=Text(";");
		assert(t.count(';') == 1);
	}
	/**
	文字列カウント。


	Examples:
	* --------------------------------
	* Text t;
	*
	* t=Text("123aa456aa789");
	* assert(t.count(Text("aa")) == 2);
	*
	* t=Text("123aa456aa789aa");
	* assert(t.count(Text("aa")) == 3);
	*
	* t=Text("aaaaaaaa");
	* assert(t.count(Text("aa")) == 4);
	*
	* t=Text("aaaaaaaaa");
	* assert(t.count(Text("aa")) == 4);
	*
	* t=Text("");
	* assert(t.count(Text("aa")) == 0);
	* --------------------------------

	
	History:
		1.100:
			[B] 保持している文字列長が0でRange violation。
	
		1.00β14:
			新規作成。
	*/
	const size_t count(in Text Word)
	in {
		assert(Word.length);
	}
	body {
		if(!text.length) {
			return 0;
		}
		
		immutable WordLength = Word.length;
		
		if(WordLength == 1) {
			return count(Word[0]);
		}

		size_t Count=0;
		for(auto i=0; i < text.length - WordLength+1; i++) {
			if(text[i .. i + WordLength] == Word.text) {
				i += WordLength-1;
				Count++;
				continue;
			}
		}

		return Count;
	}
	debug(text) unittest {
		Text t;

		t=Text("123aa456aa789");
		assert(t.count(Text("aa")) == 2);
		
		t=Text("123aa456aa789aa");
		assert(t.count(Text("aa")) == 3);
		
		t=Text("aaaaaaaa");
		assert(t.count(Text("aa")) == 4);
		
		t=Text("aaaaaaaaa");
		assert(t.count(Text("aa")) == 4);
		
		t=Text("");
		assert(t.count(Text("aa")) == 0);
	}

	/**
	History:
		1.100:
			[F] 文字列に対応。
			[B] std.string.lastIndexOfの挙動と違う。
			[S] 引数変更。
	*/
	private static int Find(in tchar[] SrcWord, in tchar[] SrcSub, CASE_SENSITIVE CaseSensitive, START Start) {
		if(!SrcSub.length || SrcSub.length != 1 && SrcWord.length < SrcSub.length) {
			return -1;
		}
		/+
		auto Match= Ignore ? &IgnoreMatch: &DistinguishMatch;
		+/
		/+
		bool function(dchar a, dchar b) Match;
		final switch(CaseSensitive) {
			case CASE_SENSITIVE.YES: Match = &DistinguishMatch; break;
			case CASE_SENSITIVE.NO:  Match = &IgnoreMatch;      break;
		}
		+/
		auto Match=GetCaceSensitiveFunc(CaseSensitive);
		/+
		auto Word = Head ? SrcWord.idup: SrcWord.idup.reverse;
		auto Sub  = Head ? SrcSub.idup:  SrcSub.idup.reverse;
		+/
		tstring Word, Sub;
		final switch(Start) {
			case START.HEAD:
				Word = SrcWord.idup;
				Sub  = SrcSub.idup;
				break;
			case START.TAIL:
				Word = SrcWord.idup.reverse;
				Sub  = SrcSub.idup.reverse;
				break;
		}

		immutable MaxCount=Sub.length;
		bool Hit=true;
		for(auto i=0, j=0; i < Word.length || !Hit && Word.length < MaxCount; i++) {
			Hit = Match(Word[i], Sub[j]);
			if(Hit) {
				if(++j == MaxCount) {
					//if(Head) {//}
					if(Start == START.HEAD) {
						return i - (j - 1);
					} else {
						return Word.length - i - 1;
					}
				}
			} else {
				j = 0;
			}
		}

		return -1;
	}

	/**
	Deprecated:
		Text.indexOfを使用すべき。
	*/
	deprecated {
		/**
		Note:
			多すぎる。
			std.string.indexOf, std.string.lastIndexOfにあわせるべき。
		*/
		const int find(in tchar c) {
			return indexOf(c, CASE_SENSITIVE.YES, START.HEAD);
		}
		/// ditto
		const int ifind(in tchar c) {
			return indexOf(c, CASE_SENSITIVE.NO, START.HEAD);
		}
		/// ditto
		const int rfind(in tchar c) {
			return indexOf(c, CASE_SENSITIVE.YES, START.TAIL);
		}
		/// ditto
		const int irfind(in tchar c) {
			return indexOf(c, CASE_SENSITIVE.NO, START.TAIL);
		}
		/// ditto
		const int find(in Text Word) {
			return indexOf(Word, CASE_SENSITIVE.YES, START.HEAD);
		}
		/// ditto
		const int ifind(in Text Word) {
			return indexOf(Word, CASE_SENSITIVE.NO, START.HEAD);
		}
		/// ditto
		const int rfind(in Text Word) {
			return indexOf(Word, CASE_SENSITIVE.YES, START.TAIL);
		}
		/// ditto
		const int irfind(in Text Word) {
			return indexOf(Word, CASE_SENSITIVE.NO, START.TAIL);
		}
		debug(text) unittest {
			Text t="123456789abcdABCD";
			assert(t.find('1') == 0);
			assert(t.find('9') == 8);
			assert(t.find('a') == 9);
			assert(t.find('A') == 13);
			assert(t.find(Text("1")) == 0);
			assert(t.find(Text("abcd")) == 9);

			assert(t.ifind('1') == 0);
			assert(t.ifind('9') == 8);
			assert(t.ifind('a') == 9);
			assert(t.ifind('A') == 9);
			assert(t.find(Text("1")) == 0);
			assert(t.find(Text("12")) == 0);
			assert(t.ifind(Text("ABCD")) == 9);
			assert(t.ifind(Text("123456789abcdABCD")) == 0);
			assert(t.ifind(Text("123456789abcdABCDE")) == -1);

			assert(t.rfind('D') == 16);
			assert(t.irfind('d') == 16);
			assert(t.rfind('1') == 0);
			assert(t.rfind(Text("12")) == 0);
			assert(t.rfind(Text("23")) == 1);
			assert(t.rfind(Text("abcd")) == 9);
			assert(t.irfind(Text("ABCD")) == 13);
			assert(t.irfind(Text("abcd")) == 13);
		}
	}
	/+
	/*
	debug(text) unittest {
		auto n=reg.match("abcabcabab", reg.regex("ab"));
		wl("%s", n.hit);
	}
	*/
	+/
	/**
	文字検索。

	Params:
		c = 検索する文字。
	
		CaseSensitive = 大文字小文字の区別。
	
		Start = 検索位置。

	Returns:
		見つかった場合はそのインデックス、
		見つからなかった場合は-1を返す。

	Examples:
	* --------------------------------
	* Text t;
	* t = Text("123456789abcdABCD");
	*
	* assert(t.indexOf('1', CASE_SENSITIVE.YES, START.HEAD) == 0);
	* assert(t.indexOf('9', CASE_SENSITIVE.YES, START.HEAD) == 8);
	* assert(t.indexOf('a', CASE_SENSITIVE.YES, START.HEAD) == 9);
	* assert(t.indexOf('A', CASE_SENSITIVE.YES, START.HEAD) == 13);
	*
	* assert(t.indexOf('1', CASE_SENSITIVE.NO, START.HEAD) == 0);
	* assert(t.indexOf('9', CASE_SENSITIVE.NO, START.HEAD) == 8);
	* assert(t.indexOf('a', CASE_SENSITIVE.NO, START.HEAD) == 9);
	* assert(t.indexOf('A', CASE_SENSITIVE.NO, START.HEAD) == 9);
	*
	* assert(t.indexOf('D', CASE_SENSITIVE.YES, START.TAIL) == 16);
	* assert(t.indexOf('d', CASE_SENSITIVE.NO, START.TAIL) == 16);
	* assert(t.indexOf('1', CASE_SENSITIVE.NO, START.TAIL) == 0);
	* --------------------------------
	*/
	const int indexOf(tchar c, CASE_SENSITIVE CaseSensitive=CASE_SENSITIVE.YES, START Start=START.HEAD) {
		return Find(text, [c], CaseSensitive, Start);
	}
	debug(text) unittest {
		Text t;
		t = Text("123456789abcdABCD");
		
		assert(t.indexOf('1', CASE_SENSITIVE.YES, START.HEAD) == 0);
		assert(t.indexOf('9', CASE_SENSITIVE.YES, START.HEAD) == 8);
		assert(t.indexOf('a', CASE_SENSITIVE.YES, START.HEAD) == 9);
		assert(t.indexOf('A', CASE_SENSITIVE.YES, START.HEAD) == 13);
		
		assert(t.indexOf('1', CASE_SENSITIVE.NO, START.HEAD) == 0);
		assert(t.indexOf('9', CASE_SENSITIVE.NO, START.HEAD) == 8);
		assert(t.indexOf('a', CASE_SENSITIVE.NO, START.HEAD) == 9);
		assert(t.indexOf('A', CASE_SENSITIVE.NO, START.HEAD) == 9);
		
		assert(t.indexOf('D', CASE_SENSITIVE.YES, START.TAIL) == 16);
		assert(t.indexOf('d', CASE_SENSITIVE.NO, START.TAIL) == 16);
		assert(t.indexOf('1', CASE_SENSITIVE.NO, START.TAIL) == 0);
	}
	/**
	文字検索。

	Params:
		Word = 検索する文字列。
	
		CaseSensitive = 大文字小文字の区別。
	
		Start = 検索位置。

	Returns:
		見つかった場合はそのインデックス、
		見つからなかった場合は-1を返す。

	Examples:
	* --------------------------------
	* Text t;
	* t = Text("123456789abcdABCD");
	* 
	* assert(t.indexOf(Text("1"), CASE_SENSITIVE.YES, START.HEAD) == 0);
	* assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.YES, START.HEAD) == 9);
	* 
	* assert(t.indexOf(Text("1"), CASE_SENSITIVE.NO, START.HEAD) == 0);
	* assert(t.indexOf(Text("12"), CASE_SENSITIVE.NO, START.HEAD) == 0);
	* assert(t.indexOf(Text("ABCD"), CASE_SENSITIVE.NO, START.HEAD) == 9);
	* assert(t.indexOf(Text("123456789abcdABCD"), CASE_SENSITIVE.NO, START.HEAD) == 0);
	* assert(t.indexOf(Text("123456789abcdABCDE"), CASE_SENSITIVE.NO, START.HEAD) == -1);
	* 
	* assert(t.indexOf(Text("12"), CASE_SENSITIVE.YES, START.TAIL) == 0);
	* assert(t.indexOf(Text("23"), CASE_SENSITIVE.YES, START.TAIL) == 1);
	* assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.YES, START.TAIL) == 9);
	* assert(t.indexOf(Text("ABCD"), CASE_SENSITIVE.NO, START.TAIL) == 13);
	* assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.NO, START.TAIL) == 13);
	* --------------------------------

	*/
	const int indexOf(in Text Word, CASE_SENSITIVE CaseSensitive=CASE_SENSITIVE.YES, START Start=START.HEAD) {
		return Find(text, Word.text, CaseSensitive, Start);
	}
	debug(text) unittest {
		Text t;
		t = Text("123456789abcdABCD");
		
		assert(t.indexOf(Text("1"), CASE_SENSITIVE.YES, START.HEAD) == 0);
		assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.YES, START.HEAD) == 9);
		
		assert(t.indexOf(Text("1"), CASE_SENSITIVE.NO, START.HEAD) == 0);
		assert(t.indexOf(Text("12"), CASE_SENSITIVE.NO, START.HEAD) == 0);
		assert(t.indexOf(Text("ABCD"), CASE_SENSITIVE.NO, START.HEAD) == 9);
		assert(t.indexOf(Text("123456789abcdABCD"), CASE_SENSITIVE.NO, START.HEAD) == 0);
		assert(t.indexOf(Text("123456789abcdABCDE"), CASE_SENSITIVE.NO, START.HEAD) == -1);
		
		assert(t.indexOf(Text("12"), CASE_SENSITIVE.YES, START.TAIL) == 0);
		assert(t.indexOf(Text("23"), CASE_SENSITIVE.YES, START.TAIL) == 1);
		assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.YES, START.TAIL) == 9);
		assert(t.indexOf(Text("ABCD"), CASE_SENSITIVE.NO, START.TAIL) == 13);
		assert(t.indexOf(Text("abcd"), CASE_SENSITIVE.NO, START.TAIL) == 13);
	}


	
	/**
	文字列分割。

	Examples:
	* --------------------------------
	* Text t;
	* Text[] list;
	*
	* t = Text("123 456 789");
	* list = t.split(' ');
	* assert(list.length == 3);
	* assert(list[0].text == "123");
	* assert(list[1].text == "456");
	* assert(list[2].text == "789");
	*
	* t = Text("123,456,789,");
	* list = t.split(',');
	* assert(list.length == 4);
	* assert(list[0].text == "123");
	* assert(list[1].text == "456");
	* assert(list[2].text == "789");
	* assert(list[3].text == "");
	*
	* t = Text("");
	* list = t.split(',');
	* assert(!list.length);
	*
	* t = Text("123456789");
	* list = t.split(',');
	* assert(!list.length);
	*
	* t = Text("あ＠い＠う＠え＠お");
	* list = t.split('＠');
	* assert(list.length == 5);
	* 
	* --------------------------------
	
	
	History:
		1.00β14:
			分割が変だったのを修正。
	*/
	const Text[] split(in tchar c) {
		immutable Count=count(c);
		if(!Count) {
			return null;
		}
		auto list = new Text[Count+1];
		for(auto i=0, j=0, k=0; i < text.length; i++) {
			if(text[i] == c) {
				list[k++] = Text(text[j..i]);
				j = i + 1;
			} else if(k == Count) {
				// 最後
				list[k] = Text(text[j..$]);
				break;
			}
		}
		return list;
	}
	debug(text) unittest {
		Text t;
		Text[] list;
		
		t = Text("123 456 789");
		list = t.split(' ');
		assert(list.length == 3);
		assert(list[0].text == "123");
		assert(list[1].text == "456");
		assert(list[2].text == "789");
		
		t = Text("123,456,789,");
		list = t.split(',');
		assert(list.length == 4);
		assert(list[0].text == "123");
		assert(list[1].text == "456");
		assert(list[2].text == "789");
		assert(list[3].text == "");
		
		t = Text("");
		list = t.split(',');
		assert(!list.length);
		
		t = Text("123456789");
		list = t.split(',');
		assert(!list.length);
		
		t = Text("あ＠い＠う＠え＠お");
		list = t.split('＠');
		assert(list.length == 5);
	}
	/**
	文字列分割。

	Examples:
	* --------------------------------
	* Text t;
	* Text[] list;
	*
	* t = Text("123..456..789");
	* list = t.split(Text(".."));
	* assert(list.length == 3);
	* assert(list[0].text == "123");
	* assert(list[1].text == "456");
	* assert(list[2].text == "789");
	*
	* t = Text("123....456..789");
	* list = t.split(Text(".."));
	* assert(list.length ==4);
	* assert(list[0].text == "123");
	* assert(list[1].text == "");
	* assert(list[2].text == "456");
	* assert(list[3].text == "789");
	*
	* t = Text("");
	* list = t.split(Text(".."));
	* assert(!list.length);
	*
	* t = Text("テキスト文字列text文字列てきすと");
	* list = t.split(Text("文字列"));
	* assert(list.length == 3);
	* assert(list[0].text == "テキスト");
	* assert(list[1].text == "text");
	* assert(list[2].text == "てきすと");
	*
	* t = Text("abcdefg");
	* list = t.split(Text(".."));
	* assert(!list.length);
	* --------------------------------

	History:
		1.100:
			[S] 属性変更。
			[B] 保持している文字列長が0でRange violation。
			[B] 一致しない場合に戻り値がある。

		1.00β14:
			新規作成。
	*/
	const Text[] split(in Text Word) {
		if(!text.length) {
			return null;
		}
		
		immutable WordLength = Word.length;

		if(WordLength == 1) {
			return split(Word[0]);
		}

		Text[] list;
		auto app=appender(&list);
		for(auto i=0, j=0; i < text.length - WordLength+1; i++) {
			if(text[i .. i + WordLength] == Word.text) {
				if(j == i) {
					//list ~= emptyText;
					app.put(Text.init);
				} else {
					//list ~= Text(text[j..i]);
					app.put(Text(text[j..i]));
				}
				j = i + WordLength;
				i += WordLength-1;
			} else if(i >= text.length - WordLength && text[i-WordLength-1 .. i-1] == Word.text) {
				// 最後
				//list ~= Text(text[j..$]);
				app.put(Text(text[j..$]));
				break;
			}
		}

		return list;
	}
	debug(text) unittest {
		Text t;
		Text[] list;
		
		t = Text("123..456..789");
		list = t.split(Text(".."));
		assert(list.length == 3);
		assert(list[0].text == "123");
		assert(list[1].text == "456");
		assert(list[2].text == "789");
		
		t = Text("123....456..789");
		list = t.split(Text(".."));
		assert(list.length ==4);
		assert(list[0].text == "123");
		assert(list[1].text == "");
		assert(list[2].text == "456");
		assert(list[3].text == "789");
		
		t = Text("");
		list = t.split(Text(".."));
		assert(!list.length);
		
		t = Text("テキスト文字列text文字列てきすと");
		list = t.split(Text("文字列"));
		assert(list.length == 3);
		assert(list[0].text == "テキスト");
		assert(list[1].text == "text");
		assert(list[2].text == "てきすと");
		
		t = Text("abcdefg");
		list = t.split(Text(".."));
		assert(!list.length);
	}
	
	/**
	改行分割。

	Note:
		std.string.splitlines。
	
	History:
		1.100:
			新規作成。
	*/
	const Text[] splitlines() {
		return Texts(std.string.splitlines(text));
	}


	/**
	文字列置き換え。

	Textそのものに作用。
	
	History:
		1.100:
			新規作成。
	*/
	ref Text replaceRaw(in Text SrcWord, in Text NewWord) {
		if(!SrcWord.length || SrcWord == NewWord) {
			return this;
		}

		for(auto i=0; i < text.length; ) {
			//auto FindIndex=this[i .. this.length].find(SrcWord);
			auto FindIndex=this[i .. this.length].indexOf(SrcWord, CASE_SENSITIVE.YES, START.HEAD);
			if(FindIndex == -1) {
				// もう見つかんない
				return this;
			}
			if(NewWord.length) {
				//std.array.replace(text, FindIndex+i, FindIndex + i + SrcWord.length, cast(wchar[])NewWord.text);
				text = text[0 .. FindIndex+i] ~ NewWord.text ~ text[FindIndex + i + SrcWord.length .. $];
				i += FindIndex + NewWord.length;
			} else {
				text = text[0 .. FindIndex+i] ~ text[FindIndex + i + SrcWord.length .. $];
				i += FindIndex + SrcWord.length;
			}
		}

		return this;
	}
	debug(text) unittest {
		auto t1 = Text("abcdefghi");
		t1.replaceRaw(Text("b"), Text("B"));
		assert(t1.text == "aBcdefghi", t1.toString);
		t1.replaceRaw(Text("aBc"), Text("ABC"));
		assert(t1.text == "ABCdefghi", t1.toString);
		t1 = Text("abc abc abc");
		t1.replaceRaw(Text("a"), Text("A"));
		assert(t1.text == "Abc Abc Abc", t1.toString);
		t1.replaceRaw(Text(" "), Text(""));
		assert(t1.text == "AbcAbcAbc", t1.toString);
		t1.replaceRaw(Text("Abc"), Text("[ABC]"));
		assert(t1.text == "[ABC][ABC][ABC]", t1.toString);
	}
	/**
	文字列置き換え。
	
	History:
		1.100:
			新規作成。
	*/
	const ref Text replace(in Text SrcWord, in Text NewWord) {
		auto ResultText=dup;
		return ResultText.replaceRaw(SrcWord, NewWord);
	}
	debug(text) unittest {
		auto t1=Text("abcdefghi");
		auto t2 = Text("def");
		auto t3 = Text("DEF");
		assert(t1.replace(t2, t3) == Text("abcDEFghi"));
		assert(t1.replace(Text("def"), t3) == Text("abcDEFghi"));
		assert(t1.replace(Text("def"), Text("")) == Text("abcghi"));
	}
	/**
	文字列挿入。
	
	Textそのものに作用。
	
	History:
		1.100:
			新規作成。
	*/
	ref Text insertRaw(in Text Word, size_t Index) {
		std.array.insert(text, Index, Word.text);
		return this;
	}
	/**
	文字列挿入。

	History:
		1.100:
			新規作成。
	*/
	const ref Text insert(in Text Word, size_t Index) {
		auto ResultText=dup;
		return ResultText.insertRaw(Word, Index);
	}
	debug(text) unittest {
		auto t=Text("0123456789");
		assert(t.insert(Text("abc"), 0) == Text("abc0123456789"));
		assert(t.insert(Text("abc"), 10) == Text("0123456789abc"));
		assert(t.insert(Text("abc"), 9) == Text("012345678abc9"));
	}

	/**
	文字列を繰り返す。
	
	History:
		1.100:
			[P] Repeatが一回の場合にSrcを返すのではなくSrcのコピーを返す用に変更。
	
		1.00β14:
			新規作成。
	*/
	static ref Text repeat(in Text Src, size_t Repeat) {
		switch(Repeat) {
			case 0:
				return Text.init;
			case 1:
				return Src.dup;
			default:
				auto TotalText=new Text;

				TotalText.text.length = Src.length * Repeat;
				for(auto i=0; i < TotalText.text.length; i += Src.length) {
					TotalText.text[i .. i+Src.length] = Src.text;
				}

				return *TotalText;
		}
	}
	/**
	文字を繰り返す。
	
	History:
		1.100:
			新規作成。
	*/
	static ref Text repeat(tchar c, size_t Repeat) {
		return repeat(Text(c), Repeat);
	}

	/**
	現在の文字列を繰り返す。
	
	History:
		1.050:
			新規作成。
	*/
	const ref Text repeat(size_t Repeat) {
		return repeat(this, Repeat);
	}
	debug(text) unittest {
		assert(Text.repeat(Text('-'), 10).text == "----------");
		assert(Text.repeat(Text('-'), 0).text == "");
		assert(Text.repeat(Text('-'), 1).text == "-");
		assert(Text.repeat('-', 10).text == "----------");
		assert(Text.repeat('あ', 4).text == "ああああ");
		Text t;
		t = Text("abc");
		assert(Text.repeat(t, 0) == Text(``));
		assert(Text.repeat(t, 1) == t);
		auto t2=Text.repeat(t, 1);
		assert(t == t2);
		assert(t.text !is t2.text);
		t = Text("abc");
		assert(t.repeat(2).text == "abcabc");
	}

	/**
	History:
		1.061:
			[S] 属性変更。
	*/
	const ref Text quot(T)(in T left, in T right) {
		return Text(Text(left) ~ Text(text) ~ Text(right));
	}
	
	/**
	History:
		1.061:
			[S] 属性変更。
	*/
	const ref Text quot(T)(in T arg) {
		return quot(arg, arg);
	}
	debug(text) unittest {
		Text t="abc";
		assert(t.quot("'").text == `'abc'`);
	}
	/**
	文字列は数値か。

	Params:
		Separator = セパレータを無視するか。

	Note:
		std.string.isNumeric。
	*/
	const bool isNumeric(bool Separator=true) {
		return std.string.isNumeric(strip.str, Separator);
	}
	
	debug(text) unittest {
		Text t;
		t=Text("1");
		assert(t.isNumeric);
		t=Text(" 1");
		assert(t.isNumeric);
		t=Text("a");
		assert(!t.isNumeric);
	}

	/**
	フィルタ。

	フィルタに引っ掛かる文字を省く。

	Params:
		Characters = フィルタとする文字。

	Returns:
		Charactersを取り除いたText。
	*/
	const ref Text filter(in tchar[] Characters)
	in {
		assert(Characters.length);
	}
	body {
		size_t i, j;
		auto Result=new tchar[length];

		while(j < Result.length) {
			immutable Target=text[j];

			bool Hit=false;
			foreach(c; Characters) {
				if(c == Target) {
					Hit = true;
					break;
				}
			}

			if(Hit) {
				j++;
			} else {
				Result[i++] = text[j++];
			}
		}

		return Text.wrap(Result[0..i]);
	}
	/// ditto
	const ref Text filter(tchar Character, in tchar[] Characters ...) {
		return filter(Character ~ Characters);
	}
	debug(text) unittest {
		assert(Text("a_b_c").filter(['_']) == Text("abc"));
		assert(Text("a b c").filter([' ']) == Text("abc"));
		assert(Text(" _ _ ").filter(['_', ' ']) == Text(""));
		assert(Text(" _ _ ").filter('_', ' ') == Text(""));
		assert(Text(" _ _ ").filter('_') == Text("   "));
	}

	/**
	先頭文字列比較。

	文字列の先頭が指定した文字列で始まっているかを調べる。

	Params:
		Word = 比較したい文字列。
	
	History:
		1.100:
			新規作成。
	*/
	const bool startsWith(in Text Word) {
		if(!Word.length) {
			return false;
		}
		return std.algorithm.startsWith(text, Word.text) == 1;
	}
	debug(text) unittest {
		auto t=Text("abcdef");
		assert(t.startsWith(Text("abc")));
		assert(t.startsWith(Text("abcdef")));
		assert(!t.startsWith(Text("")));
		assert(!t.startsWith(Text("Abcdefg")));
		assert(!t.startsWith(Text("abcdefg")));
	}
	
	/**
	末尾文字列比較。

	文字列の末尾が指定した文字列で終わっているか調べる。

	Params:
		Word = 比較したい文字列。
	
	History:
		1.100:
			新規作成。
	*/
	const bool endsWith(in Text Word) {
		/+
		if(!Word.length) {
			return false;
		}
		if(length < Word.length) {
			return false;
		}

		return text[$-Word.length .. $] == Word.text;
		+/
		if(!Word.length) {
			return false;
		}
		return std.algorithm.endsWith(text, Word.text) == 1;
	}
	debug(text) unittest {
		auto t=Text("abcdef");
		assert(t.endsWith(Text("def")));
		assert(t.endsWith(Text("abcdef")));
		assert(!t.endsWith(Text("")));
		assert(!t.endsWith(Text("abcdefg")));
	}
}
import etc.negui.system.utility.array; // Error: struct etc.negui.system.text.Text no size yet for forward reference
import etc.negui.system.exception;

///
ref Text toText(in string s) {
	return Text(s);
}
///
ref Text toText(in wstring s) {
	return Text(s);
}
///
ref Text toText(in dstring s) {
	return Text(s);
}

/**
Text/string配列作成。

見返すと使用頻度が多いコードだったんで作成。

History:
	1.00β14:
		[F] 引数がnull時にnullを返すように変更。
*/
Text[] Texts(T...)(in T args) {
	if(!args.length) {
		return null;
	}

	auto ts=new Text[args.length];

	foreach(i, arg; args) {
		ts[i] = Text(arg);
	}

	return ts;
}
debug(text) unittest {
	auto t=Texts(1, 2, "3", '4', 5.0, "", "789");
	assert(t[0].text == "1");
	assert(t[1].text == "2");
	assert(t[2].text == "3");
	assert(t[3].text == "4");
	assert(t[4].text == "5");
	assert(t[5].text == "");
}
/// ditto
Text[] Texts(T)(in T[] args) {
	if(!args.length) {
		return null;
	}

	auto ts=new Text[args.length];

	for(auto i=0; i < ts.length; i++) {
		ts[i] = Text(args[i]);
	}

	return ts;
}
/// ditto
string[] Strings(in Text[] texts) {
	if(!texts.length) {
		return null;
	}
	auto ss=new string[texts.length];

	for(auto i=0; i < ss.length; i++) {
		ss[i] = texts[i].text8();
	}

	return ss;
}
debug(text) unittest {
	Text[] ts=[
		Text("1"),
		Text("2"),
		Text("3")
	];
	assert(Strings(ts) == ["1", "2", "3"]);
}
debug(text) unittest {
	Text t=0;
	assert(t.text == `0`);
}

///
interface IText {
	const Text toText();
}

/**
History:
	1.100:
		[S] package移動(etc.negui.window.dialog.system -> etc.negui.system.text)。
*/
interface ITitleText {
	/**
	*/
	const Text text();
	/**
	*/
	void text(in Text);
}

/**
History:
	1.100:
		新規作成。
*/
version(ITEXT) mixin template MixInITextToString(T) {
	static if(is(T: IText)) {
		override string toString() {
			return toText.str;
		}
	} else static if(is(T: Object)) {
		static assert(false);
	} else {
		const string toString() {
			return toText.str;
		}
	}
}

/**
History:
	1.00β14:
		新規作成。
*/
ref Text join(in Text[] texts, in Text sep) {
	immutable TextsLength = texts.length;

	if(!TextsLength) {
		return Text("");//Text.emptyText();
	}

	auto ret=new Text;
	auto SepLength=sep.length;

	size_t TotalLength;
	for(auto i=0; i < TextsLength; i++) {
		TotalLength += texts[i].length;
	}
	TotalLength += (SepLength * (TextsLength - 1));

	ret.text.length = TotalLength;
	size_t StartPos;

	for(auto i=0; i < TextsLength; i++) {
		if(i && SepLength) {
			ret.text[StartPos .. StartPos+SepLength] = sep.text;
			StartPos += SepLength;
		}
		if(auto TextLength=texts[i].length) {
			ret.text[StartPos .. StartPos+TextLength] = texts[i].text.dup;
			StartPos += texts[i].length;
		}
	}

	return *ret;
}
debug(text) unittest {
	assert(Texts("aaa", "bbb", "ccc").join(Text("")).text =="aaabbbccc");
	assert(Texts("aaa", "bbb", "ccc").join(Text("+")).text =="aaa+bbb+ccc");
	assert(Texts("", "bbb", "").join(Text("*")).text =="*bbb*");
	assert(Texts("", "", "", "").join(Text("*")).text =="***");
}

/**
History:
	1.100:
		名前変更(GetNoneDouble->GetUnique)。

	1.090:
		新規作成というかnemuxi.gui.window.dialog.settingdialog.groupdialogから独立。
*/
ref Text GetUnique(in Text[] TextArray, in Text Src)
in {
	assert(TextArray.length);
	assert(Src.length);
}
body {
	scope ChangeName=Src.dup;
	size_t n=1;
	for(auto i=0; i < TextArray.length; i++) {
		if(TextArray[i] == ChangeName) {
			ChangeName = Text("%s(%s)", Src, ++n);
			i=-1;
			continue;
		}
	}
	return ChangeName.dup;
}
debug(text) unittest {
	auto ts=Texts(
		"abc",
		"def",
		"ABC",
		"DEF",
		"123",
		"456",
		"あ",
		"あ(1)",
		"あ(2)",
		"あ(3)"
	);
	assert(GetUnique(ts, Text("abc")) == Text("abc(2)"));
	assert(GetUnique(ts, Text("def")) == Text("def(2)"));
	assert(GetUnique(ts, Text("XYZ")) == Text("XYZ"));
	assert(GetUnique(ts, Text("あ")) == Text("あ(4)"));
}




/**

NULL結合+終端NULL二つの文字列郡を作成。
Windowsへの橋渡し用。

Bugs:
	汎用性を持たせるべき。

History:
	1.100:
		新規作成。
*/
struct TEXTNULLLIST {
	Text[] datas;
	alias datas this;

	static ref TEXTNULLLIST opCall(in tchar* Data) {
		auto TextNullList=new TEXTNULLLIST;

		auto p=Data;
		size_t i, j;
		do {
			if(!p[i]) {
				TextNullList.datas ~= Text(p[j]);
				j = ++i;
			}
		} while(p[i++]);

		return *TextNullList;
	}
	debug(text) unittest {
		auto tnl=TEXTNULLLIST("abc\0def\0efg\0\0"w.ptr);
		assert(tnl.length == 3, Text(tnl.length).str);
	}

	const size_t dataLength() {
		size_t Length=datas.length+1;

		foreach(ref const(Text) data; datas) {
			Length += data.length;
		}

		return Length;
	}

	const void* ptr() {
		if(!datas.length) {
			return null;
		}

		size_t Length=dataLength;
		auto DataList=new wchar[Length];
		for(auto i=0,j=0; i < Length-1; j++) {
			auto Size=i + datas[j].length;
			DataList[i .. Size] = datas[j].text;
			i = Size;
			DataList[i++] = 0;
		}
		DataList[$-1] = 0;

		return cast(void*)DataList;
	}
}

