﻿/**
タイマーというか時間関係。

std.timeがあまりにもあれだったんで置き換え。

*/
module etc.negui.system.time;

debug import std.stdio: wl = writefln;
debug(time) void main() {}

public import std.date;
import std.contracts;
	import std.conv;

import win32.windows;

public import etc.negui.system.text;
public import etc.negui.system.exception;
//mport etc.negui.system.error;
import etc.negui.system.meta.memberproperty;
import etc.negui.system.utility.cmp;
import etc.negui.system.utility.number;

///
class TimeException: NeGuiException {
	mixin MixInNeGuiException;
}

///
class FileDateTimeException: TimeException {
	mixin MixInNeGuiException;
}
/**

History:
	1.100:
		新規作成。
*/
struct FILEDATETIME {
	union {
		FILETIME FileTime;
		long FullTime;
	}
	mixin(SMixInStructHiddenOriginal!(FILETIME)(q{FileTime}));

	
	mixin(SMixInStructGetSet!(DWORD)("low", q{FileTime.dwLowDateTime}));
	mixin(SMixInStructGetSet!(DWORD)("high", q{FileTime.dwHighDateTime}));
	mixin(SMixInStructGetSet!(long)("full", q{FullTime}));


	/***/
	static ref SYSTEMDATETIME toSystemDateTime(const ref FILEDATETIME FileDateTime) {
		auto SystemDateTime=new SYSTEMDATETIME;
		
		enforce(FileTimeToSystemTime(cast(FILETIME*)&FileDateTime, SystemDateTime.ptr), new FileDateTimeException(Text("toSystemDateTime")));
		
		return *SystemDateTime;
	}
	/***/
	const ref SYSTEMDATETIME toSystemDateTime() {
		return toSystemDateTime(this);
	}
	
	static ref FILEDATETIME toLocalFileDateTime(const ref FILEDATETIME FileDateTime) {
		auto LocalFileDateTime=new FILEDATETIME;
		enforce(FileTimeToLocalFileTime(FileDateTime.ptr, LocalFileDateTime.ptr), new FileDateTimeException(Text("toLocalFileDateTime")));
		return *LocalFileDateTime;
	}
	const ref FILEDATETIME toLocalFileDateTime() {
		return toLocalFileDateTime(this);
	}
	static ref FILEDATETIME toSystemFileDateTime(const ref FILEDATETIME FileDateTime) {
		auto SystemFileDateTime=new FILEDATETIME;
		enforce(LocalFileTimeToFileTime(FileDateTime.ptr, SystemFileDateTime.ptr), new FileDateTimeException(Text("toSystemFileDateTime")));
		return *SystemFileDateTime;
	
	}
	const ref FILEDATETIME toSystemFileDateTime() {
		return toSystemFileDateTime(this);
	}
	
	
	/***/
	const bool opEquals(const ref FILEDATETIME FileDateTime) {
		return FullTime == FileDateTime.FullTime;
	}
	/***/
	const int opCmp(const ref FILEDATETIME FileDateTime) {
		return CompareFileTime(this.ptr, FileDateTime.ptr);
	}
}

///
class SystemDateTimeException: TimeException {
	mixin MixInNeGuiException;
}
/**
History:
	1.100:
		新規作成。
*/
struct SYSTEMDATETIME {
	SYSTEMTIME SystemTime;
	mixin(SMixInStructHiddenOriginal!(SYSTEMTIME)(q{SystemTime}));

	mixin(SMixInStructGetSet!(WORD)("year",   q{SystemTime.wYear}));
	mixin(SMixInStructGetSet!(WORD)("month",  q{SystemTime.wMonth}));
	mixin(SMixInStructGetSet!(WORD)("day",    q{SystemTime.wDay}));
	mixin(SMixInStructGetSet!(WORD)("week",   q{SystemTime.wDayOfWeek}));
	mixin(SMixInStructGetSet!(WORD)("hour",   q{SystemTime.wHour}));
	mixin(SMixInStructGetSet!(WORD)("minute", q{SystemTime.wMinute}));
	mixin(SMixInStructGetSet!(WORD)("second", q{SystemTime.wSecond}));
	mixin(SMixInStructGetSet!(WORD)("ms",     q{SystemTime.wMilliseconds}));

	/***/
	const ref FILEDATETIME toFileDateTime(const ref SYSTEMDATETIME SystemDateTime) {
		auto FileDateTime=new FILEDATETIME;

		enforce(SystemTimeToFileTime(cast(SYSTEMTIME*)SystemDateTime.ptr, FileDateTime.ptr), new SystemDateTimeException(Text("toFileDateTime")));

		return *FileDateTime;
	}
	/***/
	const ref FILEDATETIME toFileDateTime() {
		return toFileDateTime(this);
	}

	/***/
	static ref SYSTEMDATETIME localTime() {
		auto SystemDateTime=new SYSTEMDATETIME;
		
		GetLocalTime(SystemDateTime.ptr);

		return *SystemDateTime;
	}

	
	/***/
	const bool opEquals(const ref SYSTEMDATETIME SystemDateTime) {
		return toFileDateTime == SystemDateTime.toFileDateTime;
	}
	/***/
	const int opCmp(const ref SYSTEMDATETIME SystemDateTime) {
		return Cmp(toFileDateTime, SystemDateTime.toFileDateTime);
	}

	/***/
	const Text toText() {
		return Text(
			"%s-%02s-%02sT%02s:%02s:%02s.%03s",
			SystemTime.wYear,
			SystemTime.wMonth,
			SystemTime.wDay,
			SystemTime.wHour,
			SystemTime.wMinute,
			SystemTime.wSecond,
			SystemTime.wMilliseconds
		);
	}
	/***/
	const string toString() {
		return toText.str;
	}
	/// ditto
	string toString() {
		return toText.str;
	}

	/// 最小/最大値
	enum RANGE: WORD {
		MONTH_MIN  = 1,   /// 月
		MONTH_MAX  = 12,  /// ditto
		DAY_MIN    = 1,   /// 日
		DAY_MAX    = 31,  /// ditto
		WEEK_MIN   = 0,   /// 曜日
		WEEK_MAX   = 6,   /// ditto
		HOUR_MIN   = 0,   /// 時
		HOUR_MAX   = 23,  /// ditto
		MINUTE_MIN = 0,   /// 分
		MINUTE_MAX = 59,  /// ditto
		SECOND_MIN = 0,   /// 秒
		SECOND_MAX = 59,  /// ditto
		MS_MIN     = 0,   /// ミリ秒
		MS_MAX     = 999, /// ditto
	}
	private static pure nothrow string IsRange(string Method, string RANGE) {
		return `
			static pure nothrow bool ` ~ Method ~ `(WORD Value) {
				return RANGE.` ~ RANGE ~ `_MIN <= Value && Value <= RANGE.` ~ RANGE ~ `_MAX;
			}
		`;
	}
	
	/**
	年として合法か。

	0年ならfalseとする。
	*/
	static pure nothrow bool isYear(WORD Year) {
		return Year != 0;
	}
	mixin(IsRange("isMonth",  "MONTH"));
	mixin(IsRange("isDay",    "DAY"));
	mixin(IsRange("isWeek",   "WEEK"));
	mixin(IsRange("isHour",   "HOUR"));
	mixin(IsRange("isMinute", "MINUTE"));
	mixin(IsRange("isSecond", "SECOND"));
	mixin(IsRange("isMs",     "MS"));

	/**
	閏年判定。

	In:
		isYear(Year)
	*/
	static pure nothrow bool isLeapYear(WORD Year)
	in {
		assert(isYear(Year));
	}
	body {
		return Year % 4 == 0 && (Year % 100 || !(Year % 400));
	}
	debug(time) unittest {
		assert(!isLeapYear(1));
		assert(!isLeapYear(2));
		assert(!isLeapYear(3));
		assert(isLeapYear(4));
		assert(isLeapYear(1996));
		assert(isLeapYear(2000));
		assert(isLeapYear(2004));
		assert(!isLeapYear(2100));
	}

	/// 曜日。
	enum WEEK: WORD {
		SUNDAY = 0, /// 日。
		MONDAY,     /// 月。
		TUESDAY,    /// 火。
		WEDNESDAY,  /// 水。
		THURSDAY,   /// 木。
		FRIDAY,     /// 金。
		SATURDAY,   /// 土。
	}
	/**
	曜日の取得。

	Returns:
		WEEK。

	Standards:
		Zellerの公式
	
	See_Also:
		ウィキペディア, http://ja.wikipedia.org/wiki/%E3%83%84%E3%82%A7%E3%83%A9%E3%83%BC%E3%81%AE%E5%85%AC%E5%BC%8F

	In:
		isYear(Year) && isMonth(Month) && isDay(Day)
	*/
	static pure nothrow WEEK getWeekDay(in WORD Year, in WORD Month, in WORD Day)
	in {
		assert(isYear(Year));
		assert(isMonth(Month));
		assert(isDay(Day));
	}
	out(r) {
		assert(isWeek(r));
	}
	body {
		immutable MonthPrev = Month > 2;
		immutable YEAR  = MonthPrev ? Year:  Year - 1;
		immutable MONTH = MonthPrev ? Month: Month + 12;
		//return cast(WORD)((YEAR + floor(YEAR / 4) - floor(YEAR / 100) + floor(YEAR / 400) + floor(2.6 * MONTH + 1.6) + Day) % 7);
		return cast(WEEK)((YEAR + (YEAR / 4) - (YEAR / 100) + (YEAR / 400) + (2.6 * MONTH + 1.6) + Day) % 7);
	}
	debug(time) unittest {
		assert(getWeekDay(2000, 1, 1) == 6);
		assert(getWeekDay(2000, 1, 31) == 1);
		assert(getWeekDay(2000, 2, 1) == 2);
		assert(getWeekDay(2000, 2, 28) == 1);
		assert(getWeekDay(2000, 2, 29) == 2);
		assert(getWeekDay(2000, 3, 1) == 3);
	}

	/**
	月の日数。
	
	In:
		isYear(Year) && isMonth(Month)
	*/
	static pure nothrow WORD getMonthDay(WORD Year, WORD Month)
	in {
		assert(isYear(Year));
		assert(isMonth(Month));
	}
	out(r) {
		assert(isWeek(r));
	}
	body {
		final switch(Month) {
			case 1, 3, 5, 7, 8, 10, 12:
				return 31;
			case 4, 6, 9, 11:
				return 30;
			case 2:
				return isLeapYear(Year) ? 29: 28;
		}
	}
	debug(time) unittest {
		assert(getMonthDay(2000, 1) == 31);
		assert(getMonthDay(2000, 4) == 30);
		assert(getMonthDay(2000, 2) == 29);
		assert(getMonthDay(2001, 2) == 28);
	}

	/**
	次の月を取得。

	Params:
		NowMonth = 現在の月。

	Returns:
		来月。
	
	In:
		isMonth(NowMonth)
	*/
	static pure nothrow WORD nextMonth(WORD NowMonth)
	in {
		assert(isMonth(NowMonth));
	}
	body {
		if(NowMonth == 12) {
			return 1;
		} else {
			return cast(WORD)(NowMonth + 1);
		}
	}
	debug(time) unittest {
		WORD[] src =[1, 2, 3, 4, 5, 6, 7, 8,  9, 10, 11, 12];
		WORD[] next=[2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,  1];
		for(WORD i=0; i < src.length; i++) {
			assert(nextMonth(src[i]) == next[i]);
		}
	}
	/**
	前の月を取得。

	Params:
		NowMonth = 現在の月。

	Returns:
		先月。
	
	In:
		isMonth(NowMonth)
	*/
	static pure nothrow WORD prevMonth(WORD NowMonth)
	in {
		assert(isMonth(NowMonth));
	}
	body {
		if(NowMonth == 1) {
			return 12;
		} else {
			return cast(WORD)(NowMonth - 1);
		}
	}
	debug(time) unittest {
		WORD[] next=[1, 2, 3, 4, 5, 6, 7, 8,  9, 10, 11, 12];
		WORD[] src =[2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,  1];
		for(WORD i=0; i < src.length; i++) {
			assert(prevMonth(src[i]) == next[i]);
		}
	}
	/**
	次の日を取得。

	Params:
		Year  = 現在の年。
		Month = 現在の月。
		Day   = 現在の日。

	Returns:
		次の日。
	
	In:
		isYear(Year) && isMonth(Month) && isDay(Day)
	*/
	static pure nothrow WORD nextDay(WORD Year, WORD Month, WORD Day)
	in {
		assert(isYear(Year));
		assert(isMonth(Month));
		assert(isDay(Day));
	}
	body {
		// 月末判定
		immutable LastDay=getMonthDay(Year, Month);
		if(Day >= LastDay) {
			return 1;
		} else {
			return cast(WORD)(Day + 1);
		}
	}
	debug(time) unittest {
		assert(nextDay(2000, 1, 1) == 2);
		assert(nextDay(2000, 1, 31) == 1);
		assert(nextDay(2000, 2, 28) == 29);
		assert(nextDay(2000, 2, 29) == 1);
		assert(nextDay(2001, 2, 28) == 1);
	}
	/**
	前の日を取得。

	Params:
		Year  = 現在の年。
		Month = 現在の月。
		Day   = 現在の日。

	Returns:
		前の日。
	
	In:
		isYear(Year) && isMonth(Month) && isDay(Day)
	*/
	static pure nothrow WORD prevDay(WORD Year, WORD Month, WORD Day)
	in {
		assert(isYear(Year));
		assert(isMonth(Month));
		assert(isDay(Day));
	}
	body {
		// 月初判定
		if(Day == 1) {
			return getMonthDay(Year, prevMonth(Month));
		} else {
			return cast(WORD)(Day - 1);
		}
	}
	debug(time) unittest {
		assert(prevDay(2000, 12, 1) == 30);
		assert(prevDay(2000, 1, 31) == 30);
		assert(prevDay(2000, 2, 29) == 28);
		assert(prevDay(2000, 3, 1) == 29);
		assert(prevDay(2001, 3, 1) == 28);
	}
}
debug(time) unittest {
	SYSTEMDATETIME s;
	s.year = 2000;
	s.month= 2;
	s.day= 2;
	s.second = 2;
	auto f=s.toFileDateTime();
	assert(f.FullTime == f.toSystemDateTime.toFileDateTime.FullTime);
}


///
class DateTimeException: SystemDateTimeException {
	mixin MixInNeGuiException;
}

/**
日時を扱う専用クラス。

ﾈﾑぃに特化。

Bugs:
	調子乗りすぎた。
	もうしんどい、いやや。

History:
	1.100:
		[S] 全面修正。
	
	1.00β19:
		[S] Timeを継承。
*/
class DateTime: IText {
	invariant() {
		if((MASK.YEAR & mask) == MASK.YEAR) {
			assert(SYSTEMDATETIME.isYear(SystemDateTime.year));
		}
		if((MASK.MONTH & mask) == MASK.MONTH) {
			assert(SYSTEMDATETIME.isMonth(SystemDateTime.month));
		}
		if((MASK.DAY & mask) == MASK.DAY) {
			assert(SYSTEMDATETIME.isDay(SystemDateTime.day));
		}
		if((MASK.WEEK & mask) == MASK.WEEK) {
			assert(SYSTEMDATETIME.isWeek(SystemDateTime.week));
		}
		if((MASK.HOUR & mask) == MASK.HOUR) {
			assert(SYSTEMDATETIME.isHour(SystemDateTime.hour));
		}
		if((MASK.MINUTE & mask) == MASK.MINUTE) {
			assert(SYSTEMDATETIME.isMinute(SystemDateTime.minute));
		}
		if((MASK.SECOND & mask) == MASK.SECOND) {
			assert(SYSTEMDATETIME.isSecond(SystemDateTime.second));
		}
	}
	
	protected SYSTEMDATETIME SystemDateTime;
	mixin(SMixInStructHiddenOriginal!(SYSTEMDATETIME)(q{SystemDateTime}));

	/***/
	protected const ref Text FormatSingle(in Text FormatText, in Text TargetToken, int Width, int Value) {
		if(Width) {
			Text Data;
			if(GetDigits(Value) >= Width) {
				Data = Text(Value);
				Data = Data[Data.length - Width .. Data.length];
			} else {
				Data = Text("%0.*s", Width, Value);
			}
			return FormatText.replace(TargetToken, Data);//sub(FormatTarget, RegToken, s);
		} else {
			return FormatText.replace(TargetToken, Text(Value));
		}
	}
	/***/
	protected const ref Text FormatDate(in Text FormatText, MASK Mask) {
		auto Data=FormatText.dup;

		if((MASK.YEAR & Mask) == MASK.YEAR) {
			Data = FormatSingle(Data, Text("[Y]"),    0, SystemDateTime.year);
			Data = FormatSingle(Data, Text("[YY]"),   2, SystemDateTime.year);
			Data = FormatSingle(Data, Text("[YYYY]"), 4, SystemDateTime.year);
		}
		if((MASK.MONTH & Mask) == MASK.MONTH) {
			Data.replaceRaw(Text("[@M]"), monthName());
			
			Data = FormatSingle(Data, Text("[M]"),  0, SystemDateTime.month);
			Data = FormatSingle(Data, Text("[MM]"), 2, SystemDateTime.month);
		}
		if((MASK.DAY & Mask) == MASK.DAY) {
			Data = FormatSingle(Data, Text("[D]"),  0, SystemDateTime.day);
			Data = FormatSingle(Data, Text("[DD]"), 2, SystemDateTime.day);
		}
		if((MASK.WEEK & Mask) == MASK.WEEK) {
			Data.replaceRaw(Text("[@W]"), weekName());
			
			Data = FormatSingle(Data, Text("[W]"),  0, SystemDateTime.week);
		}
		if((MASK.HOUR & Mask) == MASK.HOUR) {
			Data = FormatSingle(Data, Text("[h]"),  0, SystemDateTime.hour);
			Data = FormatSingle(Data, Text("[hh]"), 2, SystemDateTime.hour);
		}
		if((MASK.MINUTE & Mask) == MASK.MINUTE) {
			Data = FormatSingle(Data, Text("[m]"),  0, SystemDateTime.minute);
			Data = FormatSingle(Data, Text("[mm]"), 2, SystemDateTime.minute);
		}
		if((MASK.SECOND & Mask) == MASK.SECOND) {
			Data = FormatSingle(Data, Text("[s]"),  0, SystemDateTime.second);
			Data = FormatSingle(Data, Text("[ss]"), 2, SystemDateTime.second);
		}
		if((MASK.MS & Mask) == MASK.MS) {
			Data = FormatSingle(Data, Text("[ms]"),  0, SystemDateTime.ms);
			Data = FormatSingle(Data, Text("[MS]"),  3, SystemDateTime.ms);
		}

		return Data.dup;
	}

	/**
	Note:
		取り敢えずべた書き。
	*/
	protected static Throwable TogetherAssignment(ref SYSTEMDATETIME SystemDateTime) {
		/+
		Exception BaseExp=new DateTimeException(Text(`TogetherAssignment`));
		Throwable e = BaseExp;
		+/
		auto tc=THROWABLECHAIN(new DateTimeException(Text(`TogetherAssignment`)));

		// ミリ秒
		if(!SYSTEMDATETIME.isMs(SystemDateTime.ms)) {
			tc.chain(new DateTimeException(Text("ms -> %s", SystemDateTime.ms)));
			
			size_t NextUp=0;
			do {
				SystemDateTime.ms = cast(WORD)(SystemDateTime.ms - (SYSTEMDATETIME.RANGE.MS_MAX+1));
				NextUp++;
			} while(!SYSTEMDATETIME.isMs(SystemDateTime.ms));
			SystemDateTime.second = cast(WORD)(SystemDateTime.second + NextUp);
		}

		// 秒
		if(!SYSTEMDATETIME.isSecond(SystemDateTime.second)) {
			tc.chain(new DateTimeException(Text("second -> %s", SystemDateTime.second)));

			size_t NextUp=0;
			
			do {
				SystemDateTime.second = cast(WORD)(SystemDateTime.second - (SYSTEMDATETIME.RANGE.SECOND_MAX+1));
				NextUp++;
			} while(!SYSTEMDATETIME.isSecond(SystemDateTime.second));
			SystemDateTime.minute = cast(WORD)(SystemDateTime.minute + NextUp);
		}

		// 分
		if(!SYSTEMDATETIME.isMinute(SystemDateTime.minute)) {
			tc.chain(new DateTimeException(Text("minute -> %s", SystemDateTime.minute)));

			size_t NextUp=0;
			
			do {
				SystemDateTime.minute = cast(WORD)(SystemDateTime.minute - (SYSTEMDATETIME.RANGE.MINUTE_MAX+1));
				NextUp++;
			} while(!SYSTEMDATETIME.isMinute(SystemDateTime.minute));
			SystemDateTime.hour = cast(WORD)(SystemDateTime.hour + NextUp);
		}

		// 時間
		if(!SYSTEMDATETIME.isHour(SystemDateTime.hour)) {
			tc.chain(new DateTimeException(Text("hour -> %s", SystemDateTime.hour)));

			size_t NextUp=0;
			
			do {
				SystemDateTime.hour = cast(WORD)(SystemDateTime.hour - (SYSTEMDATETIME.RANGE.HOUR_MAX+1));
				NextUp++;
			} while(!SYSTEMDATETIME.isHour(SystemDateTime.hour));
			SystemDateTime.day = cast(WORD)(SystemDateTime.day + NextUp);
		}

		void Month() {
			// 月
			if(SYSTEMDATETIME.isMonth(SystemDateTime.month)) {
				auto LastDay=SYSTEMDATETIME.getMonthDay(SystemDateTime.year, SystemDateTime.month);
				if(SystemDateTime.day > LastDay) {
					tc.chain(new DateTimeException(Text("month/day -> %s/%s", SystemDateTime.month, SystemDateTime.day)));
					SystemDateTime.day = LastDay;
				}
			} else {
				tc.chain(new DateTimeException(Text("month -> %s", SystemDateTime.month)));
				
				SystemDateTime.year  = cast(WORD)(SystemDateTime.month / 12 + SystemDateTime.year);
				SystemDateTime.month = cast(WORD)(SystemDateTime.month % 12);
				if(!SystemDateTime.month) {
					SystemDateTime.month = 12;
				}

				auto LastDay=SYSTEMDATETIME.getMonthDay(SystemDateTime.year, SystemDateTime.month);
				if(SystemDateTime.day > LastDay) {
					tc.chain(new DateTimeException(Text("month/day -> %s/%s", SystemDateTime.month, SystemDateTime.day)));
					SystemDateTime.day = LastDay;
				}
			}
		}
		
		// 日
		void Day() {
			auto Check = (WORD LastDay) {
				SystemDateTime.day = cast(WORD)(SystemDateTime.day - LastDay);
				auto NextMonth = SYSTEMDATETIME.nextMonth(SystemDateTime.month);
				if(NextMonth != SystemDateTime.month) {
					tc.chain(new DateTimeException(Text("NextMonth != month -> %s != %s", NextMonth, SystemDateTime.month)));

					if(NextMonth == 1) {
						SystemDateTime.year = cast(WORD)(SystemDateTime.year + 1);
					}
					SystemDateTime.month = NextMonth;
				}
			};
			if(SYSTEMDATETIME.isDay(SystemDateTime.day)) {
				auto LastDay=SYSTEMDATETIME.getMonthDay(SystemDateTime.year, SystemDateTime.month);
				if(LastDay < SystemDateTime.day) {
					// 次の月
					tc.chain(new DateTimeException(Text("LastDay < day -> %s < %s", LastDay, SystemDateTime.day)));
					
					Check(LastDay);
				}
			} else {
				tc.chain(new DateTimeException(Text("day -> %s", SystemDateTime.day)));
				
				do {
					Check(SYSTEMDATETIME.getMonthDay(SystemDateTime.year, SystemDateTime.month));
				} while(SystemDateTime.day > SYSTEMDATETIME.getMonthDay(SystemDateTime.year, SystemDateTime.month));
			}
		}

		if(!SYSTEMDATETIME.isYear(SystemDateTime.year)) {
			tc.chain(new DateTimeException(Text("year -> %s", SystemDateTime.year)));
			SystemDateTime.year = 1;
		}
		if(!SYSTEMDATETIME.isMonth(SystemDateTime.month) && !SYSTEMDATETIME.isDay(SystemDateTime.day)) {
			tc.chain(new DateTimeException(Text("month -> %s, day -> %s", SystemDateTime.month, SystemDateTime.day)));
			SystemDateTime.month = SYSTEMDATETIME.RANGE.MONTH_MIN;
			SystemDateTime.day   = SYSTEMDATETIME.RANGE.DAY_MIN;
		}

		if(SYSTEMDATETIME.isDay(SystemDateTime.day)) {
			Month();
			Day();
		} else {
			Day();
			Month();
		}

		/+
		if(BaseExp.next) {
			return BaseExp;
		} else {
			return null;
		}
		+/
		return tc();
	}

	/***/
	protected void Initialize(MODE Mode) {
		format = Text("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss]");
		mask   = MASK.DATE | MASK.TIME;
		mode   = Mode;
		months = Texts("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
		weeks  = Texts("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
	}

	/**
	History:
		1.100:
			[P] maskがMASK.ALLであればSYSTEMDATETIME.toTextと等価。
	*/
	override const Text toText() {
		if(mask == MASK.ALL && format == Text("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[MS]")) {
			return SystemDateTime.toText;
		} else {
			return FormatDate(format, mask);
		}
	}
	version(ITEXT) {
		mixin MixInITextToString!(DateTime);
	} else {
		override string toString() {
			return toText.str;
		}
	}
	
	protected this(MODE Mode, const ref SYSTEMDATETIME SystemDateTime, bool Check) {
		Initialize(Mode);

		if(Check) {
			auto sdt=SystemDateTime;
			auto e=TogetherAssignment(sdt);
			ModeChange(sdt,e);
		} else {
		this.SystemDateTime = SystemDateTime;
			ChangeWeek();
		}
	}
	/**
	現在時刻からDateTimeの生成。

	Params:
		Mode = 日時設定時の挙動。
	*/
	this(MODE Mode=MODE.FLEXIBLE) {
		this(Mode, SYSTEMDATETIME.localTime, false);
	}
	/**
	指定日時からDateTimeの生成。
	
	Params:
		Mode = 日時設定時の挙動。

		SystemDateTime = 設定する日時情報。

	Throws:
		ModeがMODE.STRICTでSystemDateTimeが不正な場合にDateTimeException。
	*/
	this(MODE Mode, const ref SYSTEMDATETIME SystemDateTime) {
		this(Mode, SystemDateTime, true);
	}

	/**
	指定日時からDateTimeの生成。
	
	Params:
		Mode = 日時設定時の挙動。

		FileDateTime = 設定する日時情報。

	Throws:
		ModeがMODE.STRICTでFileDateTimeが不正な場合にDateTimeException。
	*/
	this(MODE Mode, const ref FILEDATETIME FileDateTime) {
		this(Mode, FileDateTime.toSystemDateTime(), true);
	}
	/**
	指定日時からDateTimeの生成。
	
	Params:
		Mode = 日時設定時の挙動。

		Year   = 年。

		Month  = 月。

		Day    = 日。

		Hour   = 時。

		Minute = 分。

		Second = 秒。

		Ms     = ミリ秒。

	Throws:
		ModeがMODE.STRICTで引数が不正な場合にDateTimeException。
	*/
	this(MODE Mode, WORD Year, WORD Month=SYSTEMDATETIME.RANGE.MONTH_MIN, WORD Day=SYSTEMDATETIME.RANGE.DAY_MIN, WORD Hour=SYSTEMDATETIME.RANGE.HOUR_MIN, WORD Minute=SYSTEMDATETIME.RANGE.MINUTE_MIN, WORD Second=SYSTEMDATETIME.RANGE.SECOND_MIN, WORD Ms=SYSTEMDATETIME.RANGE.MS_MIN) {
		SYSTEMDATETIME SystemDateTime;
		with(SystemDateTime) {
			year   = Year;
			month  = Month;
			day    = Day;
			hour   = Hour;
			minute = Minute;
			second = Second;
			ms     = Ms;
		}
		this(Mode, SystemDateTime, true);
	}

	/**
	文字列からDateTimeの生成。
	
	Params:
		Mode = 日時設定時の挙動。

		Value = 日時情報に変換可能な文字列。
		        YYYY-MM-DDThh:mm:ss.MS形式(0埋め不要)で指定。
		        
		        年以外は順々に省略可能。

	Throws:
		Valueを分解できない場合にDateTimeException。
		ModeがMODE.STRICTで引数が不正な場合にDateTimeException。
	*/
	this(MODE Mode, in Text Value) {
		enforce(Value.length, new DateTimeException(Text("Value.length -> 0")));

		Text Date, Time;
		if(auto DateAndTime=Value.split('T')) {
			// 日付と時間。
			enforce(DateAndTime.length == 2, new DateTimeException(Text("Value -> %s, Value.split('T') -> %s", Value, DateAndTime)));
			
			Date = DateAndTime[0];
			Time = DateAndTime[1];
		} else {
			// 日付のみ。
			Date = Value.dup;
		}
		enforce(Date.length, new DateTimeException(Text("Date.length -> 0")));
		
		Text Year, Month, Day;
		if(Date.count('-')) {
			auto ymd=Date.split('-');
			Year = ymd[0];
			enforce(ymd.length <= 3, new DateTimeException(Text("[YYYY-MM-DD].split('-') <= 3, [YYY-MM-DD] -> %s", ymd)));
			if(ymd.length >= 2) {
				Month = ymd[1];
			}
			if(ymd.length == 3) {
				Day = ymd[2];
			}
		} else {
			// 年のみ
			Year = Date;
		}
		// 年は必須
		enforce(Year.length, new DateTimeException(Text("Year.length -> 0")));

		Text Hour, Minute, Second, Ms;
		if(Time.length) {
			// 年月日は確定していないとワケ分からん。
			enforce(Year.length && Month.length && Day.length, new DateTimeException(Text("Year.length -> %s, Month.length -> %s, Day.length -> %s", Year, Month, Day)));
			if(Time.count(':')) {
				auto hms=Time.split(':');
				Hour = hms[0];
				enforce(hms.length <= 3, new DateTimeException(Text("[hh:mm-ss.ms].split(':') <= 3, [hh:mm-ss.ms] -> %s", hms)));
				if(hms.length >= 2) {
					Minute = hms[1];
				}
				if(hms.length == 3) {
					// ミリ秒はある？
					if(hms[2].count('.')) {
						auto SecondAndMs=hms[2].split('.');
						enforce(SecondAndMs.length == 2, new DateTimeException(Text("SecondAndMs.length == 2, SecondAndMs -> %s", SecondAndMs)));
						Second = SecondAndMs[0];
						Ms     = SecondAndMs[1];
					} else {
						Second = hms[2];
					}
				}
			} else {
				// 時間のみ
				Hour = Time;
			}
		}

		// 代入
		SYSTEMDATETIME SystemDateTime;
		auto tc=THROWABLECHAIN(new DateTimeException(Text("DateTime Assign")));
		with(SystemDateTime) {
			bool Result;
			assert(Year.length);
			year   = tc.collect(Year.toNumber!(WORD), Result);

			WORD AssignThis(in Text Value, RANGE Min) {
				WORD result;
				if(Value.length) {
					result = tc.collect(Value.toNumber!(WORD), Result);
				}
				if(result < Min) {
					result = Min;
				}
				return result;
			}
			month  = AssignThis(Month,  RANGE.MONTH_MIN);
			day    = AssignThis(Day,    RANGE.DAY_MIN);
			hour   = AssignThis(Hour,   RANGE.HOUR_MIN);
			minute = AssignThis(Minute, RANGE.MINUTE_MIN);
			second = AssignThis(Second, RANGE.SECOND_MIN);
			ms     = AssignThis(Ms,     RANGE.MS_MIN);
			/+
			if(Month .length) {
				month = tc.collect(Month.toNumber!(WORD), Result);
			}
			if(Day.length) {
				day = tc.collect(Day.toNumber!(WORD), Result);
			}
			if(Hour.length) {
				hour = tc.collect(Hour.toNumber!(WORD), Result);
			}
			if(Minute.length) {
				minute = tc.collect(Minute.toNumber!(WORD), Result);
			}
			if(Second.length) {
				second = tc.collect(Second.toNumber!(WORD), Result);
			}
			if(Ms.length) {
				ms = tc.collect(Ms.toNumber!(WORD), Result);
			}
			+/
		}
		if(tc()) {
			throw tc();
		}
		try {
			this(Mode, SystemDateTime, true);
		} catch(Throwable e) {
			tc.chain(e);
			throw tc();
		}
	}
	debug(time) unittest {
		DateTime dt;
		Text t, f;
		f = Text("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[MS]");
		
		t = Text("2010-02-10T19:30:55.999");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-10T19:30:55.999"), dt.toString);
		
		t = Text("2010-02-10T19:30:55");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-10T19:30:55.000"), dt.toString);
		
		t = Text("2010-02-10T19:30");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-10T19:30:00.000"), dt.toString);
		
		t = Text("2010-02-10T19");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-10T19:00:00.000"), dt.toString);
		
		t = Text("2010-02-10");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-10T00:00:00.000"), dt.toString);
		
		t = Text("2010-02");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-02-01T00:00:00.000"), dt.toString);
		
		t = Text("2010");
		dt = new DateTime(MODE.STRICT, t);
		dt.format = f;
		dt.mask   = DateTime.MASK.ALL;
		assert(dt.toText == Text("2010-01-01T00:00:00.000"), dt.toString);
	}

	/**
	日時書式。
	
	<table>
		<thead>
			<tr>
				<th>変換前</th>
				<th>変換後</th>
			</tr>
		</thead>
		<tbody>
			<tr>
				<td>[Y]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[YY]</td>
				<td>12345 -> 45, 1234 -> 34, 123 -> 23, 12 -> 12, 1 -> 01</td>
			</tr>
			<tr>
				<td>[YYYY]</td>
				<td>12345 -> 2345, 1234 -> 1234, 123 -> 0123, 12 -> 0012, 1 -> 0001</td>
			</tr>
			<tr>
				<td>[@M]</td>
				<td>DateTime.monthName()</td>
			</tr>
			<tr>
				<td>[M]</td>
				<td>1 -> 1, 9 -> 9 10 -> 10</td>
			</tr>
			<tr>
				<td>[MM]</td>
				<td>1 -> 01, 9 -> 09, 10 -> 10</td>
			</tr>
			<tr>
				<td>[D]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[DD]</td>
				<td>12345 -> 45, 1234 -> 34, 123 -> 23, 12 -> 12, 1 -> 01</td>
			</tr>
			<tr>
				<td>[@W]</td>
				<td>DateTime.weekName()</td>
			</tr>
			<tr>
				<td>[W]</td>
				<td>0 -> 0, 3 -> 3, 6 -> 6</td>
			</tr>
			<tr>
				<td>[h]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[hh]</td>
				<td>12345 -> 45, 1234 -> 34, 123 -> 23, 12 -> 12, 1 -> 01</td>
			</tr>
			<tr>
				<td>[m]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[mm]</td>
				<td>12345 -> 45, 1234 -> 34, 123 -> 23, 12 -> 12, 1 -> 01</td>
			</tr>
			<tr>
				<td>[s]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[ss]</td>
				<td>12345 -> 45, 1234 -> 34, 123 -> 23, 12 -> 12, 1 -> 01</td>
			</tr>
			<tr>
				<td>[ms]</td>
				<td>12345 -> 12345, 1234 -> 1234, 123 -> 123, 12 -> 12, 1 -> 1</td>
			</tr>
			<tr>
				<td>[MS]</td>
				<td>12345 -> 345, 1234 -> 234, 123 -> 123, 12 -> 012, 1 -> 001</td>
			</tr>
		</tbody>
	</table>
	*/
	Text format;
	/// 月の文字列。
	Text[12] months;
	/// 曜日の文字列。
	Text[7]  weeks;

	/**
	日時情報から現在の月名を取得。
	*/
	const const(Text) monthName() {
		return months[SystemDateTime.month-1];
	}
	/**
	日時情報から現在の曜日名を取得。
	*/
	const const(Text) weekName() {
		return weeks[SystemDateTime.week];
	}

	/**
	文字列変換時の変換フラグ。

	該当しない箇所は文字列変換されない。
	*/
	enum MASK {
		YEAR   = 0b_00000001, /// 年。
		MONTH  = 0b_00000010, /// 月。
		DAY    = 0b_00000100, /// 日。
		WEEK   = 0b_00001000, /// 曜日。
		HOUR   = 0b_00010000, /// 時。
		MINUTE = 0b_00100000, /// 分。
		SECOND = 0b_01000000, /// 秒。
		MS     = 0b_10000000, /// ミリ秒。

		DATE   = YEAR | MONTH | DAY,     /// 年・月・日。
		DATES  = DATE | WEEK,            /// 年・月・日・曜日。
		TIME   = HOUR | MINUTE | SECOND, /// 時・分・秒。
		TIMES  = TIME | MS,              /// 時・分・秒・ミリ秒。

		ALL    = DATES | TIMES, /// 全て。
	}
	/**
	変換フラグ。
	*/
	MASK mask;

	/**
	日時設定時の動作。
	*/
	enum MODE {
		FLEXIBLE, /// あふれる場合は次に進めたりする。
		STRICT,   /// 問答無用で例外。
	}
	/// 挙動
	MODE mode;

	private void ModeChange(string file=__FILE__, int line=__LINE__)(const ref SYSTEMDATETIME sdt, Throwable e) {
		if(mode == MODE.FLEXIBLE) {
			SystemDateTime = sdt;
		} else {
			assert(mode == MODE.STRICT);
			if(e) {
				throw e;
			}
			SystemDateTime = sdt;
		}
		ChangeWeek();
	}

	/***/
	protected void ChangeWeek() {
		SystemDateTime.week = SYSTEMDATETIME.getWeekDay(SystemDateTime.year, SystemDateTime.month, SystemDateTime.day);
	}

	/**
	日時情報から現在の曜日を取得。
	*/
	@property const SYSTEMDATETIME.WEEK week() {
		return cast(SYSTEMDATETIME.WEEK)SystemDateTime.week;
	}

	private static pure nothrow string Property(string Method, string InStatement=null) {
		return `
			mixin(SMixInClassGet!(WORD)("` ~ Method ~ `", q{SystemDateTime.` ~ Method ~ `}, false));
		
			@property WORD ` ~ Method ~ `(WORD Value) {
				` ~ (InStatement.length ? InStatement: "") ~ `

				auto sdt=SystemDateTime;
				sdt.` ~ Method ~ ` = Value;
				
				auto e=TogetherAssignment(sdt);
				ModeChange(sdt, e);

				return SystemDateTime.` ~ Method ~ `;
			}
		`;
	}

	mixin(Property("year", q{enforce(SYSTEMDATETIME.isYear(Value), new DateTimeException(Text("year -> %s", Value)));}));
	mixin(Property("month"));
	mixin(Property("day"));
	mixin(Property("hour"));
	mixin(Property("minute"));
	mixin(Property("second"));
	mixin(Property("ms"));
	debug(time) unittest {
		auto t=new DateTime(MODE.STRICT, 2000,1,2,3,4,5);
		assert(t.second = 10 == 10);
	}
}
/***/
struct TIMESTAMP {
	///
	enum FLAG {
		NONE   = 0,
		CREATE = 0b_0001,
		ACCESS = 0b_0010,
		UPDATE = 0b_0100,

		ALL    = CREATE | ACCESS | UPDATE,
	}
	///
	FLAG flags=FLAG.ALL;
	
	DateTime create; /// 作成日時。
	DateTime access; /// アクセス日時。
	DateTime update; /// 更新日時。

	///
	const Text toText() {
		return Text(
			"Create -> %s, Access -> %s, Update -> %s, flags -> %03b",
			create, access, update, flags
		);
	}
}

debug {
import std.perf;
	scope class DTimer: PerformanceCounter {
		this() {
			start();
		}

		~this() {
			stop();
			wl("Timer:%s", microseconds);
		}
	}
}

